/* Copyright (c) 2000-2003 ActiveState SRL. 
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Generic string query dialog.
 *
 * This dialog allows you to ask the user for a single string input.
 *
 * Features:
 *  - OK and Cancel buttons.
 *  - Optional leading prompt.
 *  - Configurable dialog title and textbox label.
 *  - Dialog is resizable and it remembers its dimensions.
 *  - Can have an MRU for the textbox.
 *  - Can specify a validator for the value.
 *
 * Usage:
 *  All dialog interaction is done via an object passed in and out as the first
 *  window argument: window.arguments[0]. All these arguments are optional.
 *      .title          the dialog title
 *      .prompt         a leading <description/> to the textbox
 *      .label          the textbox's label
 *      .value          default value for textbox
 *      .mruName        if set this will be used to identify an MRU preference
 *                      set and an MRU will be provided
 *      .validator      A callable object to validate the current value.  It
 *                      will be called with the current value and should return
 *                      true iff the value is acceptable.
 *  On return window.arguments[0] has:
 *      .retval         "OK" or "Cancel" indicating how the dialog was exitted
 *  and iff .retval == "OK":
 *      .value          is the content of the textbox on exit.
 *
 */

//---- globals

var _gDimPrefPrefix = "queryStringDialog";
var _gValidator = null; // Function to validate entered value.
var _gUsingMRU = false;



//---- internal support routines

// Restores the dialog's dimensions from prefs prefixed with the given
// prefix. If no preference is set then the default is used.
//
//    "prefix" is the pref prefix to use for restoring from prefs.
//    "default*" are the defaults to use. If null, then no default it set.
//
// Returns true if restored from prefs, false if any defaults were used.
//
function _RestoreWindowDimensions(prefix, defaultWidth, defaultHeight,
                                  defaultX, defaultY)
{
    var retval = true;
    var outerHeight=null, outerWidth=null, screenX=null, screenY=null;
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].getService();
    if (prefSvc.prefs.hasLongPref(prefix+"_outerWidth")) {
        outerWidth = prefSvc.prefs.getLongPref(prefix+"_outerWidth");
    } else {
        outerWidth = defaultWidth;
        retval = false;
    }
    if (prefSvc.prefs.hasLongPref(prefix+"_outerHeight")) {
        outerHeight = prefSvc.prefs.getLongPref(prefix+"_outerHeight");
    } else {
        outerHeight = defaultHeight;
        retval = false;
    }
    if (prefSvc.prefs.hasLongPref(prefix+"_screenX")) {
        screenX = prefSvc.prefs.getLongPref(prefix+"_screenX");
    } else {
        screenX = defaultX;
        retval = false;
    }
    if (prefSvc.prefs.hasLongPref(prefix+"_screenY")) {
        screenY = prefSvc.prefs.getLongPref(prefix+"_screenY");
    } else {
        screenY = defaultY;
        retval = false;
    }
    if (outerWidth != null) window.outerWidth = outerWidth;
    if (outerHeight != null) window.outerHeight = outerHeight;
    if (screenX != null) window.screenX = screenX;
    if (screenY != null) window.screenY = screenY;
    
    return retval;
}


function _SaveWindowsDimensions(prefix)
{
    // Save Window location.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].getService();
    prefSvc.prefs.setLongPref(prefix+"_outerHeight", window.outerHeight);
    prefSvc.prefs.setLongPref(prefix+"_outerWidth", window.outerWidth);
    prefSvc.prefs.setLongPref(prefix+"_screenY", window.screenY);
    prefSvc.prefs.setLongPref(prefix+"_screenX", window.screenX);
}



//---- interface routines for queryString.xul

function OnLoad()
{
    // .title
    if (typeof window.arguments[0].title != "undefined" &&
        window.arguments[0].title != null) {
        window.title = window.arguments[0].title;
    } else {
        window.title = "Query";
    }

    // .prompt
    var descWidget = document.getElementById("querystring-prompt");
    var desc = window.arguments[0].prompt;
    if (typeof desc != "undefined" && desc != null) {
//        descWidget.setAttribute("collapsed", "false");
//        descWidget.removeAttribute("collapsed");
        var textNode = document.createTextNode(desc);
        descWidget.appendChild(textNode);
    } else {
        descWidget.setAttribute("collapsed", "true");
    }

    // .label
    var label = window.arguments[0].label;
    var labelWidget = document.getElementById("querystring-label");
    if (typeof label == "undefined" || label == null) {
        labelWidget.setAttribute("collapsed", "true");
    } else {
        labelWidget.setAttribute("value", label);
    }

    // .value
    var value = window.arguments[0].value;
    if (typeof value == "undefined" || value == null) {
        value = "";
    }
    var textboxWidget = document.getElementById("querystring-textbox");
    textboxWidget.setAttribute("value", value);

    // .mruName
    var mruName = window.arguments[0].mruName;
    if (typeof mruName != "undefined" && mruName != null) {
        textboxWidget.setAttribute("komodo-preference-set", mruName+"Mru");
        // Need to delay this until "komodo-preference-set" is properly set.
        var dummy = new MRUAutoCompleteSession(textboxWidget);
        _gUsingMRU = true;
    }

    // .validator
    var validator = window.arguments[0].validator;
    if (typeof validator != "undefined" && validator != null) {
        _gValidator = validator;
    }

    _RestoreWindowDimensions(_gDimPrefPrefix, null, null, 100, 100);

    // Need to "sizeToContent" in a timeout because it seems ineffectual
    // when run here right away.
    setTimeout('window.sizeToContent(); document.getElementById("querystring-textbox").focus();', 0);
}


function OnUnload()
{
    if (typeof window.arguments[0].retval == "undefined") {
        // This happens when "X" window close button is pressed.
        window.arguments[0].retval = "Cancel";
    } else if (window.arguments[0].retval == "OK") {
        var textboxWidget = document.getElementById("querystring-textbox");
        window.arguments[0].value = textboxWidget.value;
        if (_gUsingMRU) {
            MRU_updateFromWidget(textboxWidget);
        }
    }

    _SaveWindowsDimensions(_gDimPrefPrefix);
    opener.focus();
}


function OK()
{
    window.arguments[0].retval = "OK";
    var textboxWidget = document.getElementById("querystring-textbox");
    if (!_gValidator || _gValidator(textboxWidget.value)) {
        window.close();
    } else {
        textboxWidget.focus();
    }
}


function Cancel()
{
    window.arguments[0].retval = "Cancel";
    window.close();
}


