/* Copyright (c) 2000-2001 ActiveState SRL. 
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Komodo's Find and Replace dialog (see find.xul as well)
 *
 * The desired starting tab is passed in via "gFindDialogTabName" (it should
 * be one of "find" or "replace") global variable set in the openning window.
 * 'window.arguments' cannot be used because it is not set when the dialog
 * is launched with an existing instance.
 *
 * The core find functionality is in find_functions.js. Just the GUI stuff
 * resides here.
 *
 * TODO:
 *  - find and replace in files
 *  - shortcut keys (they don't work in the current mozilla)
 *  - key binding for Ctrl-H and Ctrl-F on find dialog to jump to the right tab
 *  - hook up a help button?
 *  - other regex syntaxes?
 *  - regex help menulist as in Rx and DevStudio7 find and replace
 *  - more search contexts
 *  - look at perf for large files (both forward and back)
 *  - interpret the inputBuffer better, capture and appropriate handle a "tab",
 *    say, for launching the replace dialog
 */


//---- globals

var parentWindow = opener;    // a hook back to the parent window

var findSvc = null; // the find XPCOM service that does all the grunt work
var _gFindContext; // the context in which to search

// variables set by SwitchToTab as appropriate for the current tab
var tabIdPrefix = null;         // e.g. for the find tab this is 'ftab',
                                // similar widgets on the different tabs
                                // have the same id with a different prefix
var patternWidget = null;       // the find pattern
var replacementWidget = null;   // the replacement string
var contextGroupWidget = null;  // the context radiogroup
var contextCurrDocWidget = null;        // context: current document
var contextSelectionWidget = null;      // context: current selection
var contextAllOpenDocsWidget = null;    // context: all open documents
var useMenulistWidget = null;   // what kind of search pattern
var caseMenulistWidget = null;
var matchWordWidget = null;
var ignoreHiddenWidget = null;
var searchBackwardWidget = null;

// find context types
var FCT_CURRENT_DOC = null;
var FCT_SELECTION = null;
var FCT_ALL_OPEN_DOCS = null;


//--- internal support routines

function _IsWordCharacter(ch)
{
    return /^\w/.test(ch);
}


function _IdPrefix2TabName(idPrefix)
{
    var tabName = null;
    if (idPrefix == "ftab") {
        tabName = "find";
    } else if (idPrefix == "rtab") {
        tabName = "replace";
    } else {
        throw("unrecognized find dialog id prefix: '" + idPrefix + "'");
    }
    return tabName
}


function _TabName2IdPrefix(tabName)
{
    var idPrefix = null;
    if (tabName == "find") {
        idPrefix = "ftab";
    } else if (tabName == "replace") {
        idPrefix = "rtab";
    } else {
        throw("unrecognized find dialog tab name: '" + tabName + "'");
    }
    return idPrefix
}


function _Initialize()
{
    SwitchToTab(parentWindow.gFindDialogTabName);

    FCT_CURRENT_DOC = Components.interfaces.koIFindContext.FCT_CURRENT_DOC;
    FCT_SELECTION = Components.interfaces.koIFindContext.FCT_SELECTION;
    FCT_ALL_OPEN_DOCS = Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS;

    //---- Preload some of the UI fields as appropriate
    // If there is selected text then preload the find pattern with it. Unless
    // it spans a line, then set the search context to the selection.
    var scimoz = parentWindow.gLastFocusedViewManager.getCurrentView()._scimoz;
    var selectedText = scimoz.selText;
    var defaultPattern = "";
    var preferredContextType = findSvc.options.preferredContextType;
    if (preferredContextType == FCT_ALL_OPEN_DOCS) {
        // See http://bugs.activestate.com/Komodo/show_bug.cgi?id=20003
        preferredContextType = FCT_CURRENT_DOC;
    }
    if (preferredContextType == FCT_CURRENT_DOC) {
        contextGroupWidget.selectedItem = contextCurrDocWidget;
    } else {
        throw("unexpected preferredContextType preference setting: " +
              preferredContextType);
    }
    if (selectedText != "") {
        // If the selected text has newline characters in it or *is* an entire
        // line (without the end-of-line) then "search within selection".
        var curLineObj = new Object;
        scimoz.getCurLine(curLineObj);
        var curLine = curLineObj.value;
        if (selectedText.search(/\n/) != -1
            || selectedText == curLine.substring(0,curLine.length-1))
        {
            contextGroupWidget.selectedItem = contextSelectionWidget;
            // If a user does a search within a selection then the "prefered"
            // context is the current document (rather than across multiple
            // docs).
            findSvc.options.preferredContextType = FCT_CURRENT_DOC;
        }
        // Otherwise, use the current selection as the search pattern.
        else {
            defaultPattern = selectedText;
        }
    }

    UpdateContext();

    // If a default pattern was not found from a selection then try
    // to parse out a whitespace delimited token from the current cursor
    // position.
    if (defaultPattern == "") {
        defaultPattern = GetWordUnderCursor(scimoz); // in interpolation_functions.js
    }
    
    // Preload with input buffer contents if any and then give focus to the
    // pattern textbox.
    // Notes:
    //   - The pattern textbox will automatically select all its contents
    //     on focus. If there are input buffer contents then we do *not*
    //     want this to happen, because this will defeat the purpose of
    //     the input buffer if the user is part way through typing in
    //     characters.
    //   - Have to set focus in a timer because this could be called within
    //     an onfocus handler, in which Mozilla does not like .focus() calls.
    var inputBufferContents = parentWindow.InputBuffer_Finish();
    if (inputBufferContents) {
        patternWidget.setAttribute('value', inputBufferContents);
        setTimeout('patternWidget.focus(); patternWidget.setSelectionRange(-1,-1);', 0);
    } else {
        patternWidget.setAttribute('value', defaultPattern);
        setTimeout('patternWidget.focus();', 0);
    }

    _UpdateOptionsUi();

    // The act of opening the find dialog should reset the find session. This
    // is the behaviour of least surprise.
    //logDebug("Reset find session, because opening the find dialog.");
    gFindSession.Reset();
}


function _UpdateOptionsUi()
{
    var opts = findSvc.options;
    var i;
    //XXX Is there a way to select the right menulist item without hardcoding 3 here?
    //XXX Yes, use the DOM and check the number of children
    switch (opts.caseSensitivity) {
    case opts.FOC_INSENSITIVE:
        for (i = 0; i < 3; i++) {
            caseMenulistWidget.selectedIndex = i;
            if (caseMenulistWidget.value == "insensitive") {
                break;
            }
        }
        break;
    case opts.FOC_SENSITIVE:
        for (i = 0; i < 3; i++) {
            caseMenulistWidget.selectedIndex = i;
            if (caseMenulistWidget.value == "sensitive") {
                break;
            }
        }
        break;
    case opts.FOC_SMART:
        for (i = 0; i < 3; i++) {
            caseMenulistWidget.selectedIndex = i;
            if (caseMenulistWidget.value == "smart") {
                break;
            }
        }
        break;
    }

    //XXX Is there a way to select the right menulist item without hardcoding 3 here?
    //XXX Yes, use the DOM and check the number of children
    switch (opts.patternType) {
    case opts.FOT_SIMPLE:
        for (i = 0; i < 3; i++) {
            useMenulistWidget.selectedIndex = i;
            if (useMenulistWidget.value == "simple") {
                break;
            }
        }
        break;
    case opts.FOT_WILDCARD:
        for (i = 0; i < 3; i++) {
            useMenulistWidget.selectedIndex = i;
            if (useMenulistWidget.value == "wildcard") {
                break;
            }
        }
        break;
    case opts.FOT_REGEX_PYTHON:
        for (i = 0; i < 3; i++) {
            useMenulistWidget.selectedIndex = i;
            if (useMenulistWidget.value == "regex-python") {
                break;
            }
        }
        break;
    }

    searchBackwardWidget.checked = opts.searchBackward;
    matchWordWidget.checked = opts.matchWord;
    ignoreHiddenWidget.checked = opts.ignoreHiddenText;
}



//---- interface routines for find.xul

function OnLoad()
{
    findSvc = Components.classes["@activestate.com/koFindService;1"].
              getService(Components.interfaces.koIFindService);
    _Initialize();
    SetWindowLocation();
}


function OnUnload()
{
    SaveWindowLocation();
    opener.focus();
}


function UpdateContext()
{
    switch (contextGroupWidget.selectedItem) {
    case contextCurrDocWidget:
        _gFindContext = Components.classes[
            "@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
        _gFindContext.type = FCT_CURRENT_DOC;
        findSvc.options.preferredContextType = FCT_CURRENT_DOC;
        break;
    case contextSelectionWidget:
        _gFindContext = Components.classes[
            "@activestate.com/koRangeFindContext;1"]
            .createInstance(Components.interfaces.koIRangeFindContext);
        _gFindContext.type = FCT_SELECTION;
        var scimoz = parentWindow.gLastFocusedViewManager.getCurrentView()._scimoz;
        _gFindContext.startIndex = scimoz.selectionStart;
        _gFindContext.endIndex = scimoz.selectionEnd;
        break;
    case contextAllOpenDocsWidget:
        _gFindContext = Components.classes[
            "@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
        _gFindContext.type = FCT_ALL_OPEN_DOCS;
        findSvc.options.preferredContextType = FCT_ALL_OPEN_DOCS;
        break;
    default:
        throw("unexpected current context widget setting");
    }
    _UpdateOptionsUi();
}

function UpdateOption(optName)
{
    // Update a specific find service option from the UI state.
    var opts = findSvc.options;

    switch (optName) {
    case "case":
        var caseValue = caseMenulistWidget.value;
        if (caseValue == "insensitive") {
            opts.caseSensitivity = opts.FOC_INSENSITIVE;
        } else if (caseValue == "sensitive") {
            opts.caseSensitivity = opts.FOC_SENSITIVE;
        } else if (caseValue == "smart") {
            opts.caseSensitivity = opts.FOC_SMART;
        } else {
            throw("unrecognized 'case sensitivity' menulist value: '" +
                  caseValue + "'");
        }
        break;

    case "use":
        var useValue = useMenulistWidget.value;
        if (useValue == "simple") {
            opts.patternType = opts.FOT_SIMPLE;
        } else if (useValue == "regex-python") {
            opts.patternType = opts.FOT_REGEX_PYTHON;
        } else if (useValue == "wildcard") {
            opts.patternType = opts.FOT_WILDCARD;
        } else {
            throw("unrecognized 'use' menulist value: '" + useValue + "'");
        }
        break;

    case "searchBackward":
        opts.searchBackward = searchBackwardWidget.checked;
        break;

    case "matchWord":
        opts.matchWord = matchWordWidget.checked;
        break;

    case "ignoreHiddenText":
        opts.ignoreHiddenText = ignoreHiddenWidget.checked;
        break;

    default:
        throw("Don't know how to update option '" + optName + "'");
    }
}


function SwitchToTab(tabName)
{
    // buffer values for the "from" tab
    var fromTabIdPrefix = tabIdPrefix;  // could be null on first call
    var fromPatternWidget = patternWidget; 
    var fromContextGroupWidget = contextGroupWidget;
    var fromContextCurrDocWidget = contextCurrDocWidget;
    var fromContextSelectionWidget = contextSelectionWidget;
    var fromContextAllOpenDocsWidget = contextAllOpenDocsWidget;
    var ftabfindnextbutton = document.getElementById("ftab-findnext-button");
    var rtabreplacebutton = document.getElementById("rtab-replace-button");
    
    // determine the XUL element id prefix for the new tab
    tabIdPrefix = _TabName2IdPrefix(tabName);
     
    // setup appropriate globals
    patternWidget = document.getElementById(tabIdPrefix+"-pattern");
    if (tabIdPrefix == "ftab") {
        replacementWidget = null;
        if (rtabreplacebutton.hasAttribute('default'))
            rtabreplacebutton.removeAttribute('default');
        ftabfindnextbutton.setAttribute('default', 'true');
    } else if (tabIdPrefix == "rtab") {
        replacementWidget = document.getElementById('rtab-replacement');
        if (ftabfindnextbutton.hasAttribute('default'))
            ftabfindnextbutton.removeAttribute('default');
        rtabreplacebutton.setAttribute('default', 'true');
    }
    contextGroupWidget = document.getElementById(tabIdPrefix+"-context");
    contextCurrDocWidget = document.getElementById(tabIdPrefix+"-context-currentdoc");
    contextSelectionWidget = document.getElementById(tabIdPrefix+"-context-selection");
    contextAllOpenDocsWidget = document.getElementById(tabIdPrefix+"-context-allopendocs");
    useMenulistWidget = document.getElementById(tabIdPrefix+"-use-menulist");
    caseMenulistWidget = document.getElementById(tabIdPrefix+"-case-menulist");
    searchBackwardWidget = document.getElementById(tabIdPrefix+"-search-backward");
    ignoreHiddenWidget = document.getElementById(tabIdPrefix+"-ignore-hidden");
    matchWordWidget = document.getElementById(tabIdPrefix+"-match-word");

    // copy transferable settings
    if (fromTabIdPrefix) {
        // to transfer: pattern, context, options
        patternWidget.setAttribute("value", fromPatternWidget.value);
        if (fromContextGroupWidget.selectedItem == fromContextCurrDocWidget) {
            contextGroupWidget.selectedItem = contextCurrDocWidget;
        } else if (fromContextGroupWidget.selectedItem ==
                   fromContextSelectionWidget) {
            contextGroupWidget.selectedItem = contextSelectionWidget;
        } else if (fromContextGroupWidget.selectedItem ==
                   fromContextAllOpenDocsWidget) {
            contextGroupWidget.selectedItem = contextAllOpenDocsWidget;
        } else {
            throw("unexpected find context radio group setting");
        }
        _UpdateOptionsUi();
    }

    // make the actual switch and give the "Find what" entry the focus
    var tabWidget = document.getElementById("tabs");
    tabWidget.selectedTab = document.getElementById(tabIdPrefix);
    patternWidget.focus();
}


function CheckDefaultKeys(event, onOK) {
    if (event.keyCode == 13) {
        // fire given function for "Enter" key
        onOK();
        event.preventDefault();
    }
}


function doFindNext() {
    //XXX if the context is "selection" and there isn't one, then should say so
    //XXX perhaps find_functions.js should be responsible for that
    var pattern = patternWidget.value;
    if (pattern != "") {
        MRU_updateFromWidget(patternWidget);
        var tabName = _IdPrefix2TabName(tabIdPrefix);
        var foundOne = Find_FindNext(parentWindow, _gFindContext, pattern, tabName);
        if (!foundOne) {
            // If no match was hilighted then it is likely that the user will
            // now want to enter a different pattern. (Copying Word's behaviour
            // here.)
            patternWidget.focus();
        }
    }
}


function doFindAll() {
    // Find all occurrences of the current pattern in the current context.
    var pattern = patternWidget.value;
    if (pattern != "") {
        MRU_updateFromWidget(patternWidget);
        SetBusyCursor(parentWindow, true);
        var foundSome = Find_FindAll(parentWindow, _gFindContext, pattern);
        SetBusyCursor(parentWindow, false);
        if (foundSome) {
            window.close();
        }
    }
}


function doMarkAll() {
    // Find all occurrences of the current pattern in the current context.
    var pattern = patternWidget.value;
    if (pattern != "") {
        MRU_updateFromWidget(patternWidget);
        SetBusyCursor(parentWindow, true);
        var foundSome = Find_MarkAll(parentWindow, _gFindContext, pattern);
        SetBusyCursor(parentWindow, false);
        if (foundSome) {
            window.close();
        }
    }
}


function doReplace() {
    var pattern = patternWidget.value;
    var replacement = replacementWidget.value;
    if (pattern != "") {
        MRU_updateFromWidget(patternWidget);
        if (replacement)
            MRU_updateFromWidget(replacementWidget);

        var foundOne = Find_Replace(parentWindow, _gFindContext, pattern, replacement);
        if (!foundOne) {
            // If no match was hilighted then it is likely that the user will
            // now want to enter a different pattern. (Copying Word's behaviour
            // here.)
            patternWidget.focus();
        }
    }
}

function doReplaceAll() {
    var pattern = patternWidget.value;
    var replacement = replacementWidget.value;
    if (pattern != "") {
        MRU_updateFromWidget(patternWidget);
        if (replacement)
            MRU_updateFromWidget(replacementWidget);
        SetBusyCursor(parentWindow, true);
        var foundSome = Find_ReplaceAll(parentWindow, _gFindContext, pattern,
                                        replacement);
        SetBusyCursor(parentWindow, false);
        if (foundSome) {
            window.close();
        } else {
            patternWidget.focus();
        }
    }
}


function doFindAllOrReplaceAll()
{
    var tabName = _IdPrefix2TabName(tabIdPrefix);
    if (tabName == "find") {
        doFindAll();
    } else if (tabName == "replace") {
        doReplaceAll();
    }
}


function doReplaceIfOnReplaceTab()
{
    var tabName = _IdPrefix2TabName(tabIdPrefix);
    if (tabName == "replace") {
        doReplace();
    }
}


