/* Copyright (c) 2000-2002 ActiveState Tool Corporation.
   See the file LICENSE.txt for licensing information. */

// -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
//
// The interface for using the run output window (where run command output goes
// to, in the bottom pane of the Komodo workspace).
//
// Expected usage:
//  - Someone calls RunOutput_Init() at startup and RunOutput_Fini() at
//    shutdown.
//  - When a command is to be run in the output window do this:
//      // announce intention to start session
//      RunOutput_StartSession(...);
//      //... setup and start running the actual command calling
//      //    RunOutput_GetTerminal() and RunOutput_Show() as needed
//      RunOutput_SetProcessHandle(p);  // to allow user to kill process
//      //... setup RunOutput_EndSession() to be run when the process
//      //    terminates.
//


//---- globals

var _gTerminal = null;
var _gProcess = null;
var _gParseOutput = 0;


//---- internal support routines

// Clear the terminal window and associated UI. This should be called prior to
// running a command in the output window.
//  "clearContent" (boolean) can be used to _not_ have the output
//      contents be cleared.
function _RunOutput_Clear(clearContent /* =true */)
{
    if (typeof clearContent == 'undefined') clearContent = true;

    if (clearContent) {
        _gTerminal.clear();
    }
    _RunOutput_UpdateSortIndicators(window, null);
    _gParseOutput = 0;

    var descWidget = document.getElementById("runoutput-desc");
    addStyle(descWidget, "color", "black");
    descWidget.removeAttribute("value");
    descWidget.removeAttribute("_command");
}



//---- public interface

function RunOutput_Init() {
    var scinWidget = document.getElementById("runoutput-scintilla");
    scinWidget.init();

    // Setup the style for the output window using the user's text settings
    scinWidget.ISciMoz.lexer = scinWidget.ISciMoz.SCLEX_NULL;
    var styleApplier = Components.classes['@activestate.com/koScintillaStyleApplier;1']
                       .createInstance(Components.interfaces.koIScintillaStyleApplier);
    // Force the font to be fixed width by making a preference item with that
    // single pref
    var fakeViewPrefs = Components.classes['@activestate.com/koPreferenceSet;1']
                        .createInstance(Components.interfaces.koIPreferenceSet);
    fakeViewPrefs.setBooleanPref('editUseFixedFont', true);
    // Set the style and apply it.
    styleApplier.init(scinWidget.ISciMoz, 'Text', fakeViewPrefs);
    styleApplier.applyStyles();

    // Create a terminal for the output window. The terminal has std handlers
    // that proxy read/write events between the scintilla and a spawned child
    // process.
    _gTerminal = Components.classes['@activestate.com/koRunTerminal;1']
                 .createInstance(Components.interfaces.koIRunTerminal);

    _gTerminal.setScintilla(scinWidget.ISciMoz);
    scinWidget.ISciMoz.hookEvents(_gTerminal);
    scinWidget.controllers.insertControllerAt(0, _gTerminal.controller);

    var outlinerWidget = document.getElementById("runoutput-outliner");
    outlinerWidget.boxObject
                  .QueryInterface(Components.interfaces.nsIOutlinerBoxObject)
                  .view = _gTerminal;

    _RunOutput_Clear();
}


function RunOutput_Fini()
{
    var scinWidget = document.getElementById("runoutput-scintilla");
    scinWidget.controllers.removeController(_gTerminal.controller);
    //XXX Does the scinWidget have to be finalized?
}


function RunOutput_GetTerminal()
{
    return _gTerminal;
}


// Start a terminal session in the output window with the given command.
// This raises an exception if the run output window is currently busy.
//    "command" is the command being run
//    "parseOutput" is a boolean indicating whether to parse output lines
//    "parseRegex" is the regular expression to use to parse output lines
//    "cwd" is the directory in which the command is being run
//    "filename" is the current editor filename (if any)
//    "clearContent" is a boolean indicating whether to clear the
//      output window content (by default "true", i.e. the window _is_
//      cleared).
function RunOutput_StartSession(command, parseOutput, parseRegex, cwd,
                                filename, clearContent /* =true */)
{
    if (typeof clearContent == 'undefined') clearContent = true;

    if (_gTerminal.active) {
        throw("A previous command is currently running in the "+
              "Command Output window. That command must finish before "+
              "starting another.");
    }
    _RunOutput_Clear(clearContent);

    // Setup the terminal.
    var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                       .getService(Components.interfaces.koILastErrorService);
    _gTerminal.startSession();
    _gParseOutput = parseOutput
    if (_gParseOutput) {
        try {
            _gTerminal.setParseRegex(parseRegex);
        } catch (ex) {
            _gTerminal.endSession();
            errmsg = lastErrorSvc.getLastErrorMessage();
            errmsg = "Error in regex '"+parseRegex+"': "+errmsg;
            throw(errmsg);
        }
    }
    _gTerminal.setCwd(cwd);
    if (filename) _gTerminal.setCurrentFile(filename);

    var descWidget = document.getElementById("runoutput-desc");
    //XXX Bug: This attribute change doesn't actually show up until the cursor
    //    is moved, triggering some gecko redraw. Cannot use
    //    window.sizeToContent() because that does bad (though non-fatal)
    //    things to komodo.xul.
    // XXX Maybe want to add cwd to this string?
    descWidget.setAttribute("value", "Running '" + command + "'...");
    // Store the command name for later use.
    descWidget.setAttribute("_command", command);

    // If not clearing the output window and we are not parsing output this
    // time around, do not muck with the parsed output view (no need to disturb
    // possible results from the previous command).
    if (clearContent || _gParseOutput) {
        var listButton = document.getElementById("runoutput-list-button");
        if (_gParseOutput) {
            listButton.removeAttribute("disabled");
        } else {
            listButton.setAttribute("disabled", "true");
        }

        // Determine what columns to show.
        if (_gParseOutput) {
            var fileCol = document.getElementById("runoutput-outliner-file");
            if (parseRegex.indexOf("?P<file>") == -1) {
                fileCol.setAttribute("hidden", "true");
            } else {
                fileCol.setAttribute("hidden", "false");
            }
            var lineCol = document.getElementById("runoutput-outliner-line");
            if (parseRegex.indexOf("?P<line>") == -1) {
                lineCol.setAttribute("hidden", "true");
            } else {
                lineCol.setAttribute("hidden", "false");
            }
            var columnCol = document.getElementById("runoutput-outliner-column");
            if (parseRegex.indexOf("?P<column>") == -1) {
                columnCol.setAttribute("hidden", "true");
            } else {
                columnCol.setAttribute("hidden", "false");
            }
        }
    }
}


// Complete a terminal session. The command exited with the given value.
function RunOutput_EndSession(retval)
{
    //dump("XXX RunOutput_EndSession(retval="+retval+")\n");
    _gTerminal.endSession();
    
    var descWidget = document.getElementById("runoutput-desc");
    var command = descWidget.getAttribute("_command");
    //XXX Bug: This attribute change doesn't actually show up until the cursor
    //    is moved, triggering some gecko redraw. Cannot use
    //    window.sizeToContent() because that does bad (though non-fatal)
    //    things to komodo.xul.
    var msg = "'" + command + "' returned " + retval + ".";
    if (retval != 0) {
        addStyle(descWidget, "color", "#bb0000"); // dark red to not appear "neon" against grey.
    } else {
    }
    descWidget.setAttribute("value", msg);

    _gProcess = null;
    var closeButton = document.getElementById("runoutput-close-button");
    closeButton.setAttribute("disabled", "true");
}


// Show the command output window.
//    "editor" is the XUL window holding the command output window.
//    "showParseOutputList" is a boolean indicating whether to show the
//        outliner parsed output list.
function RunOutput_Show(editor, showParsedOutputList)
{
    // open output pane in komodo.xul
    var splitterWidget = editor.document
                         .getElementById("redirect-splitter");
    splitterWidget.removeAttribute('state');
    
    // switch to the run output tab
    var tabsWidget = editor.document.getElementById("oTabbox");
    var tabWidget = editor.document.getElementById("runoutput_tab");
    tabsWidget.selectedTab = tabWidget;
    
    // open the proper command output view
    _RunOutput_SetView(editor, showParsedOutputList);
}


// Pass a koIRunProcess reference to the output window so it can manipulate the
// process that is being run in its terminal, if necessary.
function RunOutput_SetProcessHandle(process)
{
    if (_gTerminal.active) {
        _gProcess = process;
        var closeButton = document.getElementById("runoutput-close-button");
        closeButton.setAttribute("disabled", "false");
    }
}


// Kill the process currently running in the output window's terminal, if any.
function RunOutput_Kill(retval)
{
    if (_gProcess) {
        _gProcess.kill(retval);
    }
}


function _RunOutput_SetView(editor, showParsedOutputList)
{
    var deckWidget = editor.document.getElementById("runoutput-deck");
    if (showParsedOutputList) {
        deckWidget.setAttribute("index", 1);
    } else {
        deckWidget.setAttribute("index", 0);
    }
}


function RunOutput_ToggleView()
{
    var deckWidget = document.getElementById("runoutput-deck");
    if (deckWidget.getAttribute("index") == 1) {
        _RunOutput_SetView(window, 0);
    } else {
        _RunOutput_SetView(window, 1);
    }
}



//---- Run output outliner UI support routines.

function RunOutput_ScintillaOnClick(event)
{
    // double-click in scintilla
    if (event.detail == 2 && _gParseOutput) {
        // Get the line double-clicked on.
        var x = event.layerX < 0 ? 0 : event.layerX;
        var y = event.layerY < 0 ? 0 : event.layerY;
        var scimoz = document.getElementById("runoutput-scintilla").ISciMoz;
        var lineNum = scimoz.lineFromPosition(scimoz.positionFromPoint(x,y));
        var lineObj = new Object();
        scimoz.getCurLine(lineObj);
        var line = lineObj.value;
        
        // Get the parsed result.
        var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                           .getService(Components.interfaces.koILastErrorService);
        var fileObj = new Object();
        var lineNumObj = new Object();
        var columnNumObj = new Object();
        try {
            _gTerminal.parseLine(line, fileObj, lineNumObj, columnNumObj);
        } catch (ex) {
            var errmsg = lastErrorSvc.getLastErrorMessage();
            alert(errmsg);
            return;
        }
        
        _RunOutput_JumpToResult(fileObj.value, lineNumObj.value,
                                columnNumObj.value);
        //XXX Does this cancel work?
        event.cancelBubble = true;
        event.preventDefault();
    }
}


function RunOutput_OutlinerOnClick(event)
{
    // c.f. mozilla/mailnews/base/resources/content/threadPane.js
    var t = event.originalTarget;

    // single-click on a column
    if (t.localName == "outlinercol") {
        _gTerminal.sort(t.id);
        _RunOutput_UpdateSortIndicators(window, t.id);
    }

    // double-click in the outliner body
    else if (event.detail == 2 && t.localName == "outlinerbody") {
        var outlinerWidget = document.getElementById("runoutput-outliner");
        var i = outlinerWidget.currentIndex;
        _RunOutput_JumpToResult(_gTerminal.getFile(i),
                                _gTerminal.getLine(i),
                                _gTerminal.getColumn(i));

        // If the invoked result is the last visible one and there are more,
        // then scroll the list of find results by one.
        // XXX Could also scroll up if double-clicking on top visible item.
        var box = outlinerWidget.boxObject;
        if (box.getLastVisibleRow() <= i && _gTerminal.rowCount > i+1) {
            box.scrollByLines(1);
        }
    }
}


function _RunOutput_JumpToResult(file, line, column)
{
    // Jump to the parsed command output result.
    //XXX Not sure if URLs will pass through doOpenPath() as I might hope.
    //    This would be fairly rare, though.
    if (file) {
        var rv = doOpenPath(file);
    }
    var view = window.getFocusedView();
    if (!view) return;
    var scimoz = view._scimoz;

    // If have line and column information then select that region.
    // XXX Have not tested this with line and column ranges yet.
    if (line) {
        var start, end; // start and end positions to select
        
        var firstLine, lastLine;
        if (line.indexOf("-") == -1) {  // e.g., line == "4"
            firstLine = line;
            lastLine = line;
        } else { // e.g., line == "3-6"
            var lines = line.split("-");
            firstLine = lines[0];
            lastLine = lines[1];
        }
        
        if (column) {
            var firstCol, lastCol;
            if (column.indexOf("-") == -1) {  // e.g., column == "4"
                firstCol = column;
                lastCol = column;
            } else { // e.g., column == "3-6"
                var columns = column.split("-");
                firstCol = columns[0];
                lastCol = columns[1];
            }
            
            if (firstLine == lastLine && firstCol == lastCol) {
                start = scimoz.positionFromLine(firstLine-1) + (firstCol-1);
                end = start; // don't select if no range is given
            } else {
                start = scimoz.positionFromLine(firstLine-1) + (firstCol-1);
                end = scimoz.positionFromLine(lastLine-1) + (lastCol-1);
            }
            
        } else {
            start = scimoz.positionFromLine(firstLine - 1);
            end = scimoz.getLineEndPosition(lastLine - 1);
        }
        
        scimoz.selectionStart = start;
        scimoz.selectionEnd = end;
    }

    DoTheScimozWiggle(scimoz); //XXX Is this necessary, DavidA?

    // transfer focus to the editor
    view.manager.setFocus();
}


function RunOutput_OutlinerOnKeyPress(event)
{
    if (event.keyCode == 13) {
        // Open and/or switch to the appropriate file.
        var outlinerWidget = document.getElementById("runoutput-outliner");
        var i = outlinerWidget.currentIndex;
        var file = _gTerminal.getFile(i);
        var line = _gTerminal.getLine(i);
        var column = _gTerminal.getColumn(i);
        _RunOutput_JumpToResult(file, line, column);
        return false;
    }
    return true;
}


function _RunOutput_UpdateSortIndicators(mainWindow, sortId)
{
    var sortedColumn = null;

    // set the sort indicator on the column we are sorted by
    if (sortId) {
        sortedColumn = mainWindow.document.getElementById(sortId);
        if (sortedColumn) {
            var sortDirection = sortedColumn.getAttribute("sortDirection");
            if (sortDirection && sortDirection == "ascending") {
                sortedColumn.setAttribute("sortDirection", "descending");
            } else {
                sortedColumn.setAttribute("sortDirection", "ascending");
            }
        }
    }

    // remove the sort indicator from all the columns
    // except the one we are sorted by
    var currCol = mainWindow.document.getElementById("runoutput-outliner").firstChild;
    while (currCol) {
        while (currCol && currCol.localName != "outlinercol") {
            currCol = currCol.nextSibling;
        }
        if (currCol && (currCol != sortedColumn)) {
            currCol.removeAttribute("sortDirection");
        }
        if (currCol) {
            currCol = currCol.nextSibling;
        }
    }
}

