# Copyright (c) 2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# Scintilla event handler

import sys
import os
import time
import threading
import urllib
import urlparse
import md5
import codecs
import stat
import re

from xpcom import components, ServerException, nsError, COMException, _xpcom
from xpcom.file import URIFile
import xpcom.server

import koUnicodeEncoding
import eol
import koEditorController
from koLogger import logError, logWarning, logDebug
from koSciMozEvents import SciMozEvents
import convertURL



sci_constants = components.interfaces.ISciMoz

def RGB(r,g,b): return r+g*256+b*256*256

def MakePoint(pos):
    return pos >> 16, pos & 0xFF

def nonFuncChar(ch):
    return ch in ' \t\n\r\v*/%+-=('

# Attempt to determine the language of a file based on it's contents. If
# this code gets more sophisticated then it should probably become part
# of language services. But sending the entire file contents over XPCOM
# might be expensive and file load time is something that we want to
# minimize right now so we'll leave it here for now.

shebangPatterns = [
    ('Perl', re.compile(ur'^#!.*perl.*$', re.IGNORECASE)),
    ('PHP', re.compile(ur'^#!.*php.*$', re.IGNORECASE)),
    ('Python', re.compile(ur'^#!.*python.*$', re.IGNORECASE)),
    ('Ruby', re.compile(ur'^#!.*ruby.*$', re.IGNORECASE)),
    ('Tcl', re.compile(ur'^#!.*expect.*$', re.IGNORECASE)),    
    ('Tcl', re.compile(ur'^#!.*tcl.*$', re.IGNORECASE)),    
    ('Tcl', re.compile(ur'^#!.*wish.*$', re.IGNORECASE)),    
]

re_firstline = re.compile(ur'(.*?)(?:\r|\n|$)')

def guessLanguageFromContents(buffer):
    if buffer.startswith(u'<?xml'): # Found an XML declaration
        return ('XML', ['Text', 'XSLT'])
    elif buffer.startswith(u'#!'): # Found shebang line
        buffer = re_firstline.match(buffer).groups(0)[0]
        possibleLanguages = []
        for language, pattern in shebangPatterns:
            if pattern.match(buffer):
                possibleLanguages.append(language)                
        if len(possibleLanguages) == 0:
            return (None, None)            
        elif len(possibleLanguages) == 1:
            return (possibleLanguages[0], ['Text'])            
        else:
            return (None, possibleLanguages + ['Text'])
    else:
        return (None, None)

# this class uses a thread to kick off autosave.  More autosaves
# can be scheduled while the thread is working on one document
# the thread quits when no more documents need to be saved,
# and will be restarted on the next autosave request
class koAutoSaveScheduler:
    
    def __init__(self):
        self._editors = []
        self._thread = None
        self._tlock = threading.Lock()
        self._cv = threading.Condition()
    
    def add(self, editorWrapper):
        self._cv.acquire()
        self._editors.append(editorWrapper)
        self._cv.notify()
        self._cv.release()
        
        # see if we need to start a new thread
        self._tlock.acquire()
        if not self._thread:
            self._thread = threading.Thread(target = koAutoSaveScheduler._process,
                                         args = (self,))
            self._thread.start()
        self._tlock.release()
    
    def _process(self):
        
        while 1:
            self._cv.acquire()

            if len(self._editors) < 1:
                # If no results arrive soon, let the thread die
                self._cv.wait(1)
                if len(self._editors) < 1:
                    self._cv.release()
                    break
            editorWrapper = self._editors.pop(0)

            self._cv.release()

            apply(koEditorWrapper._autoSaveNow, (editorWrapper,))

        self._tlock.acquire()
        self._thread = None
        self._tlock.release()


class koEditorWrapper(SciMozEvents):
    """Each Scintilla buffer has associated with it one instance of this
    class.  This class isolates the rest of the world from having to
    know how Scintilla works.
    """
    
    _com_interfaces_ = [components.interfaces.ISciMozEvents,
                        components.interfaces.koIEditorWrapper,
                        components.interfaces.nsIObserver]
    _reg_desc_ = "Scintilla Editor Wrapper"
    _reg_contractid_ = "@activestate.com/koEditorWrapper;1"
    _reg_clsid_ = "{C9767BA9-21BE-488f-9374-4A8FA7B649B3}"
    
    _autoSaveScheduler = koAutoSaveScheduler()
    
    def __init__(self):
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"]\
                            .getService(components.interfaces.nsIObserverService)

        self._scintilla = None
        #self._braceCount = 0
        self._show_folds = 0
        self._last_class_browse = None
        self._controller = None
        self.extensionUI = None
        self.commandCompletionSvc = None
        self._triggers = None
        self._triggersCallTip = None
        self._eolText = eol.EOL_LF
        # diskfile dirty
        self._lastModificationDate_lastCheck = 0
        self._lastModificationDate = 0
        self._fileSize_lastCheck = 0
        self._fileSize = 0
        self._diskFileDirty = 0
        
        # *sigh* - these should not need to be here, but the whole init/load
        # process sucks.  Will be set correctly by the time we come to edit, but not
        # as we are loaded.
        self._doAutoComplete = 0
        self._doStartCallTip = 0

        self.isScintillaDirty = 0
        self.isEncodingDirty = 0
        self._writable = 1
        
        self.encodingServices = components.classes['@activestate.com/koEncodingServices;1'].\
                         getService(components.interfaces.koIEncodingServices);

        # Set the default encoding for the file
        
        self.encoding = components.classes['@activestate.com/koEncoding;1'].\
                                 createInstance(components.interfaces.koIEncoding)
        
        # Grab a reference to the global preference service.
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
                            getService(components.interfaces.koIPrefService).prefs

        # Create a preference set to hold view preferences for this editor and
        # hook it into the preference hierarchy.
        self.viewPrefs = components.classes['@activestate.com/koPreferenceSet;1'].\
                                 createInstance(components.interfaces.koIPreferenceSet);
        self.viewPrefs.parent = self._globalPrefs
        self._doLaterKP = None

        # Grab a reference to the style manager service, and get the style prefs set
        self._styleManager = components.classes["@activestate.com/koStyleManager;1"].\
                             getService(components.interfaces.koIStyleManager)
        self.stylePrefs = self._styleManager.getStylePrefs("editor")

        self._styleApplier = None
        self._lastchar = None        
        self._lexer = None
        self.languageOb = None
        self._language = None
        self._inBatchOperation = 0
        
    def init(self, scintilla, viewName, scratchDocument, language, viewstate):
        self.viewPrefs.id = viewName
        self.viewstate = viewstate
        self._scratchDocument = scratchDocument

        # Watch for preference set changes from these pref sets.
        self._globalPrefs.AddObserver(self)
        self.viewPrefs.AddObserver(self)
        #self.stylePrefs.AddObserver(self) # We don't directly listen to this - our style applier does.

        self._scintilla = scintilla
        self._autoSaveFileName = None
        self.url = None  # url==None means the document is untitled
        self.set_isEditable(1) # XXX time consumer in this function
 
        self.sourceFile = components \
                      .classes["@activestate.com/koSourceFile;1"] \
                      .createInstance(components.interfaces.koISourceFile)
        self.sourceFile.init(self._scintilla)

        self.encoding.python_encoding_name = self._globalPrefs.getStringPref('encodingDefault')
        if language: # Will be false for opened documents
            self._set_language(language)
            languagePref = self._globalPrefs.getPref('languages')\
                             .getPref('languages/' + language)

            lang_enc = languagePref.getStringPref(language+'/newEncoding')
            if lang_enc != 'Default Encoding':
                self.encoding.python_encoding_name = languagePref.getStringPref(language+'/newEncoding')
                self.encoding.byte_order_marker = languagePref.getBooleanPref(language+'/newBOM')
            
            
        # Scintilla Document Pointer
        self._docPointer = self._scintilla.createDocument()
        self._scintilla.docPointer = self._docPointer
        self._scintilla.codePage = 0
        if self.encoding.python_encoding_name == 'utf-8':
            self._scintilla.codePage = 65001 # Unicode
        elif self.encoding.python_encoding_name.startswith('cp'):
            # likely will only happen on windows
            try:
                self._scintilla.codePage = int(self.encoding.python_encoding_name[2:])
            except ValueError,e:
                #not an int
                self._scintilla.codePage = 0
        
        #jump to the beginning
        self._scintilla.selectionStart = 0
        self._scintilla.selectionEnd = 0

        # set scroll width
        self._scintilla.scrollWidth = self._globalPrefs.getLongPref("scrollWidth")

        # set end-of-line marker
        self._scintilla.eOLMode = self.getPrefEndOfLineAsScintillaConstant()

    def destroy(self):
        #print "koEditorWrapper::destroy()"
        
        # remove observers
        # XXX these cause exceptions regarding weakref/null pointer, but
        # NOT doing this we continue to leak editor wrappers.
        self.viewPrefs.RemoveObserver(self)
        self._globalPrefs.RemoveObserver(self)
        if self._styleApplier:
            self._styleApplier.destroy()
            
        #print "   killing docPointer %d" % self._docPointer
        self._scintilla.releaseDocument(self._docPointer)
        self._styleApplier = None
        self._controller = None

    def get_isDirty(self):        
        return self.isEncodingDirty or self.isScintillaDirty or self._diskFileDirty

    
    def __setattr__(self, name, value):
        self.__dict__[name] = value
        
        if name in ["isEncodingDirty", "isScintillaDirty"]:
            self._notifyDirtynessChange()
    
    def _notifyDirtynessChange(self):
        try:
            self._observerSvc.Notify(self, 'dirtyness', 'ignored')
        except COMException, e:
            # do nothing: Notify sometimes raises an exception if (???)
            # receivers are not registered?
            pass
        
    def get_controller(self):
        # xpcom readonly attribute - returns nsIController
        if self._controller is None:
            self._controller = koEditorController.GetControllerForLanguage(self)
        return self._controller

    def get_leaf_name(self,filename):
        return os.path.basename(filename)

    def get_isWritable(self):
        if self.url.startswith('koremote://'):
            # Get the writable status from koIFTPFile when we open the file?
            # Note: This can apparently raise an exception.
            return self._writable
        else:
            localPath = convertURL.LocalPathFromURL(self.url)
            if localPath and os.path.exists(localPath):
                return os.access(localPath, os.W_OK)
            else:
                return 1

    def get_existing_line_endings(self):
        endings, recommended = eol.detectEOLFormat(self._scintilla.text)
        if endings == eol.EOL_NOEOL:
            return self.get_new_line_endings()
        else:
            return endings

    def set_existing_line_endings(self, ending):
        self._scintilla.beginUndoAction()
        try:
            self._scintilla.text = eol.convertToEOLFormat(self._scintilla.text, ending)
        finally:
            self._scintilla.endUndoAction()

    def translateFromEolToScintilla(self, eolConstant):
        if eolConstant == eol.EOL_CR:
            return components.interfaces.ISciMoz.SC_EOL_CR
        elif eolConstant == eol.EOL_LF:
            return components.interfaces.ISciMoz.SC_EOL_LF
        elif eolConstant == eol.EOL_CRLF:
            return components.interfaces.ISciMoz.SC_EOL_CRLF
        else:
            assert 0, "unexpected eol constant"
            
    def translateFromScintillaToEol(self, scintillaConstant):
        if scintillaConstant == components.interfaces.ISciMoz.SC_EOL_CR:
            return eol.EOL_CR
        elif scintillaConstant == components.interfaces.ISciMoz.SC_EOL_LF:
            return eol.EOL_LF
        elif scintillaConstant == components.interfaces.ISciMoz.SC_EOL_CRLF:
            return eol.EOL_CRLF
        else:
            assert 0, "unexpected scintilla constant"
        
    def get_new_line_endings(self):
        return self.translateFromScintillaToEol(self._scintilla.eOLMode);

    def set_new_line_endings(self, ending):
        self._scintilla.eOLMode = self.translateFromEolToScintilla(ending)
        self._eolText = eol.eolMappings[ending]
            
    def get_isEncodable(self):
        try:
            self.encode()
            return 1
        except UnicodeError:
            return 0

    def get_isEditable(self):
        return self._isEditable

    def forceEncodingFromEncodingName(self, encoding_name):
        encoding = components.classes['@activestate.com/koEncoding;1'].\
                                 createInstance(components.interfaces.koIEncoding)
        encoding.python_encoding_name = encoding_name
        self._set_encoding(encoding, 'ignore')

    def setEncodingFromEncodingName(self, encoding_name):
        encoding = components.classes['@activestate.com/koEncoding;1'].\
                                 createInstance(components.interfaces.koIEncoding)
        encoding.python_encoding_name = encoding_name
        self._set_encoding(encoding)

    def set_encoding(self, encoding):
        self._set_encoding(encoding)
        
    def _set_encoding(self, encoding, mode='strict'):
        if self.encoding.python_encoding_name != encoding.python_encoding_name or\
           self.encoding.byte_order_marker != encoding.byte_order_marker:
            
            # Need to update the styleapplier here so we can set a new
            # encoding to the file.  This will replace fonts, etc.  as a
            # user changes the encoding.  We also have to handle
            # changing back and forth between utf-8 and single byte
            # encodings
            
            if not self.encoding.python_encoding_name.startswith('utf') \
               and encoding.python_encoding_name.startswith('utf') \
               and self._scintilla.codePage != 65001:
                try:
                    encodedText = koUnicodeEncoding.\
                              makeUTF8(self._scintilla.text,
                                    self.encoding.python_encoding_name)
                except Exception, e: # Figure out the real exception types
                    raise "Unable to convert text to UTF-8 from %s, %s" % \
                        (encoding.python_encoding_name,e)
                self._scintilla.codePage = 65001 # Unicode
                self._scintilla.text = encodedText
                self.encoding = encoding
                self.isEncodingDirty = 1
            elif self.encoding.python_encoding_name.startswith('utf') \
               and not encoding.python_encoding_name.startswith('utf'):
                old_encoding = self.encoding
                self.encoding = encoding
                try:
                    encodedText = self.encode(mode)[0]
                except Exception, e: # Figure out the real exception types
                    # try to encode to the new encoding directly
                    #encodedText = koUnicodeEncoding.\
                    #              tryEncoding(self._scintilla.text,
                    #                          encoding.python_encoding_name)
                    #if not encodedText:
                    self.encoding = old_encoding
                    raise "Unable to convert text to %s from UTF-8, %s" % \
                        (encoding.python_encoding_name,e)
                
                self._scintilla.codePage = 0
                if encoding.python_encoding_name.startswith('cp'):
                    # likely will only happen on windows
                    try:
                        self._scintilla.codePage = int(encoding.python_encoding_name[2:])
                    except ValueError,e:
                        #not an int
                        self._scintilla.codePage = 0
                        
                self._scintilla.text = encodedText
                self.isEncodingDirty = 1
            else:
                # 8bit encodings do not require text changes
                # and we do not need to dirty the buffer for it
                # the change will be saved in view prefs
                self.encoding = encoding
            
            self.encoding.byte_order_marker = encoding.byte_order_marker
            
            # reset fonts for this encoding
            self._styleApplier.setEncoding(self.encoding.python_encoding_name)
            self._styleApplier.applyStyles()
            
            try:
                self._observerSvc.Notify(self,
                                         'current_view_encoding_changed',
                                         self.encoding.python_encoding_name)
            except COMException, e:
                # do nothing: Notify sometimes raises an exception if (???)
                # receivers are not registered?
                pass
        
    def set_isEditable(self, state):
        self._isEditable = state
        self._scintilla.readOnly = not state
        # XXX this call is a time consumer, 120ms debug build
        self._scintilla.sendUpdateCommands("load")  # for paste command       

    def set_diskFileDirty(self, dirty):
        self._diskFileDirty = dirty

    def get_diskFileDirty(self):
        return self._diskFileDirty

    def beginBatchOperation(self):
        # Does *not* disable undo collection - do that yourself :)
        
        # When batch op is complete, we may need to send a command
        # update - but only when the selection changes _shape_, not
        # location (ie, only when it changes from no selection to
        # selection, or vice-versa.
        self._hadSelBeforeBatchOp = self._scintilla.selectionEnd != self._scintilla.selectionStart
        self._inBatchOperation = 1

    def endBatchOperation(self):
        self._inBatchOperation = 0
        sm = self._scintilla
        haveSelNow = sm.selectionEnd != sm.selectionStart
        if self._hadSelBeforeBatchOp != haveSelNow:
            sm.sendUpdateCommands("select")
        
    # the following two functions handle information for changes to the file on disk
    # this currently only works for local files, and does not work for MAC files!
    def isDiskFileDirty(self):
        # if no file loaded, return false
        if not self.url or not self.url.startswith('file://'): return 0
        
        info = self._statFile(self.url)
        if not info: return -1
        lastModificationDate = info[stat.ST_MTIME] #st_mtime
        fileSize = info[stat.ST_SIZE] # st_size
            
        # if we already checked, and did not reload, return false
        if lastModificationDate == self._lastModificationDate_lastCheck and \
           fileSize == self._fileSize_lastCheck: return 0
        self._lastModificationDate_lastCheck = lastModificationDate
        self._fileSize_lastCheck = fileSize
        
        if lastModificationDate != self._lastModificationDate or \
            fileSize != self._fileSize: return 1
        
        return 0

    def _guessLanguage(self, url, buffer = None, default = 'Text'):
        basename = os.path.basename(urlparse.urlparse(url)[2])
        registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
                                     getService(components.interfaces.koILanguageRegistryService)

        fileNameLanguage = registryService.suggestLanguageForFile(basename)        
        if (self._globalPrefs.getBooleanPref('langCharacteristics') or (fileNameLanguage == '')) and (buffer is not None):
            contentLanguage, contentOtherLanguages = guessLanguageFromContents(buffer)
        else:
            contentLanguage = None

        if contentLanguage is None:
            language = fileNameLanguage or default
        else:
            if fileNameLanguage in contentOtherLanguages:
                language = fileNameLanguage
            else:
                language = contentLanguage

        return language

    def _statFile(self, url):
        """Return stat info on the given url if it is an existing local file.
        Otherwise return None.
        """
        path = convertURL.LocalPathFromURL(url)
        if path and os.path.exists(path):
            info = os.stat(path)
            return info
        return None
        
    def _setFileInfo(self, url):
        # store the file info
        info = self._statFile(url)
        if info:
            self._lastModificationDate = info[stat.ST_MTIME]
            self._fileSize = info[stat.ST_SIZE]
    
    def _getAutoSaveFileName(self, url):
        koDirs = components.classes["@activestate.com/koDirs;1"].\
                 getService(components.interfaces.koIDirs)
        dname = os.path.join(koDirs.userDataDir, "autosave")
        dname = str(dname) # Unicode filenames can apparently break things.
        if not os.path.exists(dname):
            os.mkdir(dname)
        fname = os.path.join(dname, md5.new(url).hexdigest())
        fname = str(fname) # Unicode filenames can apparently break things.
        return fname

    def getPrefEndOfLineAsScintillaConstant(self):
        prefString = self._globalPrefs.getStringPref('endOfLine')

        if prefString == 'CRLF':
            return components.interfaces.ISciMoz.SC_EOL_CRLF
        elif prefString == 'LF':
            return components.interfaces.ISciMoz.SC_EOL_LF
        elif prefString == 'CR':
            return components.interfaces.ISciMoz.SC_EOL_CR
        else:
            assert 0, 'unknown "endOfLine" preference'

    def getPrefEndOfLineAsEolContant(self):
        return self.translateFromScintillaToEol(self.getPrefEndOfLineAsScintillaConstant())

    def templateReplacement(self, buffer):
        # XXX - This code should probably be relocated
        year, month, day, hour, minute, second, weekday, julian_day, dst_flag = time.localtime(time.time())
        variableSubstituions = { 'local_year' : str(year),
                               'local_month' : str(month),
                               'local_day' : str(day),
                               'local_hour' : str(hour),
                               'local_minute' : str(minute),
                               'local_second' : str(second),
                               'local_weekday' : str(weekday),
                               'local_julian_day' : str(julian_day),
                                   }

        # XXX Fix this when nested scopes are available
        def replace(match, variableSubstituions = variableSubstituions):
            def genGUID():
                if sys.platform.startswith("win"):
                    import pythoncom
                    return str(pythoncom.CreateGuid())[1:-1] # Strip of the {}'s, which are not always used
                else:
                    return '<insert GUID here>'

            variable = match.group(1).lower()

            if variableSubstituions.has_key(variable):
                return variableSubstituions[variable]
            else:
                if variable == 'guid':
                    return genGUID()
                elif variable.startswith('guid'):
                    variableSubstituions[variable] = genGUID()
                    return variableSubstituions[variable]
                else:
                    return '[komodo-variable "%s" not supported]' % (variable)
        
        return re.sub(r'\[komodo-variable:\s*\$(.*?)\s*\]', replace, buffer)
        
    def _load(self, url, loadAutoSavedFile, eolTranslation=None,
              doTemplateReplacement=0):
        """Load the given URL into Komodo editor.
        
        "url" is the url to load.
        "loadAutoSavedFile" is a boolean indicating if the content
            should be loaded from the URL's autosave file.
        "eolTranslation" (optional) if None indicates that no EOL
            translation should be done on the URL content. Otherwise it
            indicates the EOL to translate the content to. Valid EOL
            values the EOL_* flags from eol.py.
        "doTemplateReplacement" (optional) indicates whether this ...
        
        Returns a tuple (errno, errmsg) where "errno" is a valid nsError
        error number.
        """
        #print "XXX koEditorWrapper._load(url=%r, loadAutoSavedFile=%r, "\
        #      "eolTranslation=%r, doTemplateReplacement=%r)"\
        #      % (url, loadAutoSavedFile, eolTranslation,
        #         doTemplateReplacement)

        #XXX stringify for urlparse, apparently it needs it
        self.url = str(url)
        

        #---- Determine from where to load the file content.
        # Set the autosave file name and determine whether we are
        # reloading an autosaved file.
        self._autoSaveFileName = self._getAutoSaveFileName(self.url)
        if loadAutoSavedFile:
            urlToLoad = "file://"+self._autoSaveFileName
            self._diskFileDirty = 1
        else:
            urlToLoad = self.url
            self._diskFileDirty = 0


        #---- Load the file content.
        self._writable = 1
        try:
            # Special handling for koremote URLs.
            scheme = urlparse.urlparse(urlToLoad)[0]
            if scheme == "koremote":
                # Parse out the meaningful data from a koremote:// URL.
                #    koremote://<server-alias>/<the-real-url>
                url_parts = re.match(r'^(.*?)://(.*?)/(.*)$',
                                     urlToLoad).groups()
                server_alias = url_parts[1]
                urlToLoad = url_parts[2]
                scheme = urlparse.urlparse(urlToLoad)[0]
            else:
                server_alias = None
                
            # Load the file content (into 'data' variable).
            #print "XXX koEditorWrapper: the real URL to load: '%s'" % urlToLoad
            if scheme == "ftp":
                infile = components.classes["@activestate.com/koFTPFile;1"]\
                         .createInstance(components.interfaces.koIFTPFile)
                try:
                    infile.initServer(urlToLoad, "r", server_alias)
                    self._writable = infile.isWritable
                    if not infile.isReadable:
                        # XXX not the best way to handle file not found, but works.
                        return nsError.NS_ERROR_FILE_NOT_FOUND,\
                               "Permissions disallow reading file"
                    data = infile.read(-1) #-1 == entire file
                except COMException, e:
                    err = infile.ftpfile.connection.lastError
                    if not err: raise
                    return e.errno, err
            elif scheme == "file":
                localPath = convertURL.LocalPathFromURL(urlToLoad)
                infile = open(localPath, 'rb')
                data = infile.read()
            elif scheme not in ["http","ftp"]:
                infile = URIFile(urlToLoad)
                data = str(infile.read())
            else:
                infile = urllib.urlopen(urlToLoad)
                data = infile.read()
        except IOError, e:
            return e.errno, e.strerror
        except COMException, e:
            #XXX Perhaps we should fix URIFile() to convert from
            #XXX COMExceptions to IOErrors for us.
            if e.errno == nsError.NS_ERROR_FILE_NOT_FOUND:
                return e.errno, "File not found."
            else:
                logError("Unknown error %d." % (e.errno))
                return e.errno, "Unknown error."
        # Invariants at this point:
        # - 'data' variable holds the content to load

        self._setFileInfo(self.url)


        #---- Determine the encoding of the file content.
        # XXX Our encoding detection does not detect between latin
        #     encodings.
        buffer = None
        bom = 0
        encoding_name = self._globalPrefs.getStringPref('encodingDefault')
        encoding = self.encodingServices.get_encoding_info(encoding_name)\
                       .python_encoding_name

        if self.viewstate and self.viewstate.hasPref('encoding'):
            tryencoding = self.viewstate.getStringPref('encoding')
        else:
            tryencoding = None
        tryxmldecl = self._globalPrefs.getBooleanPref('encodingXMLDec')
        trymeta = self._globalPrefs.getBooleanPref('encodingHTML')
        autodetect = self._globalPrefs.getBooleanPref('encodingAutoDetect')
        if autodetect or tryencoding is not None:
            buffer, encoding, bom =\
                koUnicodeEncoding.autoDetectEncoding(data, tryxmldecl,
                                                     trymeta, tryencoding,
                                                     encoding)
            bom = len(bom) > 0
        elif encoding == 'utf-8':
            try:
                buffer = unicode(data, encoding)
                if (len(buffer) > 0) and (buffer[0] == unichr(0xfeff)):
                    bom = 1
                    buffer = buffer[1:]
                else:
                    bom = 0
            except UnicodeError, e:
                return -1,\
                    'The file cannot be opened using the encoding "%s". '\
                    'You can change the encoding used to open files in the '\
                    '"File Settings" tab of the "Preferences" dialog.'\
                    % self.encodingServices.get_encoding_info(encoding)\
                          .friendly_encoding_name
        else:
            buffer = data
            
        if buffer is None:
            encName = self.encodingServices.get_encoding_info(encoding)\
                          .friendly_encoding_name
            return -1, 'The file encoding "%s" is not supported.' % encName

        # Make sure that there are no null bytes in the buffer
        if buffer.find(unichr(0)) != -1:
            return -1, 'Komodo cannot open files containing null characters.'
        # Invariants at this point:
        # - the file content is in 'buffer'
        # - 'encoding' and 'bom' are determined

        #XXX Dev Notes:
        #   After this you are on your own. This function is a little
        #   bit insane after this. It switches between 'data' and
        #   'buffer' as the variables holding the content to use. Some
        #   transformations work on one, some on the other.


        if doTemplateReplacement:
            data = self.templateReplacement(data)
            

        # Convert to UTF-8 for scintilla
        codepage = 0
        if encoding.startswith('utf'):
            #XXX Is this necessary, test without, may only be necessary
            #    for utf-16 files
            data = codecs.utf_8_encode(buffer)[0]
            codepage = 65001 # Unicode
        elif encoding.startswith('cp'):
            # likely will only happen on windows
            try:
                codepage = int(encoding[2:])
            except ValueError, ex: # not an int
                codepage = 0
            

        if self.viewstate:
            self.viewstate.setStringPref('encoding', encoding)
        
        self.encoding.python_encoding_name =\
            self.encodingServices.get_canonical_python_encoding_name(encoding)
        self.encoding.byte_order_marker = bom
        try:
            self._observerSvc.Notify(None, 'current_view_encoding_changed',
                                     self.encoding.python_encoding_name)
        except COMException, e:
            pass    # ignore notification error

        # Try to figure out which language we are dealing with
        language = self._guessLanguage(self.url, buffer)
        self._set_language(language)

        # Turn off undo collection as we set the text.
        scin = self._scintilla
        scin.undoCollection = 0
        scin.emptyUndoBuffer()
        scin.readOnly = 0

        if eolTranslation is not None:
            data = eol.convertToEOLFormat(data, eolTranslation)
            scin.eOLMode = self.translateFromEolToScintilla(eolTranslation)
        else:
            format, suggestedFormat = eol.detectEOLFormat(data)
            if suggestedFormat == -1:
                scin.eOLMode = self.getPrefEndOfLineAsScintillaConstant()
            else:
                scin.eOLMode = self.translateFromEolToScintilla(suggestedFormat)
        
        scin.text = data
        scin.codePage = codepage
        scin.undoCollection = 1
        scin.setSavePoint()

        # We have to specifically scan the buffer here, since scintilla
        # text is not set when we set the language!
        if self.commandCompletionSvc is not None:
            self.commandCompletionSvc.scanBufferCompletion(scin, self.url)
            
        return 0, '' # return success and no error message
        
    def load(self, url):
        return self._load(url, 0)

    def loadFileTemplate(self, template):
        loadResult = self._load(template.fileURI, 0, self.getPrefEndOfLineAsEolContant(), 1)
        # Reset it to an untitled document
        self.url = None
        return loadResult

    def activate(self):
        "Configures Scintilla for use with this buffer."
        self._lastchar = None
        self._scintilla.docPointer = self._docPointer
        self._setupScintillaForLanguage()
 
    def deactivate(self):
        "Unconfigures Scintilla for us with this buffer."
        # cancel autocompletion/tips
        if self._scintilla.autoCActive():
            #print "canceling autocomplete"
            self._scintilla.autoCCancel()
        if self._scintilla.callTipActive():
            self._scintilla.callTipCancel()
        self._scintilla.docPointer = self._scratchDocument

    def encode(self, mode='strict'):
        if self._scintilla.codePage != 65001:
            return (self._scintilla.text,0)
        encoder = codecs.lookup(self.encoding.python_encoding_name)[0]
        
        decodedText, length = codecs.utf_8_decode(self._scintilla.text)
        if self.encoding.byte_order_marker:
            decodedText = unichr(0xfeff) + decodedText
            
        return encoder(decodedText, mode)

    def removeUnencodeable(self):
        self._scintilla.beginUndoAction()
        try:
            #XXX This is broken, where is 'encoding_name' coming from?!
            encoder = codecs.lookup(encoding_name)[0]
            
            decodedText, length = codecs.utf_8_decode(self._scintilla.text)
            if self.encoding.byte_order_marker:
                try:
                    encoder(unichr(0xfeff))
                except UnicodeError:
                    self.encoding.byte_order_marker = 0
                    logDebug("A byte order marker was enabled for an encoding that doesn't support it. This should be impossible")
            
            self._scintilla.text = codecs.utf_8_encode(encoder(decodedText, 'ignore')[0])[0]
        finally:
            self._scintilla.endUndoAction()
    
    def removeAutoSaveFile(self):
        if not self._autoSaveFileName or\
           not os.path.exists(self._autoSaveFileName):
            return
        os.remove(self._autoSaveFileName)
        
    def haveAutoSave(self, url):
        url = str(url) # this converts from Unicode to string
                       # which is apparently needed by the urlparse module 
        
        # set the autosave file name
        self._autoSaveFileName = self._getAutoSaveFileName(url)

        fileinfo = self._statFile(url)
        if not fileinfo: return 0
        tmpfileinfo = self._statFile("file://"+self._autoSaveFileName)
        if not tmpfileinfo: return 0
        
        if tmpfileinfo[stat.ST_MTIME] > fileinfo[stat.ST_MTIME]:
            #print "tmpfile is newer than original"
            return 1;
        else:
            # we have an autosave file, but it's old, delete it
            #print "removing old temporary auto save file ",self._autoSaveFileName
            os.remove(self._autoSaveFileName)
        return 0
        
    def doAutoSave(self, scin_work):
        self.scin_work = scin_work
        self._autoSaveScheduler.add(self)
        return 1
    
    def get_encodedText(self):
        return self.encode()[0]
    
    def _autoSaveNow(self):
        if self._scintilla == self.scin_work:
            # we have to wrap the scintilla to get the text since this is
            # the active document
            wrapSelf = xpcom.server.WrapObject(self, components.interfaces.koIEditorWrapper)
            editor = _xpcom.getProxyForObject(1, components.interfaces.koIEditorWrapper,
                                              wrapSelf, _xpcom.PROXY_SYNC | _xpcom.PROXY_ALWAYS)
            encodedText = editor.encodedText

        else:
            # this entire operation can happen safely in the thread
            # some magic to work around one-scintilla-per-tab
            scratch = self.scin_work.docPointer
            self.scin_work.docPointer = self._docPointer

            # change our scintilla to the work scintilla
            scin_tmp = self._scintilla
            self._scintilla = self.scin_work
            
            # get the text we need to save
            encodedText = self.encode()[0]

            # reset scintilla
            self.scin_work.docPointer = scratch
            self._scintilla = scin_tmp
        
        try:
            f = open(self._autoSaveFileName, 'wb')
        except IOError, e:
            #return e.errno, e.strerror
            return 0
        f.write(encodedText)
        f.close()
        return 1
    
    def restoreAutoSave(self, url):
        return self._load(url, 1)
        
    def _getMarkerData(self):
        s = self._scintilla
        _markers = []
        lineNo = -1
        marker_mask = -1 # really a mask of 32 1's.
        while 1:
            lineNo = s.markerNext(lineNo+1, marker_mask)
            if lineNo < 0:
                break
            _markers.append((lineNo, s.markerGet(lineNo)))
        return _markers

    def _setMarkerData(self, data):
        s = self._scintilla
        for (line, marker) in data:
            if not marker: continue # optimize for common case
            for i in range(32):  # there are 32 possible spots in scintilla for markers
                if marker & 1 << i:
                    s.markerAdd(line, i)

    def _getFoldPoints(self):
        _foldpoints = []
        i = 0
        s = self._scintilla
        _haveThem = {}
        for i in range(s.lineCount):
            if not s.getLineVisible(i):
                parent = s.getFoldParent(i)
                if not s.getFoldExpanded(parent) and \
                   not parent in _haveThem:
                    _foldpoints.append(parent)
                    _haveThem[parent] = 1
        return _foldpoints
    
    def _setFoldPoints(self, foldPoints):
        if not foldPoints: return
        s = self._scintilla
        s.colourise(0, -1)
        for line in foldPoints:
            s.toggleFold(line)

    def _clean(self, ensureFinalEOL, cleanLineEnds):
        """ This function cleans the text of scintilla according to the three prefs,
        any combination of which is valid.  It is careful to preserve the selection
        markers (currentPos and anchor)."""
        # in the case of any failure at all, we don't change the text
        try:
            _markers = self._getMarkerData()
            _foldpoints = self._getFoldPoints()
            _fvl = self._scintilla.firstVisibleLine
            _xOffset = self._scintilla.xOffset
            ANCHOR = '<<Komodo:Anchor>>'
            CURRENTPOS = '<<Komodo:CurrentPos>>'
            text = self._scintilla.text
            eolMode = self._scintilla.eOLMode
            while self._scintilla.getWCharAt(self._scintilla.currentPos-1) in ' \t':
                self._scintilla.currentPos -= 1
            while self._scintilla.getWCharAt(self._scintilla.anchor-1) in ' \t':
                self._scintilla.anchor -= 1
            currentPos = self._scintilla.currentPos
            anchor = self._scintilla.anchor
            if currentPos < anchor:
                text = text[:anchor] + ANCHOR + text[anchor:]
                text = text[:currentPos] + CURRENTPOS + text[currentPos:]
            else:
                text = text[:currentPos] + CURRENTPOS + text[currentPos:]
                text = text[:anchor] + ANCHOR + text[anchor:]
            eol = {components.interfaces.ISciMoz.SC_EOL_CR: '\r',
                   components.interfaces.ISciMoz.SC_EOL_CRLF: '\r\n',
                   components.interfaces.ISciMoz.SC_EOL_LF: '\n'}[eolMode]
            if cleanLineEnds:
                pattern = "(.*?)([ \t]+?)?(\r\n|\n|\r)"
                regex = re.compile(pattern, re.MULTILINE)
                text = regex.sub(r"\1"+eol, text+eol)[:-len(eol)]
            if ensureFinalEOL:
                if not text.endswith(eol):
                    text += eol
            if currentPos < anchor:
                # do currentPos replacement first, to not screw up position shifts do to the
                # replacement of the markers.
                newCurrentPos = text.index(CURRENTPOS)
                text = text[:newCurrentPos] + text[newCurrentPos + len(CURRENTPOS):]
                newAnchor = text.index(ANCHOR)
                text = text[:newAnchor] + text[newAnchor + len(ANCHOR):]
            else:
                newAnchor = text.index(ANCHOR)
                text = text[:newAnchor] + text[newAnchor + len(ANCHOR):]
                newCurrentPos = text.index(CURRENTPOS)
                text = text[:newCurrentPos] + text[newCurrentPos + len(CURRENTPOS):]
            self._scintilla.targetStart = 0
            self._scintilla.targetEnd = self._scintilla.length
            self._scintilla.replaceTarget(len(text), text)
            self._setMarkerData(_markers)
            self._setFoldPoints(_foldpoints)
            self._scintilla.currentPos = newCurrentPos
            self._scintilla.anchor = newAnchor
            fvl = self._scintilla.firstVisibleLine
            self._scintilla.lineScroll(0, _fvl-fvl)
            self._scintilla.xOffset = _xOffset
        except Exception, e:
            logError(e)
            return

    def save(self, url):
        ensureFinalEOL = self._globalPrefs.getBooleanPref("ensureFinalEOL")
        cleanLineEnds = self._globalPrefs.getBooleanPref("cleanLineEnds")

        if ensureFinalEOL or cleanLineEnds:
            self._clean(ensureFinalEOL, cleanLineEnds)

        # Encode in strict mode. It should be impossible for the encode to fail at this point.
        encodedText = self.encode()[0]
        #print "attempting to save: ",url
        server_alias = None
        remote_url = None
        if (url.startswith('koremote://')):
            # this is a remote file, need to parse the url
            remote_url = url
            url_parts = re.match(r'^(.*?)://(.*?)/(.*)$', url).groups()
            server_alias = url_parts[1]
            url = url_parts[2]
            #print "attempting to save: ",url

        if (url.startswith('ftp://')):
            outfile = components.classes["@activestate.com/koFTPFile;1"]\
                  .createInstance(components.interfaces.koIFTPFile)
            try:
                outfile.initServer(url, "w", server_alias)
                outfile.write(encodedText)
                outfile.close()
            except:
                err = outfile.ftpfile.connection.lastError
                if not err: raise
                return -1, err
        else:
            path = convertURL.LocalPathFromURL(url)
            #print "attempting to save: ",path
            #logDebug("save file: url('%s') ==> path('%s')\n" % (url,path));
            try:
                f = open(path, 'wb')
                f.write(encodedText)
                f.close()
            except IOError, e:
                return e.errno, e.strerror

        # note, if we got this far, the file is writable!
        self._writable = 1
        
        # restore the original koremote url so we can save again correctly
        if remote_url: url = remote_url
            
        self._setFileInfo(url)
        self._diskFileDirty = 0

        self.url = url
        # _dont_ empty undo buffer at save time - no reason you can't undo from a save-point!
        self.isScintillaDirty = 0
        self.isEncodingDirty = 0
        self._scintilla.setSavePoint()
        self.removeAutoSaveFile()
        self._autoSaveFileName = self._getAutoSaveFileName(url)
        # language may have changed
        language = self._guessLanguage(self.url, default=self._language)
        if self._language != language:
            self.set_language(language)        
        elif self.commandCompletionSvc is not None:
            self.commandCompletionSvc.scanBufferCompletion(self._scintilla, self.url)
        return 0, '' # return success and no error message

    def _enableFolding(self, whichMargin, foldstyle):
        scin = self._scintilla
        scin.setMarginTypeN(whichMargin, scin.SC_MARGIN_SYMBOL)
        scin.setMarginSensitiveN(whichMargin, 1)
        if foldstyle == 'arrows':
            #Arrow pointing right for contracted folders, arrow pointing down for expanded
            back = RGB(0x50, 0x50, 0x50)
            fore = RGB(0xaa, 0xaa, 0xaa)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPEN, scin.SC_MARK_ARROWDOWN)
            scin.markerDefine(scin.SC_MARKNUM_FOLDER, scin.SC_MARK_ARROW)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERSUB, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERTAIL, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEREND, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPENMID, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERMIDTAIL, scin.SC_MARK_EMPTY)
        elif foldstyle == 'plusminus':
            # Plus for contracted folders, minus for expanded
            fore = RGB(0xff, 0xff, 0xff)
            back = RGB(0x0, 0x0, 0x0)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPEN, scin.SC_MARK_MINUS)
            scin.markerDefine(scin.SC_MARKNUM_FOLDER, scin.SC_MARK_PLUS)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERSUB, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERTAIL, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEREND, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPENMID, scin.SC_MARK_EMPTY)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERMIDTAIL, scin.SC_MARK_EMPTY)
        elif foldstyle == 'curvytree':
            # Like a flattened tree control using circular headers and curved joins
            fore = RGB(0xff, 0xff, 0xff)
            back = RGB(0x40, 0x40, 0x40)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPEN, scin.SC_MARK_CIRCLEMINUS)
            scin.markerDefine(scin.SC_MARKNUM_FOLDER, scin.SC_MARK_CIRCLEPLUS)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERSUB, scin.SC_MARK_VLINE)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERTAIL, scin.SC_MARK_LCORNERCURVE)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEREND, scin.SC_MARK_CIRCLEPLUSCONNECTED)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPENMID, scin.SC_MARK_CIRCLEMINUSCONNECTED)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERMIDTAIL, scin.SC_MARK_TCORNERCURVE)
        elif foldstyle == 'squaretree':
            # Like a flattened tree control using square headers
            fore = RGB(0xff, 0xff, 0xff)
            back = RGB(0x80, 0x80, 0x80)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPEN, scin.SC_MARK_BOXMINUS)
            scin.markerDefine(scin.SC_MARKNUM_FOLDER, scin.SC_MARK_BOXPLUS)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERSUB, scin.SC_MARK_VLINE)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERTAIL, scin.SC_MARK_LCORNER)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEREND, scin.SC_MARK_BOXPLUSCONNECTED)
            scin.markerDefine(scin.SC_MARKNUM_FOLDEROPENMID, scin.SC_MARK_BOXMINUSCONNECTED)
            scin.markerDefine(scin.SC_MARKNUM_FOLDERMIDTAIL, scin.SC_MARK_TCORNER)
        scin.markerSetFore(scin.SC_MARKNUM_FOLDERMIDTAIL, fore)
        scin.markerSetBack(scin.SC_MARKNUM_FOLDERMIDTAIL, back)
        scin.markerSetFore(scin.SC_MARKNUM_FOLDEROPEN, fore)
        scin.markerSetBack(scin.SC_MARKNUM_FOLDEROPEN, back)
        scin.markerSetFore(scin.SC_MARKNUM_FOLDER, fore)
        scin.markerSetBack(scin.SC_MARKNUM_FOLDER, back)
        scin.markerSetFore(scin.SC_MARKNUM_FOLDERSUB, fore)
        scin.markerSetBack(scin.SC_MARKNUM_FOLDERSUB, back)
        scin.markerSetFore(scin.SC_MARKNUM_FOLDERTAIL, fore)
        scin.markerSetBack(scin.SC_MARKNUM_FOLDERTAIL, back)
        scin.markerSetFore(scin.SC_MARKNUM_FOLDEREND, fore)
        scin.markerSetBack(scin.SC_MARKNUM_FOLDEREND, back)
        scin.markerSetFore(scin.SC_MARKNUM_FOLDEROPENMID, fore)
        scin.markerSetBack(scin.SC_MARKNUM_FOLDEROPENMID, back)
        scin.setMarginMaskN(whichMargin, scin.SC_MASK_FOLDERS)
        scin.setProperty("fold", "1")
        scin.setFoldFlags(16)

    def doingFolding(self):
        return self._show_folds

    def _set_language(self, language):
        self.sourceFile.suggestLanguage(language)
        self._controller = None # So we get a new controller for the new language
        self._onLanguageChange()
        self.notify_language_change(language)

    def set_language(self, language):
        self._set_language(language)
        self._setupScintillaForLanguage()
        scin = self._scintilla
        scin.clearDocumentStyle()
        scin.colourise(0, scin.positionFromLine(scin.firstVisibleLine+scin.linesOnScreen))

    def notify_language_change(self, language):        
        try:
            self._observerSvc.Notify(self, "language_changed_"+self.viewPrefs.id, language);
        except COMException, e:
            # do nothing: Notify fails if a receiver fails!
            pass
        
    def get_language(self):
        return self._language

    def _setupScintillaForLanguage(self):
        assert self._styleApplier is not None and self._lexer is not None
        scin = self._scintilla
        self._styleApplier.initWithEncoding(self._scintilla, self._language, self.viewPrefs, self.encoding.python_encoding_name)

        lexer = self._lexer
        lexer.setCurrent(scin)
        
        styleBits = self.languageOb.styleBits
        scin.styleBits = styleBits
        # Determine how many indicators we can have.
        indicBits = self.languageOb.indicatorBits

        #XXX Scintilla renames the indicators depending on the number of style bits :-(.
        #XXX Compensate.
        indicOffset = 8 - styleBits - indicBits
        if indicBits == 2:
            scin.indicSetStyle(indicOffset + 1, scin.INDIC_SQUIGGLE);
            scin.indicSetFore(indicOffset + 1, RGB(0xff,0,0));
            scin.indicSetStyle(indicOffset, scin.INDIC_SQUIGGLE);
            scin.indicSetFore(indicOffset, RGB(0,0x80,0));
        else:
            scin.indicSetStyle(indicOffset, scin.INDIC_SQUIGGLE);
            scin.indicSetFore(indicOffset, RGB(0xff,0,0));
            
        self._show_folds = lexer.supportsFolding

        # We need to apply the prefs here for a number of reasons:
        # If activating, because if we guessed file
        # indentation, we may need to adjust indentation settings (and thus
        # items like visual indent markers).
        # On language change, so the font prefs update.
        self._applyViewPrefs()

    # call this after any operation that might change the language
    def _onLanguageChange(self):
        self.languageOb = self.sourceFile.getLanguageBlockAtPosition(0).language
        self._language = self.languageOb.name
        # reset the controller for the new language
        # the new language should have been set before we reach this point.
        # XXX this is VERY BAD.  We have to resinstall controllers any
        # time we change a language, it needs to be fixed to work
        # correctly.
        self._controller = None
        self._show_folds = 0
        scin = self._scintilla

        if self._styleApplier:
            self._styleApplier.destroy()
            self._styleApplier = None
            
        self._styleApplier = components.classes["@activestate.com/koScintillaStyleApplier;1"].\
                             createInstance()

        self._lexer = self.languageOb.getLanguageService(components.interfaces.koILexerLanguageService)

        self.commandCompletionSvc = self.languageOb.getLanguageService(components.interfaces.koICompletionLanguageService)

        if self.commandCompletionSvc is not None:
            self._triggers = self.commandCompletionSvc.triggers
            self._triggersCallTip = self.commandCompletionSvc.triggersCallTip
            scin.autoCSeparator = self.commandCompletionSvc.completionSeparator
            self.commandCompletionSvc.scanBufferCompletion(scin, self.url)

    def set_folding_border_visibility(self, yesorno):
        pass
  
    def doCommandCompletion(self, char):
        if self.commandCompletionSvc is None: return
        if self._doAutoComplete and (not self._triggers or char in self._triggers):
            self.commandCompletionSvc.AutoComplete(char, self._scintilla)
        # MUST have a trigger to do calltips!
        if self._doStartCallTip and self._triggersCallTip and char in self._triggersCallTip:
            self.commandCompletionSvc.StartCallTip(char, self._scintilla)

    def onSavePointReached(self):
        self.isScintillaDirty = 0
    
    def onSavePointLeft(self):
        self.isScintillaDirty = 1
    
    def onCharAdded(self, char):
        self._lastchar = char
        

    def onDoubleClick(self):
        # get the selection, and see if a variable is selected,
        # if so, make sure the entire variable is selected
        sm = self._scintilla
        haveSelNow = sm.selectionEnd != sm.selectionStart
        if sm.selectionStart > 0 and haveSelNow:
            indicators = None
            if self.languageOb: indicators = self.languageOb.variableIndicators
            if indicators:
                char = sm.getWCharAt(sm.selectionStart-1)
                if char in indicators:
                    sm.selectionStart = sm.selectionStart - 1

    def onUpdateUI(self):
        if not self.languageOb:
            # We can get an onUpdateUI call too early, before there is a language,
            # let alone a controller
            return 

        controller = self.get_controller()

        # hmm, we might be able to now catch tabs, etc. right here
        if self._lastchar and self._lastchar is not "":
            try:
                # XXX - this is evil - an earlier function sets self._lastchar to an integer, but
                # here we transform it to a character.
                # If this code ever fails, then *evil* things happen:
                # * self._lastchar is converted to a char.
                # * this function fails, preventing self._lastchar from being reset to None
                # * this function gets called again, this time failing to convert self._lastchar
                #  ( which is a char from last time around) back into a char
                # Thus, this function *again* fails, causing the whole process to be repeated.
                #
                # The correct fix is probably to set a local variable, then set self._lastchar to None
                # at the top of this function.  This function should only ever use the local variable.
                # This will ensure that self._lastchar is only ever an integer or None.
                #
                # (This was not done at the time of adding this comment as we are frozen for 1.2.
                self._lastchar = unichr(int(self._lastchar))
            except ValueError: # Occasional Unicode related exceptions here - see above for why!!
                print "Bad integer - self._lastchar =", repr(self._lastchar)
                raise
            self.doCommandCompletion(self._lastchar)
            controller.keyPressed(self._lastchar)
            
        # this is where we do things like brace matching and guide highlighting
        if controller.doBraceMatching:
            # only go through the pain of bracematching if we realy need to
            # idealy, the braces would be in the language service, but not necessary right now
            if self._lastchar is None or self._lastchar in "{}[]():":
                controller.showMatchingBrace()
            elif self._scintilla.highlightGuide >= 0:
                self._scintilla.braceHighlight(-1, -1)
                self._scintilla.highlightGuide = -1
        self._lastchar = None

    def onModified( self, position, modificationType, text,
                    linesAdded, line, foldLevelNow, foldLevelPrev ):
        # Result: void (None)
        # In: param0: int32
        # In: param1: int32
        # In: param2: string
        # In: param4: int32
        # In: param5: int32
        # In: param6: int32
        # In: param7: int32
        
        # If we are performing a mass-edit, don't check autocomplete etc
        if self._inBatchOperation: return

        if modificationType & sci_constants.SC_MOD_DELETETEXT \
           and self.commandCompletionSvc is not None and self._doAutoComplete\
           and (not self._triggers or (text and (text[:1] in self._triggers))) and self._scintilla.autoCActive():
            #print "canceling autocomplete"
            self._scintilla.autoCCancel()
        
        # if we delete or insert text, signal command completion to update information
        # don't do it on a single char change, looking for eol or mass paste/delete
        if (len(text) > 1 or linesAdded > 0) and (modificationType & sci_constants.SC_MOD_DELETETEXT \
           or modificationType & sci_constants.SC_MOD_INSERTTEXT) \
           and self.commandCompletionSvc is not None and self._doAutoComplete:
           self.commandCompletionSvc.scanBufferCompletion(self._scintilla, self.url)

    def onMarginClick(self, modifiers, position, margin):
        if margin == 1:
            # If the language doesnt support folding, we can't get
            # a click on the margin!
            s = self._scintilla
            lineclick = s.lineFromPosition(position)
            if (s.getFoldLevel(lineclick) & sci_constants.SC_FOLDLEVELHEADERFLAG):
                if s.getFoldExpanded(lineclick):
                    # check to see if cursor is inside fold, if so, move it to a visible place
                    level = s.getFoldLevel(lineclick)
                    lineMaxSubord = s.getLastChild(lineclick, level)
                    currentLine = s.lineFromPosition(s.currentPos)
                    if currentLine > lineclick and currentLine <= lineMaxSubord:
                        pos = s.positionFromLine(lineclick)
                        s.selectionStart = pos
                        s.selectionEnd = pos
                        s.currentPos = pos
                s.toggleFold(lineclick)
        else:
            pass

    def onZoom(self):
        if self._scintilla.getMarginWidthN(0):
            self._updateLineNumberMargin()
    
    def _updateLineNumberMargin(self):
        s = self._scintilla    
        s.setMarginWidthN(0, s.textWidth(0, str(max(1000,s.lineCount*2)))+5)

    # nsIObserver interface
    def Observe(self, prefSet, topic, data):
        if topic == self.viewPrefs.id or topic == self._globalPrefs.id or topic == self.stylePrefs.id:
            # Dispatch a preference change...
            self._dispatchPrefChange(prefSet, data)
        else:
            logWarning("Unexpected Observe() topic %s" % data)

    # Probably should make this function table-based to reduce
    # duplication of effort.
    def _dispatchPrefChange(self, prefSet, prefName):
        if hasattr(self, "_apply_" + prefName):
            getattr(self, "_apply_" + prefName)(prefSet)

    def _apply_showWhitespace(self, prefSet):
        self._scintilla.viewWS = prefSet.getBooleanPref('showWhitespace')

    def _apply_editUseAutoWrap(self, prefSet):
        self.editUseAutoWrap= prefSet.getBooleanPref('editUseAutoWrap')
        if self.editUseAutoWrap:
            self._scintilla.wrapMode = self._scintilla.SC_WRAP_WORD
            self._scintilla.layoutCache = self._scintilla.SC_CACHE_PAGE
        else:
            self._scintilla.wrapMode = self._scintilla.SC_WRAP_NONE
            self._scintilla.layoutCache = self._scintilla.SC_CACHE_NONE

    def _apply_editUseEdge(self, prefSet):
        self._updateEdge(prefSet)

    def _updateEdge(self, prefSet):
        if prefSet.getBooleanPref('editUseEdge'):
            self._scintilla.edgeColumn= prefSet.getLongPref('editAutoWrapColumn')
            if self.viewPrefs.getBooleanPref('editUseFixedFont'):
                self._scintilla.edgeMode = self._scintilla.EDGE_LINE
            else:
                self._scintilla.edgeMode = self._scintilla.EDGE_BACKGROUND
        else:
            self._scintilla.edgeMode = self._scintilla.EDGE_NONE
            
    def _apply_editAutoWrapColumn(self, prefSet):
        self.editAutoWrapColumn = prefSet.getLongPref('editAutoWrapColumn')
        self._updateEdge(prefSet)

    def _apply_showIndentationGuides(self, prefSet):
        self._scintilla.indentationGuides = prefSet.getBooleanPref('showIndentationGuides')

    def _apply_editUseAutoComplete(self, prefSet):
        self._doAutoComplete = self.viewPrefs.getBooleanPref('editUseAutoComplete')

    def _apply_editUseCallTips(self, prefSet):
        self._doStartCallTip = self.viewPrefs.getBooleanPref('editUseCallTips')

    def _apply_showEOL(self, prefSet):
        self._scintilla.viewEOL = prefSet.getBooleanPref('showEOL')

    def _apply_editUseFixedFont(self, prefSet):
        self._styleApplier.applyStyles()
        self._updateEdge(prefSet)

    def _apply_showLineNumbers(self, prefSet):
        toggle = prefSet.getBooleanPref('showLineNumbers')
        if toggle:
            self._updateLineNumberMargin()
        else:
            self._scintilla.setMarginWidthN(0, 0)

    def _apply_useTabs(self, prefSet):
        self._controller.useTabs =  prefSet.getBooleanPref('useTabs')
        # Make sure to actually call the configure, so SciMoz gets
        # the changes reflected in
        self._controller.configureIndentationParams()

    def _apply_indentWidth(self, prefSet):
        self._controller.indentWidth =  prefSet.getLongPref('indentWidth')
        # Make sure to actually call the configure, so SciMoz gets
        # the changes reflected in
        self._controller.configureIndentationParams()

    def _apply_tabWidth(self, prefSet):
        self._controller.tabWidth =  prefSet.getLongPref('tabWidth')
        # Make sure to actually call the configure, so SciMoz gets
        # the changes reflected in
        self._controller.configureIndentationParams()

    def _apply_editFoldStyle(self, prefSet):
        # use margin 1 for folding
        if not self._show_folds: return
        foldstyle = prefSet.getStringPref('editFoldStyle')
        if foldstyle != 'none':
            self._enableFolding(1, foldstyle)
            lastLine = self._scintilla.firstVisibleLine + self._scintilla.linesOnScreen
            self._scintilla.colourise(0, self._scintilla.positionFromLine(lastLine))
            self._scintilla.setMarginWidthN(1, 15)
        else:
            self._scintilla.showLines(0, self._scintilla.lineCount-1)
            for line in range(self._scintilla.lineCount):
                if self._scintilla.getFoldLevel(line) & self._scintilla.SC_FOLDLEVELHEADERFLAG:
                    self._scintilla.setFoldExpanded(line, 1)
            # If we don't do this, folding trails off into other buffers.
            self._scintilla.setProperty("fold", "0")
            self._scintilla.setMarginWidthN(1, 0)
            
    def _apply_encoding(self, prefSet):
        if prefSet.hasStringPref('encoding'):
            encoding_name = prefSet.getStringPref('encoding')
            if self.encoding.python_encoding_name != encoding_name:
                self.setEncodingFromEncodingName(encoding_name)
    
    # A list of view-related preferences.
    _viewPrefList = ['showWhitespace', 'showIndentationGuides', 'showEOL',
                    'editUseFixedFont', 'showLineNumbers',
                     'editUseAutoComplete', 'editUseCallTips',
                     'editUseAutoWrap', 'editAutoWrapColumn', 'editUseEdge',
                     'encoding', 'editFoldStyle',
                ]

    def _applyViewPrefs(self):
        for prefName in koEditorWrapper._viewPrefList:
            self._dispatchPrefChange(self.viewPrefs, prefName)
        self.get_controller().applyIndentationParams()
            
    def _apply_Default_fixed(self, prefSet):
        self._styleApplier.applyStyles()
            
    def _apply_Default_proportional(self, prefSet):
        self._styleApplier.applyStyles()


    def getUnsavedChanges(self):
        import difflib
        eolMode = self._scintilla.eOLMode
        eol = {components.interfaces.ISciMoz.SC_EOL_CR: '\r',
               components.interfaces.ISciMoz.SC_EOL_CRLF: '\r\n',
               components.interfaces.ISciMoz.SC_EOL_LF: '\n'}[eolMode]
        inmemory = self._scintilla.text.splitlines(1)
        ondisk = open(convertURL.LocalPathFromURL(self.url), 'rb').read().splitlines(1)
        difflines = difflib.ndiff(ondisk, inmemory)
        # difflib adds \n to its new lines, regardless of what was sent in
        fixed_difflines = []
        for line in difflines:
            if line.startswith('? '):
                fixed_difflines.append(line[:-1] + eol)
            else:
                fixed_difflines.append(line)
        return ''.join(fixed_difflines)


