# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

# An implementation of a ftp file objects
#
# Contributors:
# * Shane Caraveo

from xpcom import file, components, COMException, ServerException, server, nsError
from xpcom.server.enumerator import SimpleEnumerator

from koLogger import *
import tempfile
import codecs

# needed for ftpfile
import urlparse, urllib, re, os
import ftplib, StringIO, string
import time, strptime
import koUnicodeEncoding

re_parse_unix_ftp_line = re.compile(r'^.([rwxXstugo-]*)\s+(.*?\s+.*?\s+.*?)\s*(\d*?)\s+(\w+)\s+(\d+)\s+([\d:]+)\s(.*)$')
re_parse_dos_ftp_line = re.compile(r'^(..-..-..\s+..:....)\s+(.*?)\s+(.*)$')

def addslash(path):
    if len(path) < 1: return "/"
    if path[len(path)-1] != "/": path += "/"
    return path

def normalizePath(path):
    path = os.path.normpath(path) #normalize the path
    path = string.replace(path, "\\","/") # os.normpath changes slashes on win
    if path[0] != "/": path = "/"+path
    return path

#must persist the last username/password in order to properly open files from koEditorWrapper
sessionData = {}
def saveSession(key, data):
    global sessionData
    sessionData[key] = data
    
def getSession(key):
    global sessionData
    if sessionData.has_key(key):
        return sessionData[key]
    return None

class koFTPConnection:
    _com_interfaces_ = [components.interfaces.koIFTPConnection]
    _reg_desc_ = "Komodo FTP Connection"
    _reg_contractid_ = "@activestate.com/koFTPConnection;1"
    _reg_clsid_ = "{76A0956E-044C-45fc-BD9B-07BC5C659E4A}"

    def __init__(self):
        self.connection = None
        self._lasterror = ""
        self._directoryCache = {}
        # can only read/write one file at a time
        self.filedata = StringIO.StringIO()
        self._globalPrefs = \
            components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
    
    def doAuthentication(self):
        # if we don't have it, pop up a dialog asking for it!
        self.authAttempt = self.authAttempt + 1
        dialogproxy = components.classes['@activestate.com/asDialogProxy;1'].\
                    createInstance(components.interfaces.asIDialogProxy)
        title = "Authenticate"
        if self._lasterror:
            title = "Authentication Failed, try again..."
        prompt = "Please enter the username and password for "+self.server
        authinfo = dialogproxy.authenticate(title, prompt, self.username, 1, 0)
        if authinfo:
            #print "got back ",authinfo
            p = string.split(authinfo, ":",1)
            self.username = p[0]
            self.password = p[1]
            return 1
        return 0
            
    def open(self, server, port, username, password):
        self.authAttempt = 0
        self.server = server
        if port > 0: self.port = port
        else: self.port = 21
        self.username = username
        self.password = password
        
        key = "%s:%s"%(self.server,self.port)
        #print "session data %s:%s:%s"%(self.server,self.port,self.username)
        if not self.username or not self.password:
            sess = getSession(key)
            if sess:
                self.username = sess[0]
                self.password = sess[1]
        
        if not self.username or not self.password:
            if not self.doAuthentication():
                self._lasterror = "Authentication information required"
                raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
                
        self._lasterror = ""
        # open the connection
        try:
            self.connection = ftplib.FTP()
            self.connection.connect(self.server, self.port)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP OPEN ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        
        # attempt authentication
        while 1:
            try:
                self.connection.login(self.username, self.password)
                break
            except ftplib.all_errors, e:
                self._lasterror = e.args[-1]
                if self._lasterror[:3] == "530":
                    # username/password error
                    if self.authAttempt < 3 and self.doAuthentication(): continue
                print "FTP OPEN ERROR: ", self._lasterror
                raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
                #return 0
        
        #save the session
        saveSession(key,[self.username,self.password])
        return 1

    def retrlines(self, cmd, callback):
        '''Retrieve data in line mode.
        The argument is a RETR or LIST command.
        The callback function (2nd argument) is called for each line,
        with trailing CRLF stripped.  This creates a new port for you.
        print_line() is the default callback.'''
        resp = self.connection.sendcmd('TYPE I')
        conn = self.connection.transfercmd(cmd)
        fp = conn.makefile('rb')
        while 1:
            line = fp.readline()
            if not line:
                break
            if line[-2:] == '\r\n':
                line = line[:-2]
            elif line[-1:] == '\n':
                line = line[:-1]
            callback(line)
        fp.close()
        conn.close()
        return self.connection.voidresp()
    
    def list(self, path, refresh):
        #print "list ",path
        info = []
        self._lasterror = ""
        orig_path = None
        pwd = None
        try:
            if not refresh and self._directoryCache.has_key(path):
                return self._directoryCache[path]
            
            pwd = self.currentDirectory()
            # for some reason, using LIST does not work with quoted paths
            # but we can cwd to a LFN path, then get a list of files
            # to handle LFN
            try:
                self.connection.cwd(path)
            except ftplib.all_errors, e:
                # we have a filename, lets get the parent path
                # and cleanup at the end
                orig_path = path
                path = os.path.dirname(path)
                self.connection.cwd(path)
            
            self.retrlines("LIST -al", info.append)
            #self.connection.retrlines("LIST -al", info.append)
                    
            # detect directories, parse for cache
            # this is a directory
            lines = []
            for line in info:
                # XXX this is still not entirely correct but
                # at least allows us to view unicode filenames
                # for the most part.
                line, enc, bom = koUnicodeEncoding.\
                                autoDetectEncoding(line)
                if not line:
                    print "Encounted undeciferable encoding ",line
                    continue
                leafName = ""
                # don't cache a directory this way
                if line[0] == "d": 
                    #unix line
                    fig = re_parse_unix_ftp_line.search(line)
                    if fig:
                        fi = fig.groups()
                        if fi[6]=="." or fi[6]=="..": continue
                elif line[0] in ["-","l"]: 
                    #unix line
                    fig = re_parse_unix_ftp_line.search(line)
                    if fig:
                        fi = fig.groups()
                        leafName = fi[6]
                else:
                    #try dos line
                    fig = re_parse_dos_ftp_line.search(line)
                    if fig: 
                        fi = fig.groups()
                        if str(fi[1][:5]) != '<DIR>':
                            leafName = fi[2]
                    else: continue
                if leafName:
                    fpath = addslash(path) + leafName
                    self._directoryCache[fpath] = [line]
                
                # if we got here, add the line to the directory cache
                lines.append(line)
                        
            self._directoryCache[path] = lines
                    
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP LIST ERROR: ", self._lasterror
            # don't raise, this is expected behaviour if creating a directory or file
            pass

        # reset to the directory we were in before.                
        if pwd:
            self.changeDirectory(pwd)
                
        # get this from cache to properly handle long file name issues
        if orig_path:
            if self._directoryCache.has_key(orig_path):
                return self._directoryCache[orig_path]
            # if we don't have a cache entry, the file did not exist!
            return []

        return info
    
    def refresh(self, path):
        # remove path and all children from cache
        if not self._directoryCache.has_key(path): return
        for key in self._directoryCache.keys():
            if key[:len(path)] == path:
                del self._directoryCache[key]
        self.list(path, 1)
    
    def haveCacheEntry(self, path):
        return self._directoryCache.has_key(path)
    
    def havePath(self, path):
        haveit = self._directoryCache.has_key(path)
        if not haveit:
            info = self.list(path, 0)
            haveit = len(info) > 0
        return haveit

    def rename(self, oldName, newName):
        #print "rename %s to %s" % (oldName, newName)
        try:
            self.connection.rename(oldName, newName)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP RENAME ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def removeFile(self, name):
        #print "remove ",name
        try:
            self.connection.delete(name)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP REMOVE ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def changeDirectory(self, path):
        #print "cwd ",path
        try:
            self.connection.cwd(path)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP CWD ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def currentDirectory(self):
        #print "cwd ",path
        try:
            return self.connection.pwd()
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP PWD ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def getHomeDirectory(self):
        #print "cwd ",path
        try:
            pwd = self.connection.pwd()
            self.connection.cwd("~")
            home = self.connection.pwd()
            self.connection.cwd(pwd)
            return home
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP HOMEDIR ERROR: ", self._lasterror
        return "/"
            
    def removeDirectory(self, name):
        #print "remove ",name
        try:
            self.connection.rmd(name)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP REMOVE ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def createDirectory(self, name, permissions):
        #print "createDirectory %s, %d" % (name, permissions)
        try:
            self.connection.mkd(name) 
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP CREATE ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)

    def close(self):
        self._lasterror = ""
        try:
            self.connection.quit() 
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP CLOSE ERROR: ", self._lasterror
            pass

    def _read_cb(self, block):
        #print "read %d bytes"  % (len(block))
        self.filedata.write(block)

    def readFile(self, filename):
        try:
            if self.filedata.tell() == 0:
                # retreive the file
                self.connection.retrbinary("RETR "+filename, self._read_cb)
                self.filedata.seek(0,0)
            data = self.filedata.getvalue()
            # reset filedata
            self.filedata.close()
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP RETR ERROR: ", self._lasterror
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        self.filedata = StringIO.StringIO()
        return data
    
    def writeFile(self, filename, data):
        self.filedata.write(data)
        self.filedata.seek(0,0)
        self._lasterror = ""
        try:
            self.connection.storbinary("STOR "+filename, self.filedata, 8192)
        except ftplib.all_errors, e:
            self._lasterror = e.args[-1]
            print "FTP STOR ERROR: ", self._lasterror
            self.filedata.close()
            self.filedata = StringIO.StringIO()
            raise ServerException(nsError.NS_ERROR_FAILURE, self._lasterror)
        self.filedata.close()
        self.filedata = StringIO.StringIO()
    
    def get_lastError(self):
        return self._lasterror
    
    
class koFTPNode:
    _com_interfaces_ = [components.interfaces.koIFTPNode, components.interfaces.koIFTPNodeFile]
    _reg_desc_ = "Komodo FTP Node"
    _reg_contractid_ = "@activestate.com/koFTPNode;1"
    _reg_clsid_ = "{2747A621-9F3B-400d-B724-B6CA0F5F3027}"

    def __init__(self):
        self.connection = None
        self._parent = None
        self._children = []
        self._path = None
        self._readable = 0
        self._writeable = 0
        self._isDirectory = 0
        self._isFile = 0
        self._isHidden = 0
        self._isSymlink = 0
        self.fileSize = 0
        self.lastModificationDate = 0
        self.leafName = ""
        self.server = ""
        self.port = 21
        self.username = ""
        self.password = ""
        self.followlinks = 1
        self.URL = ""
        self.diskSpaceAvailable = 0
        self.persistentDescriptor = ""
        self._target = ""
        
        # no support for the following attributes yet
        self.NORMAL_FILE_TYPE = 0
        self.DIRECTORY_TYPE = 1
        self.permissions = 0
        self.permissionsOfLink = 0
        self.lastModificationDateOfLink = 0
        self.fileSizeOfLink = 0

        self._globalPrefs = \
            components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
            
    def init(self, server, port, username, password):
        if not self.connection:
            self.connection = \
                components.classes['@activestate.com/koFTPConnection;1'].\
                createInstance(components.interfaces.koIFTPConnection)
            #self.connection = koFTPConnection()
            #print "%s %d %s %s" %(server,port,username,password)
            self.connection.open(server, port, username, password)

    def normalize(self):
        self._path = normalizePath(self._path)

    def getChildren(self):
        return self._children
    
    def get_parent(self):
        self._getParent()
        return self._parent
    
    def _getParent(self):
        if not self._parent and self._path != "/":
            self._parent = components.classes['@activestate.com/koFTPNode;1'].\
                        createInstance(components.interfaces.koIFTPNode)
            # get the parent path
            parts = string.split(self._path, "/")
            parent_path = ""
            if len(parts) > 2:
                parts = parts[:len(parts)-1]
                parent_path = string.join(parts, "/") + "/"
            if not parent_path:
                parent_path = "/"
            #parent_path = self._path[:len(self._path)-len(self.leafName)]
            #print "getting parent for ",self._path," lf: ",self.leafName," at ",parent_path
            self._parent.initWithConnection(self.connection, parent_path)
    
    def initClone(self, parent, path):
        self._parent = parent
        if parent: self._children = parent.getChildren()
        self.initWithConnection(parent.connection, path)
            
    def initWithConnection(self, connection, path):
        self.connection = connection
        self.set_path(path)

    def initChild(self, parent, info):
        self._parent = parent
        self.initWithConnectionInfo(parent.connection, info)

    def initWithConnectionInfo(self, connection, info):
        self.connection = connection
        self._parseFTPLine(info)
        if self._isSymlink and self.followlinks:
            # automaticly resolve symlinks
            #print "initWithConnectionInfo attempting to resolve link: ",self._target
            leafname = self.leafName
            self.set_path(self._target)
            if not self.leafName:
                if leafname[0]=="/": leafname = leafname[1:]
                self.leafName = leafname
            #print "initWithConnectionInfo path : %s leafname : %s " %(self._path,self.leafName)
        self._path = addslash(self._parent.path) + self.leafName
        self._generateURI()
    
    def get_unicodeURL(self):
        return self.URL
    
    def set_unicodeURL(self,url):
        self.set_URL(url)
        
    def set_URL(self,url):
        # XXX the path should be reset if URL is set
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    def _generateURI(self):
        # we do not include username/password in the URI!
        # those must be retrieved from prefs
        uri = "ftp://"+self.connection.server+":"+str(self.connection.port)+self._path
        self.URL = uri

    def _parseFTPLine(self, line):
        #print "parseline: ",line
        if line[0] in ["-","d","l"]:
            # unix style lines- not done
            fig = re_parse_unix_ftp_line.search(line)
            if not fig: return
            fi = fig.groups()
            
            self._isDirectory = line[0] == "d"
            self._isFile = line[0] == "-"
            self._isHidden = 0
            self._isSymlink = line[0] == "l"
            # we unfortunately dont know if we are owner or group member, so guess again
            self._readable = "r" in fi[0]
            self._writeable = "w" in fi[0]
            
            if fi[2]: self.fileSize = string.atoi(fi[2])
            else: self.fileSize = 0

            # do we have time, or year?
            month,day,year = fi[3:6]
            if year.find(":") > -1:
                hour = year
                # fix time to be 5 digit always
                if len(hour)<5: hour = "0"+hour
                year = "%d"%(time.gmtime(time.time())[0])
            else:
                hour = "00:00"
            
            if len(day) < 2: day = "0"+day
            
            date = string.join([month,day,year,hour]," ")
            #print "DATE: ",date
            t = strptime.strptime(date,'%b %d %Y %H:%M')
            
            # print "date trans: ",repr(t)
            self.lastModificationDate = time.mktime(t)*1000
            
            if self._isSymlink:
                self.leafName, self._target = fi[6].split(" -> ")
                #print "Symlink %s -> %s" %(self.leafName, self._target)
            else:
                self.leafName = fi[6]
            #if self._isDirectory: self.leafName += "/"
            #print "D %d R %d W %d Name %s" % (self._isDirectory, self._readable, self._writeable, self.leafName)
        else:
            # msdos style lines
            fig = re_parse_dos_ftp_line.search(line)
            if not fig: return
            fi = fig.groups()
            # we simply don't know, so say yes!

            self._readable = 1
            self._writeable = 1
            self._isDirectory = 0
            self._isFile = 1
            self._isHidden = 0
            self._isSymlink = 0
            
            try:
                self.fileSize = string.atoi(fi[1])
            except ValueError, e:
                # this is a directory
                if str(fi[1][:5]) == '<DIR>':
                    self._isDirectory = 1
                    self._isFile = 0
                    self.fileSize = 0
                else:
                    pass
            t = strptime.strptime(fi[0],'%m-%d-%y  %I:%M%p')
            # print "date trans: ",repr(t)
            self.lastModificationDate = time.mktime(t)*1000
            self.leafName = fi[2]
            #print "D %d R %d W %d Size %d Time %d Name %s" % (self._isDirectory, self._readable, self._writeable, self.fileSize, self.lastModificationDate, self.leafName)
    
    def get_path(self):
        return self._path
    
    def set_path(self, path):
        # parse path for specific info
        if not path or path == "": path = "/"
        
        self._path = path
        self.normalize()
        if not self.connection: return
        info = self.connection.list(path, 0)
        num_resp = len(info)
        # if num_resp == 0, file does not exist, we may be attempting
        # to write a new file, so we will set defaults simular to directory
        if num_resp <> 1:
            #if file wasn't found, lastError will contain an error, otherwise
            # this is likely a directory
            if self.connection.lastError == "":
                # this is a directory
                # XXX, how do I get it's info?
                self._isDirectory = 1
                self._readable = 1
                self.lastModificationDate = time.time()*1000
                self.fileSize = 1
            else:
                logError("ERROR: "+self.connection.lastError)
                self._isDirectory = 0
                self._readable = 0
                self.lastModificationDate = 0
                self.fileSize = 0
                    
            self._writeable = 1
            self._isFile = 0
            self._isHidden = 0
            self._isSymlink = 0
            self.leafName = ""
                
        elif num_resp==1:
            line, enc, bom = koUnicodeEncoding.\
                autoDetectEncoding(info[0])
            if not line:
                raise "Encounted undeciferable encoding %s"%line
            # definitly a file
            self._parseFTPLine(line)
            # we still may be looking at a directory with one entry, figure it out
            if not self._isDirectory and \
               os.path.basename(self._path) <> self.leafName:
                # this is a directory with a single entry, revert to a directory
                self._isDirectory = 1
                self._readable = 1
                self.lastModificationDate = time.time()*1000
                self.fileSize = 1
                self._writeable = 1
                self._isFile = 0
                self._isHidden = 0
                self._isSymlink = 0
                self.leafName = ""
                    
            if self._isSymlink and self.followlinks:
                # automaticly resolve symlinks
                #print "attempting to resolve link: ",self._target
                self._path = addslash(self._path) + self.leafName
                self.normalize()
            #print "path : %s leafname : %s " %(repr(self._path),self.leafName)
                
        self._generateURI()
            
    def _generateChildren(self, path):
        children = self.connection.list(self._path, 0)
        nodes = []
        # we have more listings lets make some nodes
        for info in children:
            node = components.classes['@activestate.com/koFTPNode;1'].\
                        createInstance(components.interfaces.koIFTPNode)
            node.initChild(self, info)
            nodes.append(node)
        return nodes
            
    def get_directoryEntries(self):
        if not self.connection:
            return SimpleEnumerator([])
        if len(self._children) < 1:
            # or not self.connection.haveCacheEntry(self._path):
            self._children = self._generateChildren(self._path)
        return SimpleEnumerator(self._children)
    
    def exists(self):
        if not self.connection: return 0
        return self.connection.havePath(self._path)
    
    def close(self):
        self.connection.close()
        self.connection = None
    
    def equals(self, node):
        if self._path == node.path and self.fileSize == node.fileSize \
            and self.lastModificationDate == node.lastModificationDate:
                return 1
        return 0
    
    def clone(self):
        node = components.classes['@activestate.com/koFTPNode;1'].\
                    createInstance(components.interfaces.koIFTPNode)
        parent = self.get_parent()
        if parent:
            node.initClone(parent, self._path)
        else:
            node.initWithConnection(self.connection, self._path)
        return node
    
    def append(self, relativePath):
        if not relativePath or relativePath[0] == "/":
            # see nsLocalFileWin.cpp for related functionality
            raise ServerException(nsError.NS_ERROR_FILE_UNRECOGNIZED_PATH, relativePath)
        path = self._path
        if path == "/":
            path += relativePath
        else:
            path += "/"+relativePath
        #print "new path is ",path
        self.set_path(path)
    
    def appendUnicode(self, relativePath):
        self.append(relativePath)

    def dump(self):
        print "node: %s leafname: %s isdir: %d uri: %s" % \
                (self._path, self.leafName, self._isDirectory, self.uri)

    #############################################################
    # koIFTPNodeFile interface functions
    #############################################################
    def initWithPath(self, filePath):
        self.set_path(filePath)
    
    def initWithUnicodePath(self, filePath):
        self.set_path(filePath)
            
    #[noscript] PRFileDescStar openNSPRFileDesc(in long flags, in long mode);
    def openNSPRFileDesc(self, flags, mode):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    #[noscript] FILE           openANSIFileDesc(in string mode);
    def openANSIFileDesc(self, mode):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    #[noscript] PRLibraryStar  load();
    def load(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    def appendRelativePath(self, relativeFilePath):
        self.append(relativeFilePath)
    
    def appendRelativeUnicodePath(self, relativeFilePath):
        self.appendUnicode(relativeFilePath)
    
    def reveal(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    def launch(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    
    #############################################################
    # realy basic stuff
    #############################################################
    
    def get_unicodeLeafName(self):
        return self.leafName
    
    def set_unicodeLeafName(self, leafName):
        self.leafName = leafName

    def get_target(self):
        if self._isSymlink:
            return self._target
        else:
            raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def get_unicodeTarget(self):
        return self.get_target()
    
    def get_unicodePath(self):
        return self.get_path()
    
    def isWritable(self):
        return self._writeable != 0
    
    def isReadable(self):
        return self._readable != 0
    
    def isExecutable(self):
        return 0
    
    def isHidden(self):
        return self._isHidden != 0
    
    def isDirectory(self):
        return self._isDirectory != 0

    def isFile(self):
        return self._isFile != 0

    def isSymlink(self):
        return self._isSymlink != 0

    def isSpecial(self):
        return 0

    def moveTo(self, newParentDir, newName):
        # moveTo only does a rename in ftp
        parent = self.get_parent()
        if newParentDir is not None:
            newname = addslash(newParentDir)+newName
        elif parent:
            newname = addslash(parent.path)+newName
        else:
            path = re.search(r'^(.*)/', self._path)
            newname = path + "/"+newName

        if string.find(newname, "..") > -1:
            newname = normalizePath(newName)
        
        self.connection.rename(self._path, newname)
        if parent:
            self.connection.refresh(parent.path)

    def moveToUnicode(self, newParentDir, newName):
        #either param can be None
        self.moveTo(newParentDir, newName)

    def remove(self, recursive):
        if self._isDirectory:
            self.connection.removeDirectory(self._path)
        else:
            self.connection.removeFile(self._path)
        parent = self.get_parent()
        if parent:
            self.connection.refresh(parent.path)

    def createUnique(self, suggestedName, type, permissions):
        # since we append the new directory name before calling createUnique,
        # self._isDirectory will always be false
        #if not self._isDirectory:
        #        raise COMException(nsError.NS_ERROR_NOT_IMPLEMENTED)
        if type == self.DIRECTORY_TYPE:
            # XXX
            # duplicate behaviour of nsIFile, it seems to ignore the suggestedName,
            # and creates the directory that is the path.  You must call append to
            # get the file object to have the correct path, before calling createUnique
            # 
            # path = addslash(self._path)+suggestedName
            # path = normalizePath(path)
            # self.connection.createDirectory(path, permissions)
            self.connection.createDirectory(self._path, permissions)
            parent = self.get_parent()
            if parent:
                self.connection.refresh(parent.path)
        else:
            raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    #############################################################
    # NOT IMPLEMENTED!
    #############################################################
    
    def create(self, type, permissions):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def copyTo(self, newParentDir, newName):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
            
    def copyToUnicode(self, newParentDir, newName):
        self.copyTo(newParentDir, newName)

    def copyToFollowingLinks(self, newParentDir, newName):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def copyToFollowingLinksUnicode(self, newParentDir, newName):
        self.copyToFollowingLinks(newParentDir, newName)

    def spawn(self, args):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def contains(self, inFile,  recursive):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)


class koFTPNodeFile(koFTPNode):
    _com_interfaces_ = [components.interfaces.koIFTPNode, components.interfaces.koIFTPNodeFile, components.interfaces.nsILocalFile]
    _reg_desc_ = "Komodo FTP Node File"
    _reg_contractid_ = "@activestate.com/koFTPNodeFile;1"
    _reg_clsid_ = "{CA20E22B-99D9-4e40-B156-521017549B95}"
    
    def __init__(self):
        koFTPNode.__init__(self)

    #  def initWithPath(self, filePath):
    #          koFTPNode.initWithPath(self, filePath)
    #  
    #  def initWithUnicodePath(self, filePath):
    #          koFTPNode.initWithUnicodePath(self, filePath)
    #          
    #  #[noscript] PRFileDescStar openNSPRFileDesc(in long flags, in long mode);
    #  def openNSPRFileDesc(self, flags, mode):
    #          return koFTPNode.openNSPRFileDesc(self, flags, mode)
    #          
    #  #[noscript] FILE           openANSIFileDesc(in string mode);
    #  def openANSIFileDesc(self, mode):
    #          return koFTPNode.openANSIFileDesc(self, mode)
    #          
    #  #[noscript] PRLibraryStar  load();
    #  def load(self):
    #          return koFTPNode.load()
    #
    #  def appendRelativePath(self, relativeFilePath):
    #          koFTPNode.appendRelativePath(self, relativeFilePath)
    #  
    #  def appendRelativeUnicodePath(self, relativeFilePath):
    #          koFTPNode.appendRelativeUnicodePath(self, relativeFilePath)
    #  
    #  def reveal(self):
    #          koFTPNode.reveal(self)
    #  
    #  def launch(self):
    #          koFTPNode.launch(self)


# FTP File support
class FTPFile:
    _com_interfaces_ = [components.interfaces.koIFTPFile]
    _reg_desc_ = "Komodo FTP File"
    _reg_contractid_ = "@activestate.com/koFTPFile;1"
    _reg_clsid_ = "{010656B6-E830-45ae-B7D9-F22C1E4AFDEC}"

    def __init__(self):
        self.password = None
        self.username = None
        self._globalPrefs = None
        self.filename = ""
        self.filedata = StringIO.StringIO()
        self.ftpfile = None
            
    def initServer(self, url, mode="r", server_alias=""):
        #print "initServer url=",url
        self.server_opts = None
        if server_alias:
            # get username and password from prefs
            if not self._globalPrefs:
                self._globalPrefs = \
                    components.classes["@activestate.com/koPrefService;1"].\
                    getService(components.interfaces.koIPrefService).prefs
            deploymentServerList = self._globalPrefs.getPref("deploymentServerList");

            for i in range(0, deploymentServerList.length):
                dserver = deploymentServerList.getPref(i).QueryInterface(components.interfaces.koIOrderedPreference)
                if server_alias == dserver.getStringPref(0):
                    # we found our match
                    so = dserver.getStringPref(1)
                    self.server_opts = so.split(":")
                    #print "server_opts=",so, repr(self.server_opts)
                    if self.server_opts[0].lower() != "ftp":
                            raise ValueError, "FTPFile Mismatched Protocol "
                    self.username = self.server_opts[4]
                    self.password = self.server_opts[5]
                    crypto = components.classes['@activestate.com/koPasswordCrypto;1'].\
                            getService(components.interfaces.koIPasswordCrypto)
                    # if the password is empty, that's ok, we'll ask for it later
                    if self.password:
                        try:
                            self.password = crypto.decodePassword(self.password)
                        except:
                            logWarning("Dropping password that could not be decoded. See bug 18752.")
                    #print "username: %s password %s" % (self.username, self.password)
                    break
        self.init(url, mode)
            
    def init(self, url, mode="r"):
        parts = urlparse.urlparse(url)
        if parts[0] <> "ftp":
            raise ValueError, "FTPFile Unknown Protocol"
        
        # get the username:password@domain
        #u = re.search("^(([^:@]+)(:([^:@]+))?@)?([^:@]+)(:(.*))?",parts[1]).groups()
        #print "host info: ",parts[1]
        ug = re.search("^(([^:@]+)(:([^:]+))?@)?([^:@]+)(:(.*))?",parts[1])
        if not ug:
            raise ValueError, "Regular Expression Failure in %s" % (parts[1])
            return
        u = ug.groups()
        #print "re: ",repr(u)
        if not self.username: self.username = u[1]
        if not self.password: self.password = u[3]
        self.domain = u[4]
        self.port = u[6] # not used yet
        if not self.port or self.port == "0": self.port = "21"
        #print "%s:%s@%s" %(username,password,domain)

        self.fullpath = parts[2]
        self.mode = mode
        # make the connection to the ftp site
        self.ftpfile = components.classes["@activestate.com/koFTPNodeFile;1"] \
                        .createInstance(components.interfaces.koIFTPNodeFile);
        self.ftpfile.init(self.domain, string.atoi(self.port), self.username, self.password)
        self.ftpfile.initWithUnicodePath(self.fullpath)
    
    def __del__(self):
            self.close()

    def read(self, n = -1):
        
        #print "attempting to read ftp file, size", n
        if not self.filedata or not self.ftpfile or not self.ftpfile.isReadable():
            logWarning("unable to read file "+self.fullpath+", readable: "+self.ftpfile.isReadable())
            return None
        if self.filedata.tell() == 0:
            # retrieve the file
            data = self.ftpfile.connection.readFile(self.fullpath)
            #print "read in ",len(data)," bytes"
            self.filedata.write(data)
            self.filedata.seek(0,0)
        if n == -1:
            return self.filedata.getvalue()
        else:
            return self.filedata.read(n)

    def write(self, data):
        if not self.filedata or not self.ftpfile or not self.ftpfile.isWritable():
            logWarning("unable to write file "+self.fullpath+", writeable"+self.ftpfile.isWritable())
            return None
        self.filedata.write(data)

    def _store(self):
        if self.mode == 'w':
            self.filedata.seek(0,0)
            self.ftpfile.connection.writeFile(self.fullpath, self.filedata.getvalue())
        self.filedata.close()
        self.filedata = None
                    
            
    def close(self):
        if self.ftpfile:
            self._store()
            self.ftpfile = None

    def get_isWritable(self):
        if self.ftpfile:
            return self.ftpfile.isWritable()
        logWarning("no ftpfile "+self.fullpath+" available")
        return 0
    
    def get_isReadable(self):
        if self.ftpfile:
            return self.ftpfile.isReadable()
        logWarning("no ftpfile "+self.fullpath+" available")
        return 0
    
    
    