#!python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

# Provide general (read-only) information about a Komodo build/installation
#
# This is accomplished by reading komodo-info.txt, distributed with
# a Komodo build/installation.

import sys
import os
import re
import strptime
import time
from xpcom import components, nsError, ServerException, COMException
from koLogger import logError, logWarning, logDebug

# Defines not in the nsI IDL.
PROXY_SYNC    = 0x0001  # acts just like a function call.
PROXY_ASYNC   = 0x0002  # fire and forget.  This will return immediately and
                        # you will lose all return information.
PROXY_ALWAYS  = 0x0004  # ignore check to see if the eventQ is on the same
                        # thread as the caller, and alway return a proxied
                        # object.

class KoInfoService:
    _com_interfaces_ = [components.interfaces.koIInfoService]
    _reg_clsid_ = "{EB22F329-1D99-427a-B0E1-19DFF13AFBF7}"
    _reg_contractid_ = "@activestate.com/koInfoService;1"
    _reg_desc_ = "Komodo Information Service"

    def _getInfo(self, infoURL):
        #logDebug("Getting Komodo information from '%s'.\n" % infoURL)
        import xpcom.file
        fin = xpcom.file.URIFile(infoURL)
        datumRe = re.compile("^(?P<name>\w+)=(?P<value>.*)$")
        info = {}
        for line in fin.readlines():
            line = line.strip()
            datumMatch = datumRe.search(line)
            if datumMatch:
                info[datumMatch.group("name")] = datumMatch.group("value")
            else:
                logError("Bogus data line, '%s', in Komodo information "\
                         "file, '%s'.\n" % (line, infoURL))
                raise ServerException(nsError.NS_ERROR_UNEXPECTED)
        return info

    def __init__(self):
        self.platform = sys.platform
        self.isWindows = sys.platform.startswith("win")
        import platform
        self.osSystem = platform.system()
        self.osRelease = platform.release()
        self.osVersion = platform.version()
        if sys.platform.startswith("win"):
            import win32con
            self.VER_PLATFORM_WIN32_WINDOWS = win32con.VER_PLATFORM_WIN32_WINDOWS
        
        infoURL = "chrome://komodo/content/resources/komodo-info.txt"
        info = self._getInfo(infoURL)
        self.version = info["version"]
        self.buildNumber = info["buildNumber"]
        self.buildASCTime = info["buildASCTime"]
        releaseYear = int(info["releaseYear"])
        releaseMonth = int(info["releaseMonth"])
        releaseDay = int(info["releaseDay"])
        self.mozBinDir = info["mozBinDir"]
        self.buildType = info["buildType"]
        self.installerType = info["installerType"]
        self.productType = info["productType"]

        self.licensingType = info["licensingType"]

        if self.licensingType == "timebomb":
            expirationTime = int(info["timebombExpirationTime"])
            now = time.time()
            self.licenseValid = expirationTime >= now
            self.expires = 1
            self.daysUntilExpiration = int((expirationTime - now)/(60*60*24))
            self.expirationDate = time.strftime("%a %b %d, %Y",
                time.localtime(expirationTime))
            if abs(self.daysUntilExpiration) == 1:
                days = "day"
            else:
                days = "days"
            if self.daysUntilExpiration >= 0:
                self.licenseInfo = "This Komodo trial will expire on %s (%d %s left)."\
                    % (self.expirationDate, self.daysUntilExpiration, days)
            else:
                self.licenseInfo = "This Komodo trial will expire on %s (%d %s ago)."\
                    % (self.expirationDate, -self.daysUntilExpiration, days)

            self.licensee = None
            self.licenseUserName = None
            self.licenseType = None
            self.licenseToken = None
            self.licenseSerialNo = None
            self.licenseStatus = None

        elif self.licensingType == "license":
            proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"]\
                .getService(components.interfaces.nsIProxyObjectManager)
            _licSvc = components.classes["@activestate.com/koLicense;8"]\
                .getService(components.interfaces.koILicense)
            licSvc = proxyMgr.getProxyForObject(None,
                components.interfaces.koILicense, _licSvc, PROXY_SYNC)
            
            featureNames = eval(info["licenseFeaturesStr_ForPython"])
            rv, featureName, feature = licSvc.checkEx(featureNames,
                releaseYear, releaseMonth, releaseDay)
            self.licenseStatus = rv
            self.licenseValid = self.licenseStatus == licSvc.LIC_VALID
            if feature:
                self.licensee = licSvc.fetchValue(feature, "Licensee")
                self.licenseUserName = licSvc.fetchValue(feature, "UserName")
                self.licenseType = licSvc.fetchValue(feature, "LicenseType")
                self.licenseToken = licSvc.fetchValue(feature, "Token")
                self.licenseSerialNo = licSvc.fetchValue(feature, "SerialNo")
                self.daysUntilExpiration = licSvc.daysUntil(feature,
                                                            "ExpirationDate")
                if self.daysUntilExpiration == 9999:
                    self.expires = 0
                    self.expirationDate = None
                else:
                    self.expires = 1
                    expirationDate = licSvc.fetchValue(feature,
                                                       "ExpirationDate")
                    self.expirationDate = time.strftime("%a %b %d, %Y",
                        strptime.strptime(expirationDate, "%d-%m-%Y"))
            else:
                self.expires = None
                self.daysUntilExpiration = None
                self.expirationDate = None
                self.licensee = None
                self.licenseUserName = None
                self.licenseType = None
                self.licenseType = None
                self.licenseToken = None

            # Make a nice description of the license status for the user.
            if self.daysUntilExpiration and abs(self.daysUntilExpiration) == 1:
                days = "day"
            else:
                days = "days"
            if self.licenseStatus == licSvc.LIC_VALID:
                if self.expires:
                    self.licenseInfo = "Licensed to %s (%s license). Expires on %s (%d %s left)."\
                        % (self.licensee, self.licenseType, self.expirationDate,
                           self.daysUntilExpiration, days)
                else:
                    self.licenseInfo = "Licensed to %s (%s license)."\
                        % (self.licensee, self.licenseType)
            elif self.licenseStatus == licSvc.LIC_EXPIRED:
                self.licenseInfo = "License has expired. Licensed to %s (%s license). Expired on %s (%d %s ago)."\
                    % (self.licensee, self.licenseType, self.expirationDate,
                       -self.daysUntilExpiration, days)
            elif self.licenseStatus == licSvc.LIC_NOTCOVERED:
                self.licenseInfo = "Your license does not cover this version of Komodo."
            elif self.licenseStatus == licSvc.LIC_NOFEATURE:
                self.licenseInfo = "Your ActiveState license does not include a Komodo license."
            elif self.licenseStatus == licSvc.LIC_NOFILE:
                self.licenseInfo = "No ActiveState license file could be found."
        else:
            logError("Unexpected licensing type: %r" % self.licensingType)
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)

    def windows_GetVersionEx(self):
        import win32api
        return win32api.GetVersionEx()


if __name__ == "__main__":
    info = components.classes['@activestate.com/koInfoService;1'].\
        getService(components.interfaces.koIInfoService)
    print "platform: %r" % info.platform
    if info.isWindows:
        print "VER_PLATFORM_WIN32_WINDOWS: %r" % info.VER_PLATFORM_WIN32_WINDOWS
    print "osSystem: %r" % info.osSystem
    print "osRelease: %r" % info.osRelease
    print "osVersion: %r" % info.osVersion
    print "version: %r" % info.version
    print "buildNumber: %r" % info.buildNumber
    print "buildASCTime: %r" % info.buildASCTime
    print "mozBinDir: %r" % info.mozBinDir
    print "buildType: %r" % info.buildType
    print "installerType: %r" % info.installerType
    print "productType: %r" % info.productType
    print "licensingType: %r" % info.licensingType
    print "licenseValid: %r" % info.licenseValid
    print "licenseInfo: %r" % info.licenseInfo
    print "expires: %r" % info.expires
    print "daysUntilExpiration: %r" % info.daysUntilExpiration
    print "expirationDate: %r" % info.expirationDate
    print "licensee: %r" % info.licensee
    print "licenseUserName: %r" % info.licenseUserName
    print "licenseType: %r" % info.licenseType
    print "licenseStatus: %r" % info.licenseStatus
