#!python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

# Initialization Services for Komodo
#
# XXX The WebServices init stuff should just be part of a koWSService.init()
#     or the like. This module is for init stuff that must happen very early.

import sys
import os

from xpcom import components, nsError, ServerException, COMException
from koLogger import logError, logWarning, logDebug

import locale, logging

	
#---- support routines

class LoggerWithNotify(logging.Logger):
    
    def __init__(self, *args, **kw):
        logging.Logger.__init__(self, *args, **kw)
        try:
            self._observerSvc = components.classes["@mozilla.org/observer-service;1"]\
                        .getService(components.interfaces.nsIObserverService)
        except:
            pass # if the getService call fails

    def setLevel(self, level):
        logging.Logger.setLevel(self, level)
        if hasattr(self, '_observerSvc'):
            try:
                self._observerSvc.Notify(None, 'set_import_timing_logger_level', "%s:%d" % (self.name, level))
            except COMException: # if no one's noticing, we get an exception. XXX what exception?
                pass

logging.setLoggerClass(LoggerWithNotify)
import_timing_logger = logging.getLogger("perf.imports")
#import_timing_logger.setLevel(logging.INFO)

# The following imports, two globals and two methods are used by the import
# timing stuff, which is enabled by uncommenting a line in the initialize()
# call of koInitService
import __builtin__, time
_old_import = __builtin__.__import__
_imported_modules = {}
def _timed_import(name, *args, **kw):
    t1 = time.clock()
    mod = _old_import(name, *args, **kw)
    dt = time.clock() - t1
    realname = mod.__name__
    if not _imported_modules.has_key(name):
        _imported_modules[name] = dt
        import_timing_logger.info("%s\t%4.4f" % (realname, dt))
    else:
        _imported_modules[name] += dt
    return mod

module_tree = {}
current_module = module_tree
def _memory_import(name, *args, **kw):
    global current_module
    start = int(my_process_memory())
    current_module[name] = {}
    orig = current_module
    current_module = current_module[name]
    mod = _old_import(name, *args, **kw)
    end = int(my_process_memory())
    realname = mod.__name__
    if name.startswith(realname):
        realname = name
    mem = end-start
    if mem:
        import_timing_logger.info("%s\t%s" % (realname, memoryrepr(mem)))
        current_module['memory'] = memoryrepr(mem)
    current_module = orig
    if not current_module[name]:
        del current_module[name]
    return mod

from progressbar import progressStep
def _progress_bar_import(name, *args, **kw):
    if name not in sys.modules:
        progressStep("Loading "+name+'...', 1)
    t1 = time.clock()
    mod = _old_import(name, *args, **kw)
    dt = time.clock() - t1
    realname = mod.__name__
    if not _imported_modules.has_key(name):
        _imported_modules[name] = dt
        import_timing_logger.info("%s\t%4.4f" % (realname, dt))
    return mod


#---- Show a startup progress bar.
__builtin__.__import__ = _progress_bar_import



#----- interal support routines

def _isdir(dirname):
    """os.path.isdir() doesn't work for UNC mount points. Fake it.
    
    # For an existing mount point (want: _isdir() == 1)
    os.path.ismount(r"\\crimper\apps") -> 1
    os.path.exists(r"\\crimper\apps") -> 0
    os.path.isdir(r"\\crimper\apps") -> 0
    os.listdir(r"\\crimper\apps") -> [...contents...]
    # For a non-existant mount point (want: _isdir() == 0)
    os.path.ismount(r"\\crimper\foo") -> 1
    os.path.exists(r"\\crimper\foo") -> 0
    os.path.isdir(r"\\crimper\foo") -> 0
    os.listdir(r"\\crimper\foo") -> WindowsError
    # For an existing dir under a mount point (want: _isdir() == 1)
    os.path.mount(r"\\crimper\apps\Komodo") -> 0
    os.path.exists(r"\\crimper\apps\Komodo") -> 1
    os.path.isdir(r"\\crimper\apps\Komodo") -> 1
    os.listdir(r"\\crimper\apps\Komodo") -> [...contents...]
    # For a non-existant dir/file under a mount point (want: _isdir() == 0)
    os.path.ismount(r"\\crimper\apps\foo") -> 0
    os.path.exists(r"\\crimper\apps\foo") -> 0
    os.path.isdir(r"\\crimper\apps\foo") -> 0
    os.listdir(r"\\crimper\apps\foo") -> []  # as if empty contents
    # For an existing file under a mount point (want: _isdir() == 0)
    os.path.ismount(r"\\crimper\apps\Komodo\latest.komodo-devel.txt") -> 0
    os.path.exists(r"\\crimper\apps\Komodo\latest.komodo-devel.txt") -> 1
    os.path.isdir(r"\\crimper\apps\Komodo\latest.komodo-devel.txt") -> 0
    os.listdir(r"\\crimper\apps\Komodo\latest.komodo-devel.txt") -> WindowsError
    """
    if sys.platform[:3] == 'win' and dirname[:2] == r'\\':
        if os.path.exists(dirname):
            return os.path.isdir(dirname)
        try:
            os.listdir(dirname)
        except WindowsError:
            return 0
        else:
            return os.path.ismount(dirname)
    else:
        return os.path.isdir(dirname)


def _mkdir(newdir):
    """works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
    """
    if _isdir(newdir):
        pass
    elif os.path.isfile(newdir):
        raise OSError("a file with the same name as the desired " \
                      "dir, '%s', already exists." % newdir)
    else:
        head, tail = os.path.split(newdir)
        if head and not _isdir(head):
            _mkdir(head)
        #print "_mkdir %s" % repr(newdir)
        if tail:
            os.mkdir(newdir)


def _copy(src, dst):
    """works the way a good copy should :)
        - no source, raise an exception
        - destination directory, make a file in that dir named after src
        - source directory, recursively copy the directory to the destination
        - filename wildcarding allowed
    NOTE:
        - This copy CHANGES THE FILE ATTRIBUTES.
    """
    import string, glob, shutil

    assert src != dst, "You are try to copy a file to itself. Bad you! "\
                       "src='%s' dst='%s'" % (src, dst)
    # determine if filename wildcarding is being used
    # (only raise error if non-wildcarded source file does not exist)
    if string.find(src, '*') != -1 or \
       string.find(src, '?') != -1 or \
       string.find(src, '[') != -1:
        usingWildcards = 1
        srcFiles = glob.glob(src)
    else:
        usingWildcards = 0
        srcFiles = [src]

    for srcFile in srcFiles:
        if os.path.isfile(srcFile):
            if usingWildcards:
                srcFileHead, srcFileTail = os.path.split(srcFile)
                srcHead, srcTail = os.path.split(src)
                dstHead, dstTail = os.path.split(dst)
                if dstTail == srcTail:
                    dstFile = os.path.join(dstHead, srcFileTail)
                else:
                    dstFile = os.path.join(dst, srcFileTail)
            else:
                dstFile = dst
            dstFileHead, dstFileTail = os.path.split(dstFile)
            if dstFileHead and not _isdir(dstFileHead):
                _mkdir(dstFileHead)
            if _isdir(dstFile):
                dstFile = os.path.join(dstFile, os.path.basename(srcFile))
            #print "copy %s %s" % (srcFile, dstFile)
            if os.path.isfile(dstFile):
                # make sure 'dstFile' is writeable
                os.chmod(dstFile, 0755)
            shutil.copy(srcFile, dstFile)
            # make the new 'dstFile' writeable
            os.chmod(dstFile, 0755)
        elif _isdir(srcFile):
            srcFiles = os.listdir(srcFile)
            if not os.path.exists(dst):
                _mkdir(dst)
            for f in srcFiles:
                s = os.path.join(srcFile, f)
                d = os.path.join(dst, f)
                try:
                    _copy(s, d)
                except (IOError, os.error), why:
                    raise OSError("Can't copy %s to %s: %s"\
                          % (repr(s), repr(d), str(why)))
        elif not usingWildcards:
            raise OSError("Source file %s does not exist" % repr(srcFile))



#---- main component implementation

class KoInitService:
    _com_interfaces_ = [components.interfaces.koIInitService]
    _reg_clsid_ = "{BAECC764-52AC-46dc-9428-D23F05247818}"
    _reg_contractid_ = "@activestate.com/koInitService;1"
    _reg_desc_ = "Komodo Init Service"

    def __init__(self):

        #---- Special (read: hacky) import handling of win32comexts.
        # We already move pythoncomXX.dll and PyWinTypes.dll to the
        # site-packages/win32 directory to get them out of the system
        # directory where they could conflict with existing ones from
        # non-siloed Python's.  The problem is that the
        # site-packages/win32comext tree also contains binary PyWin32
        # Python extensions that need to find these DLLs.  The chosen
        # solution is to (1) move the win32comext .pyd's that we need to
        # the win32 directory, (2) pre-import them, and (3) muck around
        # with Python's import machinery to make it look like their
        # location was not changed. This cannot be done at module
        # top-level because this fails when regxpcom is run during the
        # _build_ of the installer.
        if sys.platform.startswith("win"):
            import shell
            sys.modules["win32com.shell.shell"] = sys.modules["shell"]
            import win32com.shell
            win32com.shell.__dict__['shell'] = shell
            del sys.modules["shell"]
            del shell


        self._isInitialized = 0
        koDirs = components.classes["@activestate.com/koDirs;1"].getService(components.interfaces.koIDirs)
        logDir = os.path.join(koDirs.userDataDir, 'logs')
        if not os.path.exists(logDir):
            os.mkdir(logDir)
        hdlr = logging.FileHandler(os.path.join(logDir, 'status.log'))
        fmt = logging.Formatter("%(levelname)s: %(name)s: %(message)s")
        hdlr.setFormatter(fmt)
        logging.getLogger('status').addHandler(hdlr)

    def setPlatformErrorMode(self):
        if sys.platform.startswith("win"):
            try:
                # Stop Windows displaying a dialog when removable
                # media is accessed, but not available
                # (eg, when a file from a floppy is on the MRU,
                # but no floppy is in the drive)
                SEM_FAILCRITICALERRORS = 1 # constant pulled from win32con to save memory
                import win32api
                win32api.SetErrorMode(SEM_FAILCRITICALERRORS)
            except ImportError:
                logError("Could not import win32api/win32con to set the "\
                         "Win32 Error Mode.")

    def _setencoding(self,encoding=None):
        try:
            if not encoding:
                encoding = locale.getdefaultlocale()
            if encoding is not None:
                reload(sys)
                sys.setdefaultencoding(encoding[1])
                del sys.setdefaultencoding
                return encoding
        except ValueError, e:
            # XXX Also getting this error occassionaly (as reported on
            # Komodo-beta). Should handle that.
            #    LookupError: unknown encoding
            if encoding:
                logError('Unsupported encoding for locale %s.%s'%\
                         (encoding[0],encoding[1]))
            else:
                logError('Unable to set encoding for locale %s'%\
                         os.getenv('LANG'))
        except LookupError, e:
            # windows sets codepages into the locale without
            # the leading 'cp' that python wants.  If the encoding
            # we received is a number, try cp# and see if it works.
            # eg. 1252 should be turned into cp1252
            try:
                isint = int(encoding[1])
                if sys.platform.startswith("win") and isint:
                    encoding[1] = 'cp'+encoding[1]
                    return self._setencoding(encoding)
            except ValueError,e:
                # not an int
                pass
            logDebug('Unsupported encoding for locale %s.%s'%\
                     (encoding[0],encoding[1]))
        except TypeError, e:
            if not sys.platform.startswith("win") and os.getenv("LANG") == "C":
                return ("en_US","iso8859-1")
            else:
                logDebug('Unable to set encoding for locale %s'%\
                            os.getenv('LANG'))
        except Exception, e:
            logDebug('Unable to determine system locale settings')
        return None
    
    def _setlocale(self):
        try:
            locale.setlocale(locale.LC_ALL,'')
        except:
            logDebug('Unable to set encoding for locale %s'%os.getenv('LANG'))
            return None
        return self._setencoding()

    def _removeEuroLocale(self):
        envvars = ['LC_ALL','LC_CTYPE','LANG','LANGUAGE']
        for var in envvars:
            orig_lang = os.getenv(var)
            if orig_lang and orig_lang.lower().endswith('@euro'):
                lang = orig_lang[:orig_lang.find('@euro')]
                os.environ[var] = lang
        
    def setEncoding(self):
        self._startup_locale = self._setlocale()
        if sys.platform.startswith("win"):
            # XXX - this is evil and should be dropped ASAP.
            # Changing the default encoding could impact other code.
            # Python 2.2 will have this support for all file system functions,
            # so as soon as that version is in Komodo, this MUST DIE.

            # On Windows, default python <=2.1 chokes on Unicode file
            # names.  If we change the encoding to "Latin-1", we can
            # at least support European chars transparently.
            # (and if we use "mbcs" we should be able to use
            # _any_ language supported by Windows, latin or otherwise.)
            reload(sys)
            sys.setdefaultencoding("mbcs")
            del sys.setdefaultencoding
            logDebug("encoding set to '%s'" % sys.getdefaultencoding())

        else: # Platforms other than Windows
            # Set the default encoding to that appropriate for the
            # current locale, except with UTF-8 support
            if not self._startup_locale:
                # XXX In python 2.2, setlocale fails if
                # the LANG envt variable has @euro at the end.
                # It's a bug in Python 2.2
                self._removeEuroLocale()
                self._startup_locale = self._setlocale()
            if not self._startup_locale:
                logDebug("Unable to determine the current locale "+\
                            "settings, please set LANG environment "+\
                            "variable. Defaulting to en_US.iso8859-1")
                # set the locale to something that we can
                # least let komodo startup and run with
                self._startup_locale = ['en_US','iso8859-1']
            elif self._startup_locale[1].lower() == 'utf':
                # RH 8 likes to use 'utf' rather than UTF-8,
                # but it will accept UTF-8
                self._startup_locale[1]='UTF-8'
            new_locale = self._startup_locale[0]+'.UTF-8'
            
            # We have to set pythons encoding to utf-8 so that
            # we can have some hope of interpreting file paths
            # that are written in different encodings, and to
            # make GTK show UTF8 fonts correctly.  for instance
            # saving a file with a latin-1 filename, but starting
            # komodo with encoding ISO8859-1.  If we do not do
            # this, and a latin-1 filename is in mru, komodo fails
            # to load
            
            # XXX this will break on older unix systems that do not
            # support UTF-8 -- it would be nice to detect this and
            # let the user know.
            try:
                # this bit is for file I/O
                reload(sys)
                sys.setdefaultencoding('utf-8')
                del sys.setdefaultencoding
            except Exception,e:
                logWarning('cannot set encoding to utf-8: %s'%e)
            
            # this bit is the GTK important bit
            # if it's not done, GTK won't display correctly
            try:
                locale.setlocale(locale.LC_CTYPE,new_locale)
            except Exception,e:
                logWarning("cannot reset LOCALE set to %s : %s" % (new_locale,e))

    def getStartupLocale(self):
        return "%s.%s" % \
                (self._startup_locale[0],self._startup_locale[1])

    def initProcessUtils(self):
        import koprocessutils
        koprocessutils.initialize()

    def finishInitialization(self):
        pass

    def upgradeUserSettings(self):
        # Get the current version.
        infoSvc = components.classes["@activestate.com/koInfoService;1"].getService()
        currVer = infoSvc.version.split(".")[:2]
        for i in range(len(currVer)):
            try:
                currVer[i] = int(currVer[i])
            except ValueError:
                pass
        currVer = tuple(currVer) # e.g. (2,3)
        #print "XXX upgradeUserSettings: currVer=%r" % (currVer,)

        # Get a mapping of the versions for which there are user data
        # directories to those directories.
        koDirsSvc = components.classes["@activestate.com/koDirs;1"].getService()
        existingDataDirs = {} # mapping of ver to user data directory
        for userDataDir in koDirsSvc.GetAllUserDataDirs():
            existingVer = os.path.basename(userDataDir).split(".")
            for i in range(len(existingVer)):
                try:
                    existingVer[i] = int(existingVer[i])
                except ValueError:
                    pass
            existingDataDirs[tuple(existingVer)] = userDataDir
        #print "XXX upgradeUserSettings: existingDataDirs=%r"\
        #      % (existingDataDirs,)
            
        # Files and dir in app data dir to upgrade on first run.
        # XXX Or maybe this should be defined as "everything except ..."
        filesToUpgrade = [
            "komodo-user-prefs.xml",
            "komodo-style-prefs.xml",
            "templates",
            "toolbox.kpf",
            "view-state.xmlc", #XXX This is trouble if we upgrade siloed Python.
            "webservices"
        ]

        # If any of the above exist in the current version's user data
        # dir then do NOT upgrade.
        try:
            currDataDir = existingDataDirs[currVer]
        except KeyError, ex:
            # XXX The current version should always exist in the
            #     "existingDataDirs" because komodo.exe will have
            #     created it on startup. However, because of a still not
            #     understood bug with dev builds on Linux,
            #     .GetAllUserDataDirs() will still return the empty list
            #     (problem with os.listdir()). Handle that case here.
            logDebug("NOT upgrading user settings because we hit the "\
                     "os.listdir()+Linux+devbuild bug.")
            return
        for file in os.listdir(currDataDir):
            if file in filesToUpgrade:
                logDebug("NOT upgrading user settings because '%s' already "\
                         "exists for the current version: %s"\
                         % (file, os.path.join(currDataDir, file)))
                return

        # XXX:HACK Workaround for bug 23150:
        # Because of bug 23046 (Should have new Moz build for _Linux_
        # releases of Komodo), Komodo 2.3 startup will create
        # ~/.komodo/2.0/... which will trick this logic into thinking
        # that there are valid 2.0 files to upgrade from. This _may_ not
        # be true. The workaround is to remove the 2.0 entry from
        # "existingDataDirs" if not.
        if not sys.platform.startswith("win") \
           and existingDataDirs.has_key((2,0)):
            firstFileToUpgrade = os.path.join(existingDataDirs[(2,0)],
                                              filesToUpgrade[0])
            if not os.path.exists(firstFileToUpgrade):
                del existingDataDirs[(2,0)]

        # Determine the most recent version with user settings, if any.
        existingVers = existingDataDirs.keys()
        existingVers.sort()
        currIndex = existingVers.index(currVer)
        if currIndex == 0:
            logDebug("NOT upgrading user settings because there is no "\
                     "previous version: %s" % existingVers)
            return 
        prevVer = existingVers[currIndex-1]
        
        # Upgrade.
        logDebug("upgrading user settings from version '%s'" % (prevVer,))
        for file in filesToUpgrade:
            src = os.path.join(existingDataDirs[prevVer], file)
            dst = os.path.join(existingDataDirs[currVer], file)
            try:
                _copy(src, dst)
            except OSError, ex:
                logError("Could not upgrade '%s' from '%s': %s"\
                         % (dst, src, ex))


