from xpcom import components, ServerException, COMException, nsError
from koLogger import logError, logDebug, logWarning
from invocationutils import *
import sys, os, string

INVOKE_RUN = components.interfaces.koIInvocation.INVOKE_RUN
INVOKE_DEBUG = components.interfaces.koIInvocation.INVOKE_DEBUG
INVOKE_BUILD = components.interfaces.koIInvocation.INVOKE_BUILD

class koCommandInvocation:
    _invocation_name_ = "Command"
    _invocation_description_ = "Batch file, shell script or other program"
    _invocation_chrome_category_ = "chrome://komodo/content/pref/invocations/categories/command.xul"
    _invocation_chrome_instance_ = "chrome://komodo/content/pref/invocations/instances/command.xul"

    _reg_registrar_ = RegisterAsInvocation, None
    _reg_clsid_ = "{67fdd4be-725e-43b0-8fa4-4092fc60cfc8}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=" + _invocation_name_
    _com_interfaces_ = components.interfaces.koIInvocation
    def __init__(self):
        self.currentCategoryPrefs = None
        self.currentInstancePrefs = None
        self.name = self._invocation_name_
        self.process_spawner = components.classes["@activestate.com/koProcessRunner;1"] \
                          .createInstance (components.interfaces.koIProcess)
        self.stdin = self.stdout = self.stderr = None
        self.infoSvc = components.classes["@activestate.com/koInfoService;1"] \
            .getService(components.interfaces.koIInfoService);
        self.komodoVersion = self.infoSvc.version;

    def validate(self, invokeType):
        # XXX - todo - work out what "debug" means in this context!?
        if self.currentInstancePrefs is None or self.currentCategoryPrefs is None:
            raise ServerException(nsError.NS_ERROR_NOT_INITIALIZED)
        prefset = self.currentInstancePrefs
        if not prefset.getStringPref("filename"):
            return 0, "command", "You must specify the command to execute"

        cwd = prefset.getStringPref("cwd")
        if cwd and not os.path.isdir(cwd):
            return 0, "cwd", "Directory '%s' does not exist" % cwd
        return 1, None, None

    def get_supportedInvokeTypes(self):
        return INVOKE_RUN # | INVOKE_DEBUG

    def _checkValid(self, invokeType):
        ok, id, message = self.validate(invokeType)
        if not ok:
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, message)

    def needsBuild( self ):
        # Result: boolean
        # we never need a separate build process.
        return 0

    def invoke(self, invokeType):
        self._checkValid(invokeType)
        return self._launch()

    def terminate(self, exit_code, use_force):
        if self.process_spawner:
            try:
                self.process_spawner.killProcess();
                return 1
            except COMException:
                pass
        return 0

    def getDefaultPreferences( self ):
        cat = components.classes["@activestate.com/koPreferenceSet;1"].createInstance()
        cat.setBooleanPref("use-console", 0)
        cat.setStringPref("debugger.language", "")

        inst = components.classes["@activestate.com/koPreferenceSet;1"].createInstance()
        inst.setStringPref("filename", "")
        inst.setStringPref("params", "")
        inst.setStringPref("cwd", "")
        inst.setBooleanPref("show-dialog", 1)

        inst.parent = cat
        return cat, inst

    def get_currentCategoryPreferences( self ):
        # Result: koIPreferenceSet
        return self.currentCategoryPrefs

    def set_currentCategoryPreferences( self, prefs):
        # Result: void - None
        # In: param0: koIPreferenceSet
        self.currentCategoryPrefs = prefs

    def get_currentInstancePreferences( self ):
        # Result: koIPreferenceSet
        return self.currentInstancePrefs

    def set_currentInstancePreferences( self, prefs):
        # Result: void - None
        # In: param0: koIPreferenceSet
        self.currentInstancePrefs = prefs

    def get_persistableInstancePreferences( self ):
        # Result: koIPreferenceSet
        return self.currentInstancePrefs.clone()

    def _launch(self):
        process_spawner = self.process_spawner

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        showConsole = self.currentInstancePrefs.getBooleanPref('use-console')
        args = [self.currentInstancePrefs.getStringPref('filename')]
        args += components.classes["@activestate.com/CommandLineParser;1"] \
                     .createInstance(components.interfaces.asICommandLineParser) \
                     .parse( self.currentInstancePrefs.getStringPref('params') );
        if sys.platform.startswith("linux"):
            # XXX unicode issue
            shellstring = str(string.join(args, " "))
            args = ["/bin/sh", "-c" , shellstring]
        args.append(None)

        if self.currentInstancePrefs.hasStringPref('systemEnvironment'):
            environmentStrings = self.currentInstancePrefs.getStringPref('systemEnvironment')
            environmentStrings = environmentStrings.strip().split("\n")
        else:        
            environmentStrings = components.classes["@activestate.com/koUserEnviron;1"].getService().GetEnvironmentStrings()
        # XXX:HACK - KOMODO_VERSION was made required so some PDK invocations
        # could decide how to present their version output for Komodo's PDK
        # integration would proerply determine the PDK version. This is
        # an *abomination*.
        environmentStrings.append("KOMODO_VERSION=%s" % self.komodoVersion)
        
        pio = ProcessIOHandler(self.stdin, self.stdout, self.stderr,
                               self.process_spawner)
        process_spawner.init(pio)

        if cwd:
            process_spawner.cwd = cwd
        try:
            if showConsole:
                process_spawner.startProcessInConsole(args,
                    environmentStrings)
            else:
                process_spawner.startProcess(args,
                    environmentStrings)
        except COMException:
            msg = "Executing '%s' failed: %s\n" % (args, self.process_spawner.errorText)
            logWarning(msg)
            # if for some reason the spawner has caused an exception, be sure we
            # kill stdin thread (which blocks for a write), and signal the termination listeners
            if self.stdin: self.stdin.write(None)
            pio.signalProcessEnd(-1)
            raise ServerException(nsError.NS_ERROR_FILE_EXECUTION_FAILED, msg)

