#!python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import sys
import os
import string
import re
import urlparse, mimetools, mimetypes
import threading
import tempfile

import process
import koprocessutils
from invocationutils import *

from xpcom import components, ServerException, COMException, nsError
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from koLogger import logError, logDebug, logWarning



#---- globals

INVOKE_RUN = components.interfaces.koIInvocation.INVOKE_RUN
INVOKE_DEBUG = components.interfaces.koIInvocation.INVOKE_DEBUG
INVOKE_BUILD = components.interfaces.koIInvocation.INVOKE_BUILD



#---- support routines

def _escapeArg(arg):
    """Escape the given command line argument for the shell."""
    return arg.replace('"', r'\"')

def _joinArgv(argv):
    r"""Join an arglist to a string appropriate for running.

        >>> import os
        >>> _joinArgv(['foo', 'bar "baz'])
        'foo "bar \\"baz"'
    """
    cmdstr = ""
    for arg in argv:
        if ' ' in arg or ';' in arg:
            cmdstr += '"%s"' % _escapeArg(arg)
        else:
            cmdstr += _escapeArg(arg)
        cmdstr += ' '
    if cmdstr.endswith(' '): cmdstr = cmdstr[:-1]  # strip trailing space
    return cmdstr



#---- Component implementations

class koInterpreterInvocation:
    _com_interfaces_ = components.interfaces.koIInvocation
    def __init__(self):
        self.currentCategoryPrefs = None
        self.currentInstancePrefs = None
        self.prefService = components.classes["@activestate.com/koPrefService;1"].getService();
        self.name = self._invocation_name_
        self.process_spawner = components.classes["@activestate.com/koProcessRunner;1"] \
                  .createInstance (components.interfaces.koIProcess)
        self._environUtils = components.classes[
            "@activestate.com/koEnvironUtils;1"] \
            .getService(components.interfaces.koIEnvironUtils)
        self.infoSvc = components.classes["@activestate.com/koInfoService;1"] \
            .getService(components.interfaces.koIInfoService);
        self.komodoVersion = self.infoSvc.version\

        self.stdin = self.stdout = self.stderr = None
        self._postData = None

    def needsBuild( self ):
        # Result: boolean
        # we never need a separate build process.
        return 0

    def invoke(self, invokeType):
        self._checkValid(invokeType)
        if invokeType == INVOKE_DEBUG:
            self._launchForDebug()
        elif invokeType == INVOKE_RUN:
            return self._launchForRun()
        else:
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE)

    def get_supportedInvokeTypes(self):
        return INVOKE_DEBUG | INVOKE_RUN

    def validate(self, invokeType):
        if self.currentInstancePrefs is None or self.currentCategoryPrefs is None:
            return 0, "", "Invocation preferences not initialized"

        # Attempt to locate the executable if none specified.
        if not self.currentCategoryPrefs.getStringPref("executable"):
            self.currentCategoryPrefs.setStringPref("executable", self._determineExecutable())

        if not (invokeType & self.get_supportedInvokeTypes()):
            return 0, "", "Invocation does not support requested invocation type"            

        return self._do_validate(invokeType)

    def _checkValid(self, invokeType):
        ok, id, message = self.validate(invokeType)
        if not ok:
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, message)

    def terminate(self, exit_code, use_force):
        try:
            self.process_spawner.killProcess();
            return 1
        except COMException:
            pass
        return 0

    def _fillExtraDefaultPreferences(self, cat, inst):
        pass
        
    def getDefaultPreferences( self ):
        cat = components.classes["@activestate.com/koPreferenceSet;1"].createInstance()
        cat.setStringPref("executable", "") # validate fills this correctly if not done before.
        cat.setStringPref("executable-params", "")
        cat.setBooleanPref("use-console", 0)
        cat.setLongPref("debugger.listener-port", self.prefService.prefs.getLongPref("remoteDebuggerListenPort"))

        inst = components.classes["@activestate.com/koPreferenceSet;1"].createInstance()
        inst.setStringPref("filename", "")
        inst.setStringPref("params", "")
        inst.setStringPref("cwd", "")
        inst.setBooleanPref("show-dialog", 1)

        inst.parent = cat
        self._getSystemEnvironment(cat, inst)
        self._getDefaultCGIPrefs(cat, inst)
        self._fillExtraDefaultPreferences(cat, inst)

        return cat, inst

    def get_currentCategoryPreferences( self ):
        # Result: koIPreferenceSet
        return self.currentCategoryPrefs

    def set_currentCategoryPreferences( self, prefs):
        # Result: void - None
        # In: param0: koIPreferenceSet
        self.currentCategoryPrefs = prefs

    def get_currentInstancePreferences( self ):
        # Result: koIPreferenceSet
        return self.currentInstancePrefs

    def set_currentInstancePreferences( self, prefs):
        # Result: void - None
        # In: param0: koIPreferenceSet
        self.currentInstancePrefs = prefs

    def get_persistableInstancePreferences( self ):
        # Result: koIPreferenceSet
        ret = self.currentInstancePrefs.clone()
        for prefName in ["CGIEnvironment", "systemEnvironment"]:
            if ret.hasPref(prefName):
                ret.deletePref(prefName)
        return ret

    def _doLaunch(self, real_args, cwd, environmentStrings):
        showConsole = self.currentInstancePrefs.getBooleanPref('use-console')
        process_spawner = self.process_spawner

        #print "Launching with args", real_args, " in ", cwd
        
        ioHandler = ProcessIOHandler(self.stdin, self.stdout, self.stderr,
                                     self.process_spawner)
        process_spawner.init(ioHandler)
        process_spawner.cwd = cwd
        try:
            if showConsole:
                process_spawner.startProcessInConsole(real_args,
                    environmentStrings)
            else:
                process_spawner.startProcess(real_args,
                    environmentStrings)

            # send post data if any exists
            if self._postData and self.stdin:
                #print "Sending %d bytes of data to engine\n", len(self._postData)
                self.stdin.write(self._postData)

        except COMException:
            msg = "Executing '%s' failed: %s\n" % (real_args, self.process_spawner.errorText)
            logWarning(msg)
            raise ServerException(nsError.NS_ERROR_FILE_EXECUTION_FAILED, msg)

    def _setPrefEnvironment(self, inst, name, newname = None):
        if not newname: newname = name
        penv = self.prefService.prefs.getStringPref(name)
        # if we have a predefined user environment, sort it first
        if len(penv) > 0:
            penv = string.split(penv, os.linesep)
            penv.sort()
            penv = string.join(penv,os.linesep)
        inst.setStringPref(newname,penv)
        
    def _getSystemEnvironment(self, cat, inst):
        environment = components.classes["@activestate.com/koUserEnviron;1"].getService().GetEnvironmentStrings()
        environment.sort()
        env = string.join(environment,"\n")
        cat.setStringPref("systemEnvironment",env)
        self._setPrefEnvironment(inst, "userEnvironment")
        
    # CGI emulation support functions
    def _getDefaultCGIPrefs(self, cat, inst):
        inst.setBooleanPref("sim-cgi",0)

        # get the webserver/browser env vars
        self._setPrefEnvironment(cat, "debuggerEnvironment", "CGIEnvironment")
        self._setPrefEnvironment(inst, "userCGIEnvironment")
        
        inst.setStringPref("documentRoot","")
        
        inst.setStringPref("getparams","")
        inst.setStringPref("postparams","")
        inst.setStringPref("posttype","application/x-www-form-urlencoded")
        inst.setStringPref("mpostparams","")
        inst.setStringPref("cookieparams","")
        inst.setStringPref("request-method","GET")
        
    def _getenv(self):
        """Return the merged list of environment strings.
        The merge order is:
            1. the system environment
            2. the default CGI environment
            3. the user environment settings
            4. the user CGI environment settings
        This tries to ensure, as best as possible, that the user's
        settings are used as expected.
        """
        # 1. the system environment
        mergedEnvStrs = self.currentInstancePrefs.getStringPref('systemEnvironment')
        mergedEnvStrs = mergedEnvStrs.strip().split("\n")

        # 2. the default CGI environment
        if self._usecgi:
            cgiEnvStrs = self.currentInstancePrefs.getStringPref('CGIEnvironment')
            cgiEnvStrs = cgiEnvStrs.strip().split("\n")
            mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                mergedEnvStrs, cgiEnvStrs)
            
        # 3. the user environment settings
        userEnvStrs = self.currentInstancePrefs.getStringPref('userEnvironment')
        if userEnvStrs:
            userEnvStrs = userEnvStrs.strip().split("\n")
            mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                mergedEnvStrs, userEnvStrs)
        
        # 4. the user CGI environment settings
        if self._usecgi:
            userCgiEnvStrs = self.currentInstancePrefs.getStringPref('userCGIEnvironment')
            if userCgiEnvStrs:
                userCgiEnvStrs = userCgiEnvStrs.strip().split("\n")
                mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                    mergedEnvStrs, userCgiEnvStrs)
            
            # add generated cgi env vars (from GET, POST, and Cookie params)
            httpEnvStrs = ["%s=%s" % item
                           for item in self._build_http_env().items()]
            mergedEnvStrs = self._environUtils.MergeEnvironmentStrings(
                mergedEnvStrs, httpEnvStrs)

        # Invariant:
        #   The MergeEnvironmentStrings() functionality ensures that
        #   there are no env. vars with an empty value.
        # Convert items to plain strings (no Unicode)
        mergedEnvStrs = [str(item) for item in mergedEnvStrs]
        # add komodo version
        mergedEnvStrs.append("KOMODO_VERSION=%s" % self.komodoVersion)
        return mergedEnvStrs
    
    def _filename_from_url(self, url):
        fn = re.match(r'.*[\/\\]([^\/\\]*)$',url);
        if fn:
            return fn.group(1)
        return ""
        
    def _build_http_env(self):
        self._postData, self._postSize, self._postType = self._build_post_data()
        requestMethod = self.currentInstancePrefs.getStringPref('request-method')
        
        # calculate the document root
        pathTranslated = self.currentInstancePrefs.getStringPref('filename')
        cookieData = self.currentInstancePrefs.getStringPref('cookieparams')
        cookies = self._build_args_list(cookieData, "; ")
        
        # XXX have to do something MUCH better for the doc root!
        scriptName = "/" + self._filename_from_url(pathTranslated)

        env = {}
        env["QUERY_STRING"]=self._build_get_line()
        env["CONTENT_LENGTH"]=self._postSize
        if self._postSize:
            #print "Content size is ",self._postSize
            env["CONTENT_TYPE"]=self._postType
        env["PATH_INFO"]=scriptName #XXX verify this!
        env["PATH_TRANSLATED"]=pathTranslated
        env["SCRIPT_NAME"]=scriptName
        env["REQUEST_METHOD"]=requestMethod
        env["HTTP_COOKIE"]=cookies
        # build any environment variables we need to emulate the web server environment
        return env
    
    def _build_args_list(self, args, sep="&"):
        argsList = string.split(string.strip(args), "\n")
        return string.join(argsList,sep)
    
    def _build_get_line(self):
        getArgs = self.currentInstancePrefs.getStringPref('getparams')
        return self._build_args_list(getArgs)

    def _build_post_data(self):
        # build the post data that must be sent down stdin to the cgi
        # XXX extremely simplistic mime building going on here

        postArgs = self.currentInstancePrefs.getStringPref('postparams')
        postFiles = self.currentInstancePrefs.getStringPref('mpostparams')
        postType = self.currentInstancePrefs.getStringPref('posttype')
        
        if not postType: postType = "application/x-unknown"
        
        if postType == "application/x-www-form-urlencoded":
            postData = self._build_args_list(postArgs)
        elif postType == "multipart/form-data":
            # handle RFC 1867 data
            out = ""
            # we have to do multpart mime encoding
            boundary = mimetools.choose_boundary()
            # generate the mime headers for regular post variables
            if postArgs:
                argsList = string.split(postArgs, "\n")
                for i in range(len(argsList)):
                    try:
                        key, value = argsList[i].split("=", 1)
                        out += "--"+boundary+"\r\n"
                        out += "Content-Disposition: form-data; name=\"%s\"\r\n\r\n" % (key)
                        out += value+"\r\n"
                    except ValueError:
                        pass
                
            if postFiles:
                fileList = string.split(postFiles, "\n")
                # read each file, encode it, then add it to the return data
                for i in range(len(fileList)):
                    # read each file, base64 encode it, and add to the request
                    if len(fileList[i]) < 3: continue
                    fvar = string.split(fileList[i], "=")
                    fn = self._filename_from_url(fvar[1])
                    out += "--"+boundary+"\r\n"
                    out += "Content-Disposition: form-data; name=\"%s\"; filename=\"%s\"\r\n" % (fvar[0], fn)
                    type, encoding = mimetypes.guess_type(fvar[1])
                    if not type: type = "text/plain"
                    out += "Content-Type: %s\r\n" % (type)
                    out += "Content-Transfer-Encoding: binary\r\n\r\n"

                    infile = open(fvar[1], "rb")
                    out += infile.read()
                    infile.close()
                    
            # XXX - not rfc compliant, but works right, should have additional -- at end
            out += "--"+boundary+"\r\n"
            postData = out
            postType = "multipart/form-data; boundary=\"%s\"" % (boundary)
        
        else:
            # support raw post data entered into the post args field
            postData = postArgs
        return postData,len(postData),postType


class koPythonInvocation(koInterpreterInvocation):
       # If this object needs to be registered, the following 2 are also needed.
    _reg_clsid_ = "{eef242ad-c57e-4507-80d5-d2fa7885dea5}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Python"
    _reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Python"
    _invocation_description_ = "Python"
    _invocation_chrome_category_ = "chrome://komodo/content/pref/invocations/categories/python.xul"
    _invocation_chrome_instance_ = "chrome://komodo/content/pref/invocations/instances/python.xul"

    def _fillExtraDefaultPreferences(self, cat, inst):
        cat.setStringPref("PYTHONPATH", "")
        cat.setBooleanPref("unbuffered-handles", 1)

    def _determineExecutable(self):
        languageService = components.classes[\
                "@activestate.com/koLanguage?language=Python;1"].\
                getService(components.interfaces.koILanguage)

        pythonExe = self.prefService.prefs.getStringPref("pythonDefaultInterpreter")
        if not pythonExe:
            pythonExe = languageService.getLanguageService(components.interfaces.koIAppInfoEx).executablePath
        if not pythonExe:
            logError("No Python interpreter could be found.");
            return ""
        else:
            return pythonExe

    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the executable used for the debug session")
            checkFileExists(prefset, 'filename', "You must specify the main Python script to be used")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None

    def _launchForDebug(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')

        # use the get variables as the user_args if emulating CGI        
        debug_args = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: debug_args.append(get_args)
        if len(debug_args) < 1:
            debug_args = components.classes["@activestate.com/CommandLineParser;1"] \
                     .createInstance(components.interfaces.asICommandLineParser) \
                     .parse( self.currentInstancePrefs.getStringPref('params') );

        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')

        executableArgs = []
        if os.path.basename(executable).find('jython') == -1 and self.currentInstancePrefs.getBooleanPref('unbuffered-handles'):
            executableArgs.append("-u")

        kdbArgs = []
        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        kdbArgs.append("-p%s" % (port,))
        if self.currentInstancePrefs.getBooleanPref('use-console'):
            kdbArgs.append("-w")

       # The Python interpreter that is used to launch "kdb" may be a
        # a different version than is currently run, therefore we must be
        # sure to launch kdb.py rather than kdb.pyX.
        koDirs = components.classes["@activestate.com/koDirs;1"].\
                 getService(components.interfaces.koIDirs)
        kdbFile = os.path.join(koDirs.installDir, "callkomodo", "kdb.py")
        
        real_args = [executable] + executableArgs + [kdbFile] + kdbArgs + [script] + debug_args + [None]

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = self._getenv()
        return self._doLaunch(real_args, cwd, environment)

    def _launchForRun(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        script_args = components.classes["@activestate.com/CommandLineParser;1"] \
                     .createInstance(components.interfaces.asICommandLineParser) \
                     .parse( self.currentInstancePrefs.getStringPref('params') );

        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')
        executableArgs = []
        if self.currentInstancePrefs.getBooleanPref('unbuffered-handles'):
            executableArgs.append("-u")
        
        real_args = [executable] + executableArgs + [script] + script_args + [None]
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = self._getenv()        
        return self._doLaunch(real_args, cwd, environment)

class koPerlInvocation(koInterpreterInvocation):
    _reg_clsid_ = "{3becc3bd-1e42-47b4-aafd-9974d10823dd}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Perl"
    _reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Perl"
    _invocation_description_ = "Perl"
    _invocation_chrome_category_ = "chrome://komodo/content/pref/invocations/categories/perl.xul"
    _invocation_chrome_instance_ = "chrome://komodo/content/pref/invocations/instances/perl.xul"

    def __init__(self):
        koInterpreterInvocation.__init__(self)
        proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerSvc = proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService,
            observerSvc, PROXY_ALWAYS | PROXY_SYNC)
    
    def _fillExtraDefaultPreferences(self, cat, inst):
        inst.setStringPref("warnings", "enabled")
        inst.setLongPref("debugger.io-port", 9011)

    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the Perl executable used for the debug session")
            checkFileExists(prefset, 'filename', "You must specify the main Perl script to be used")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)
            warnings = prefset.getStringPref("warnings")
            valid_warning_values = ["enabled", "disabled", "all"]
            if warnings not in valid_warning_values:
                raise PrefException("warnings", "Must be one of " + ",".join(valid_warning_values))
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None
        
    def _determineExecutable(self):
        koInfoService = components.classes["@activestate.com/koInfoService;1"].getService();
        # use pref'd default if one has been chosen
        perlExe = self.prefService.prefs.getStringPref("perlDefaultInterpreter")
        if perlExe:
            # "wPerl.exe" on Windows causes problems for "Debugging in
            # Console". We want to just use "perl.exe" even if the preference
            # states to use "wPerl.exe"
            if sys.platform.startswith("win"):
                dirName, baseName = os.path.split(perlExe)
                if os.path.splitext(baseName)[0].lower() == "wperl":
                    perlExe = os.path.join(dirName, "perl.exe")
                    if not os.path.isfile(perlExe):
                        perlExe = ""
        if not perlExe:
            sysUtils = components.classes["@activestate.com/koSysUtils;1"].getService()
            perlExe = sysUtils.Which("perl")
        if not perlExe or not os.path.isfile(perlExe):
            logError("No Perl interpreter could be found.");
            return ""
        return perlExe


    def _getExtraPerlOptions(self):
        ret = []
        warnings = self.currentInstancePrefs.getStringPref('warnings')
        if warnings == "enabled":
            ret.append("-w")
        elif warnings == "all":
            ret.append("-W")
        elif warnings == "disabled":
            ret.append("-X")
        return ret

    def _WaitCleanUpAndNotify(self, child, command, scriptFileName=None,
                              inputFileName=None):
        """Thread used by _doLaunchInConsole."""
        retval = child.wait()

        if scriptFileName:
            try:
                os.unlink(scriptFileName)
            except OSError, ex:
                logWarning("Could not remove temporary script file '%s': %s"\
                           % (scriptFileName, ex))
        if inputFileName:
            try:
                os.unlink(inputFileName)
            except OSError, ex:
                logWarning("Could not remove temporary input file '%s': %s"\
                           % (inputFileName, ex))
        
        # Send appropriate notification.
        try:
            #XXX Should eventually pass exitCode out with notification
            self._observerSvc.Notify(None, "invocation_terminated", None)
        except COMException, e:
            # do nothing if nobody listening
            pass

    def _getEnvDictFromString(self, env):
        """Convert this:
            "VAR1=value1\nVAR2=value2\n..."
        to this:
            {"VAR1": "value1", "VAR2": "value2", ...}

        (NOTE: this is copied from koRunService)
        """
        envDict = {}
        for piece in env.split('\n'):
            equalSign = piece.find('=')
            envDict[piece[:equalSign]] = piece[equalSign+1:]
        return envDict

    def _mergeEnvWithParent(self, envDict):
        """Merge the given environment variables with parent process' environ.
        (NOTE: this is copied from koRunService)

        Note: koIUserEnviron is used to avoid picking up the current
              Komodo process' environment changes from getting in the way.
        """
        userEnvSvc = components.classes["@activestate.com/koUserEnviron;1"].getService()
        userEnvDict = {}
        for piece in userEnvSvc.GetEnvironmentStrings():
            equalSign = piece.find('=')
            userEnvDict[piece[:equalSign]] = piece[equalSign+1:]
        userEnvDict.update(envDict)
        return userEnvDict

    def _doLaunchInConsole(self, command, cwd, env, input=None):
        """Launch the given command string in a new console.
        (NOTE: this is mostly copied from koRunService.Run())

        Optionally an "input" string can be provided. If it is then the
        command is run via "command < input", where the input is placed
        in a temporary file.
        """
        #print "_doLaunchInConsole(command=%r, cwd=%r, env=%r, "\
        #      "input=%r)"\
        #      % (command, cwd, env, input)
        if not cwd: # Allow cwd="".
            cwd = None
        if not env:
            env = ""
        envDict = self._getEnvDictFromString(env)
        envDict = self._mergeEnvWithParent(envDict)
        # The process library requires that all env strings be ASCII
        # or all Unicode. We'll just make them all Unicode for now.
        #XXX This unicode conversion may not be necessary since
        #    _SaferCreateProcess in process.py.
        uEnvDict = {}
        for key, val in envDict.items():
            uEnvDict[unicode(key)] = unicode(val)
        envDict = uEnvDict

        # Keep a pure 'command' for reporting status to the user.
        actualCommand = command

        if input:
            # Put the input into a temp file.
            inputFileName = tempfile.mktemp()
            inputFile = open(inputFileName, "w")
            inputFile.write(input)
            inputFile.close()
            if ' ' in inputFileName:
                actualCommand += ' < "%s"' % inputFileName
            else:
                actualCommand += ' < %s' % inputFileName
        else:
            inputFileName = None

        if sys.platform.startswith("win"):
            if os.environ.has_key("SHELL"):
                shell = os.environ["SHELL"]
            elif os.environ.has_key("ComSpec"):
                shell = os.environ["ComSpec"]
            else:
                #XXX Would be nice to get full path here so can use
                #    os.spawn* calls.
                shell = "command.com"
            shellArgs = "/C"
        else:
            shell = "xterm"
            shellArgs = "-e /bin/sh"
    
        # A shell script that will be run is created.  The script
        # adds the "Press any key to continue..." feature and
        # ensures that the correct return value is returned.
        if sys.platform.startswith("win"):
            scriptFileName = tempfile.mktemp() + ".bat"
            bat = open(scriptFileName, "w")
            bat.write("@echo off\n")
            # Escape '%' once for being in a batch file plus once for
            # being on a "call ..." line.
            actualCommand = actualCommand.replace("%", "%%%%")
            bat.write("call %s\n" % actualCommand)
            bat.write("set KOMODO_COMMAND_ERRORLEVEL=%ERRORLEVEL%\n")
            bat.write("pause\n")
            # Must "exit" with the last return value to ensure that
            # cmd.exe will return that value
            #XXX does the correct retval actually get returned?
            bat.write("exit %KOMODO_COMMAND_ERRORLEVEL%\n")
            bat.close()
            #XXX does this work on Win9x???
        else:
            # Create a bash script that will be run.
            scriptFileName = tempfile.mktemp() + ".sh"
            script = open(scriptFileName, "w")
            script.write("#!/bin/sh\n")
            script.write("%s\n" % actualCommand)
            script.write("export KOMODO_COMMAND_RETVAL=$?\n")
            script.write("echo -n Press ENTER to continue . . .\n")
            script.write("read KOMODO_DUMMY\n")
            script.write("unset KOMODO_DUMMY\n")
            script.write("exit $KOMODO_COMMAND_RETVAL\n")
            script.close()
        actualCommand = "%s %s %s"\
                        % (shell, shellArgs, scriptFileName)
        #print "RUN: cmdLine is '%s'" % cmdLine
        #print "RUN: -------- %s is as follows -------" % scriptFileName
        #fin = open(scriptFileName, 'r')
        #for line in fin.readlines():
        #    print line,
        #fin.close()
        #print "RUN: ---------------------------------"
        flags = process.Process.CREATE_NEW_CONSOLE

        try:
            child = process.Process(actualCommand, cwd=cwd, env=envDict,
                                    flags=flags)
        except process.ProcessError, ex:
            _gLastErrorSvc.setLastError(ex.errno, str(ex))
            raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))

        # The "invocation_terminated" notification is sent when the
        # child terminates.  A separate thread is created to handle that
        # so this call can return immediately.
        t = threading.Thread(target=self._WaitCleanUpAndNotify,
                             kwargs={'child': child,
                                     'command': command,
                                     'scriptFileName': scriptFileName,
                                     'inputFileName': inputFileName})
        t.start()

    def _doPerlLaunch(self, real_args, cwd, environmentStrings):
        showConsole = self.currentInstancePrefs.getBooleanPref('use-console')
        if not sys.platform.startswith("win") and showConsole:
            # Perl debugging in console on Linux would often crash
            # Komodo on termination. We don't know why but suspect that
            # the problem was in koprocess.cpp (hence the hard Komodo
            # crash), which is normally used for spawning debugging in
            # Komodo. The chosen solution was to use process.py's
            # Process spawning for just this case. Eventually all Komodo
            # debugging should move to process.py.
            command = _joinArgv(real_args[:-1]) # drop the trailing None
            env = '\n'.join(environmentStrings)
            input = self._postData # Send CGI post data if any exists.
            self._doLaunchInConsole(command, cwd, env, input)
        else:
            self._doLaunch(real_args, cwd, environmentStrings)

    def _launchForDebug(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')

        # use the get variables as the user_args if emulating CGI        
        debug_args = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: debug_args.append(get_args)
        if len(debug_args) < 1 and \
           self.currentInstancePrefs.hasPref('params'):
            params = self.currentInstancePrefs.getStringPref('params')
            if params:
                clp = components.classes["@activestate.com/CommandLineParser;1"] \
                         .createInstance(components.interfaces.asICommandLineParser)
                debug_args = clp.parse(params)
        
        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')
        real_args = [executable, "-d"] + self._getExtraPerlOptions() + [script] + debug_args + [None]

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)

        cmdPort = self.prefService.prefs.getLongPref("remotePerlListenPort")
        #listenerPort = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        #print "Perl thinks the debugger port is", cmdPort
        environmentStrings = self._getenv()        
        environmentDict = environmentStringsToDict(environmentStrings)
        #environmentDict["PERLDB_OPTS"] = 'PrintRet=0 RemotePort=127.0.0.1:%d CallKomodo=127.0.0.1:%d' % (cmdPort,listenerPort)
        environmentDict["PERLDB_OPTS"] = 'PrintRet=0 ReadLine=0 RemotePort=127.0.0.1:%d ' % (cmdPort)
        environmentDict["PERL5DB"]="BEGIN { require 'perl5db.pl' }"

        # prepend Komodo's Perl utils dir to PERL5LIB
        perl5LibDirs = environmentDict.get("PERL5LIB", "").split(os.pathsep)
        koDirs = components.classes["@activestate.com/koDirs;1"].\
            getService(components.interfaces.koIDirs)
        perl5LibDirs.insert(0, koDirs.installDir)
        environmentDict["PERL5LIB"] = os.pathsep.join(perl5LibDirs)

        environmentStrings = environmentDictToStrings(environmentDict)

        self._doPerlLaunch(real_args, cwd, environmentStrings)

    def _launchForRun(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        debug_args = components.classes["@activestate.com/CommandLineParser;1"] \
                     .createInstance(components.interfaces.asICommandLineParser) \
                     .parse( self.currentInstancePrefs.getStringPref('params') );
        script = self.currentInstancePrefs.getStringPref('filename')
        executable = self.currentInstancePrefs.getStringPref('executable')
        real_args = [executable] + self._getExtraPerlOptions() + [script] +  debug_args + [None]

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)

        environmentStrings = self._getenv()        

        self._doPerlLaunch(real_args, cwd, environmentStrings)

class koPHPInvocation(koInterpreterInvocation):
    _reg_clsid_ = "{aa80dab8-b5fb-45c4-b1c5-0eec9494955b}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=PHP"
    _reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "PHP"
    _invocation_description_ = "PHP"
    _invocation_chrome_category_ = "chrome://komodo/content/pref/invocations/categories/php.xul"
    _invocation_chrome_instance_ = "chrome://komodo/content/pref/invocations/instances/php.xul"

    def __init__(self):
        koInterpreterInvocation.__init__(self)
        self.checked_can_debug = 0
        self.can_debug = 0
        self.languageService = components.classes[\
                "@activestate.com/koLanguage?language=PHP;1"].\
                getService(components.interfaces.koILanguage)

    def get_supportedInvokeTypes(self):
        rc = INVOKE_RUN
        exe = self.currentInstancePrefs.getStringPref("executable")
        ini = self.currentInstancePrefs.getStringPref("ini")
        argv = [str(exe), '-m']  # XXX str() for fear of a Unicode cmdln
        env = koprocessutils.getUserEnv()
        if ini:
            env["PHPRC"] = ini
        p = process.ProcessOpen(argv, mode='b', env=env)
        lines = p.stdout.readlines()
        # For some reason, PHP linter causes an exception on close
        # with errno = 0, will investigate in PHP later.
        try:
            p.close()
        except IOError, e:
            if e.errno == 0:  
                pass
            else:
                raise

        for line in lines:
            if line.strip() == "Active Debugger":
                self.can_debug = 1

        if self.can_debug:
            rc |= INVOKE_DEBUG
        return rc

    def _fillExtraDefaultPreferences(self, cat, inst):
        ini = self.prefService.prefs.getStringPref("phpConfigFile")
        cat.setStringPref("ini", ini)

        inst.setBooleanPref("disableOutputBuffering", 1)
        inst.setBooleanPref("enableImplicitFlush", 1)
                         
        # since we now always remote debug, php creates the cookie and passes to komodo        
        #logDebug("Don't know how to make a php cookie")
        #inst.setStringPref("debugger.cookie", "MakeMeACookie")
    
    #XXX see bug 18570 need to reconsildate validate and make sure logic is correct in here and in editor-wrapper.
    def validate(self, invokeType):
        if self.currentInstancePrefs is None or self.currentCategoryPrefs is None:
            return 0, "", "Invocation preferences not initialized"

        # Attempt to locate the executable if none specified.
        if not self.currentCategoryPrefs.getStringPref("executable"):
            self.currentCategoryPrefs.setStringPref("executable", self._determineExecutable())

        return self._do_validate(invokeType)
    
    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the PHP executable used for the debug session")
            checkFileExists(prefset, 'filename', "You must specify the main PHP script to be used")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)
            
            #verify we can run
            if invokeType & INVOKE_DEBUG:
                rc = self.get_supportedInvokeTypes()
                if not rc & invokeType:
                    return 0, "", "PHP is not configured properly for debugging."
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None
        
    def _determineExecutable(self):
        koInfoService = components.classes["@activestate.com/koInfoService;1"].getService();
        # use pref'd default if one has been chosen
        phpExe = self.prefService.prefs.getStringPref("phpDefaultInterpreter")
        if not phpExe or not os.path.isfile(phpExe):
            phpExe = self.languageService.\
               getLanguageService(components.interfaces.koIAppInfoEx).\
               executablePath
        if not phpExe or not os.path.isfile(phpExe):
            logError("No PHP interpreter could be found.");
            return ""
        return phpExe

    def _launchForDebug(self):
        self._usecgi = self.currentInstancePrefs.getBooleanPref('sim-cgi')
        script = self.currentInstancePrefs.getStringPref('filename')
        ini = self.currentInstancePrefs.getStringPref('ini')
        buffer = self.currentInstancePrefs.getBooleanPref("disableOutputBuffering")
        flush = self.currentInstancePrefs.getBooleanPref("enableImplicitFlush")
        if ini == "": ini = None
        executable = self.currentInstancePrefs.getStringPref('executable')

        # use the get variables as the user_args if emulating CGI        
        user_args = []
        if self._usecgi:
            get_args = self._build_get_line()
            if get_args: user_args.append(get_args)
        if len(user_args) < 1:
            user_args = components.classes["@activestate.com/CommandLineParser;1"] \
                     .createInstance(components.interfaces.asICommandLineParser) \
                     .parse( self.currentInstancePrefs.getStringPref('params') );
        
        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        real_args = [executable]
        if self._usecgi:
            real_args += [script] + [None]
        else:
            real_args += ["-f", script] + user_args + [None]

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)

        environment = self._getenv()
        
        # this is to handle the fact that PHP ignores command line opts if
        # it decides it is running as a CGI, so we pass the info on the environment as well.
        # so the debug extension now always uses the environment or the ini file for config
        environmentDict = environmentStringsToDict(environment)
        environmentDict["ASDBG_ENABLED"] = '5'
        #environmentDict["ASDBG_PORT"] = '9001'
        environmentDict["ASDBG_REMOTEPORT"] = str(port)
        if buffer: environmentDict["ASDBG_BUFFERING"] = '0'
        if flush: environmentDict["ASDBG_FLUSH"] = '1'
        if ini is not None: environmentDict["PHPRC"] = ini
        environment = environmentDictToStrings(environmentDict)
        
        return self._doLaunch(real_args, cwd, environment)

    def _launchForRun(self):
        self._usecgi = self.currentInstancePrefs.getLongPref('sim-cgi')
        script = self.currentInstancePrefs.getStringPref('filename')
        ini = self.currentInstancePrefs.getStringPref('ini')
        if ini == "": ini = None
        executable = self.currentInstancePrefs.getStringPref('executable')
        user_args = components.classes["@activestate.com/CommandLineParser;1"] \
                     .createInstance(components.interfaces.asICommandLineParser) \
                     .parse( self.currentInstancePrefs.getStringPref('params') );

        real_args = [executable]
        if self._usecgi:
            real_args += [script]
        else:
            real_args += ["-f", script] + user_args + [None]

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = self._getenv()
        
        if ini is not None:
            environmentDict = environmentStringsToDict(environment)
            environmentDict["PHPRC"] = ini
            environment = environmentDictToStrings(environmentDict)
        
        return self._doLaunch(real_args, cwd, environment)




class koXSLTInvocation(koInterpreterInvocation):
    _reg_clsid_ = "{20B8B708-D3A6-423b-BB41-9EDEE5604CD5}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=XSLT"
    _reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "XSLT"
    _invocation_description_ = "XSLT"
    _invocation_chrome_category_ = "chrome://komodo/content/pref/invocations/categories/xslt.xul"
    _invocation_chrome_instance_ = "chrome://komodo/content/pref/invocations/instances/xslt.xul"

    def __init__(self):
        koInterpreterInvocation.__init__(self)
        self.checked_can_debug = 0
        self.can_debug = 0

    def _fillExtraDefaultPreferences(self, cat, inst):
        inst.setStringPref("inputfile", "")

    def _getenv(self):
        env = components.classes["@activestate.com/koUserEnviron;1"].getService().GetEnvironmentStrings()

        if not sys.platform.startswith("win"):
            koDirs = components.classes["@activestate.com/koDirs;1"]\
                     .getService(components.interfaces.koIDirs)
            envToMergeIn = ["LD_LIBRARY_PATH=%s" % koDirs.mozBinDir]
            env = self._environUtils.MergeEnvironmentStrings(
                    env, envToMergeIn)
        
        return env

    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the XSLT executable used for the debug session")

            checkFileExists(prefset, 'filename', "You must specify the main XSLT script to be used")

            inputFile = prefset.getStringPref('inputfile')
            if not inputFile:
                raise PrefException('inputfile', "You must specify the input XML file to be used")
            else:
                addressingScheme, networkLocation, path, parameters, query, fragmentIdentifier = urlparse.urlparse(inputFile)
                if not addressingScheme:
                   if not os.path.isfile(inputFile):
                        raise PrefException('inputfile', "File '%s' does not exist" % inputFile)
                elif addressingScheme == 'file':
                    addressLocation = inputFile.lower().find('file:///')
                    if addressLocation == -1:
                        raise PrefException('inputfile', 'URIs that use the "file" addressing scheme must begin with "file:///"')
                    else:
                        if not os.path.isfile(inputFile[addressLocation + len('file:///'):]):
                            raise PrefException('inputfile', "File '%s' does not exist" % inputFile)

        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None
    
    def _determineExecutable(self):
        # use pref'd default if one has been chosen
        if self.prefService.prefs.hasStringPref("xsltDefaultInterpreter") and\
           self.prefService.prefs.getStringPref("xsltDefaultInterpreter"):
            runLocation = self.prefService.prefs.getStringPref("xsltDefaultInterpreter")
        else:        
            koDirs = components.classes["@activestate.com/koDirs;1"].\
                getService(components.interfaces.koIDirs)

            if sys.platform.startswith('win'): 
                runLocation =  os.path.join( koDirs.installDir, 'xslt', 'ActiveXSLTW.exe' )
            else:
                runLocation = os.path.join( koDirs.installDir, 'xslt', 'XalanDebugger' )
            
        return runLocation


    def _launchForDebug(self):
        showConsole = 0
        script = self.currentInstancePrefs.getStringPref('filename')
        inputfile = self.currentInstancePrefs.getStringPref('inputfile')
        executable = self.currentInstancePrefs.getStringPref('executable')
        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        
        if (showConsole):
            wait_args = ["-WAIT"]
        else:
            wait_args = []

        real_args = [executable, "-Q", "-CONNECTBACKDEBUG", '127.0.0.1', str( port ), "-XSL", script, "-IN", inputfile] + wait_args + [None]
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = self._getenv()
        return self._doLaunch(real_args, cwd, environment)


    def _launchForRun(self):
        showConsole = 0
        script = self.currentInstancePrefs.getStringPref('filename')
        inputfile = self.currentInstancePrefs.getStringPref('inputfile')
        executable = self.currentInstancePrefs.getStringPref('executable')

        if (showConsole):
            wait_args = ["-WAIT"]
        else:
            wait_args = []

        real_args = [executable, "-Q", "-XSL", script, "-IN", inputfile] + wait_args + [None]
        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = components.classes["@activestate.com/koUserEnviron;1"].getService().GetEnvironmentStrings()
        return self._doLaunch(real_args, cwd, environment)



class koTclInvocation(koInterpreterInvocation):
       # If this object needs to be registered, the following 2 are also needed.
    _reg_clsid_ = "{06860119-32f6-432d-a0c0-2398fcd56653}"
    _reg_contractid_ = "@activestate.com/koInvocation;1?type=Tcl"
    _reg_registrar_ = RegisterAsInvocation, None

    _invocation_name_ = "Tcl"
    _invocation_description_ = "Tcl"
    _invocation_chrome_category_ = "chrome://komodo/content/pref/invocations/categories/tcl.xul"
    _invocation_chrome_instance_ = "chrome://komodo/content/pref/invocations/instances/tcl.xul"

    def _fillExtraDefaultPreferences(self, cat, inst):
        # XXX - It doesn't look like TCL_LIBRARYPATH is being used anywhere!
        cat.setStringPref("TCL_LIBRARYPATH", "")
        inst.setStringPref("tclInterpreterType", "wish")
        
    def _determineExecutable(self):
        """determine the Tcl interpreter to use"""
        # use pref'd default if one has been chosen

        # Find a reasonable TclPro installation (possible perf issue computing this
        # even when the user has picked an explicit preference)
        tclProInstallation = components.classes["@activestate.com/koTclProInfoEx;1"].createInstance();
        if not tclProInstallation.selectDefault():
            tclProInstallation = None
            
        type = self.currentInstancePrefs.getStringPref("tclInterpreterType")
        if type == "wish":
            tclExe = self.prefService.prefs.getStringPref("wishDefaultInterpreter")
            if not tclExe and tclProInstallation:
                tclExe = tclProInstallation.wish_path
        else:        
            tclExe = self.prefService.prefs.getStringPref("tclshDefaultInterpreter")
            if not tclExe and tclProInstallation:
                tclExe = tclProInstallation.tclsh_path

        if not tclExe or not os.path.isfile(tclExe):
            logError("No TclPro interpreter could be found.");
            return ""
        else:
            return tclExe

    def _do_validate(self, invokeType):
        try:
            prefset = self.currentInstancePrefs
            checkFileExists(prefset, 'executable', "You must specify the executable used for the debug session")
            checkFileExists(prefset, 'filename',   "You must specify the Tcl script to be debugged")

            cwd = prefset.getStringPref("cwd")
            if cwd and not os.path.isdir(cwd):
                raise PrefException("cwd", "Directory '%s' does not exist" % cwd)

            tclInfo = components.classes["@activestate.com/koTclProInfoEx;1"] \
                     .getService(components.interfaces.koIAppInfoEx)
            tclInfo.installationPath = tclInfo.getInstallationPathFromBinary(prefset.getStringPref("executable"))
                
            if not tclInfo.haveLicense:
                raise PrefException("", r'An ActiveState TclPro license is required to debug Tcl scripts. See the Komodo Tcl preferences or "http://www.ActiveState.com/Products/ASPN_Tcl/" for more information.')
                
        except PrefException, exc:
            return 0, exc.prefid, exc.msg
        return 1, None, None

    def _launchForDebug(self):
        self._usecgi = 0
        debug_args = components.classes["@activestate.com/CommandLineParser;1"] \
                 .createInstance(components.interfaces.asICommandLineParser) \
                 .parse( self.currentInstancePrefs.getStringPref('params') );

        koDirs = components.classes["@activestate.com/koDirs;1"].\
            getService(components.interfaces.koIDirs)

        executable = self._determineExecutable ()
        script     = self.currentInstancePrefs.getStringPref('filename')
        dbgscript  = os.path.join( koDirs.installDir, 'tcl', 'tdbg.tcl' )

        executableArgs = []

        dbArgs = ['-file', script]

        port = self.currentInstancePrefs.getLongPref("debugger.listener-port")
        dbArgs.append("-port-komodo")
        dbArgs.append("%s" % port)

        real_args = [executable] + executableArgs + [dbgscript] + dbArgs + ['--'] + debug_args + [None]

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = self._getenv()        
        return self._doLaunch(real_args, cwd, environment)

    def _launchForRun(self):
        self._usecgi = 0
        script_args = components.classes["@activestate.com/CommandLineParser;1"] \
                     .createInstance(components.interfaces.asICommandLineParser) \
                     .parse( self.currentInstancePrefs.getStringPref('params') );

        script         = self.currentInstancePrefs.getStringPref('filename')
        executable     = self._determineExecutable ()
        executableArgs = []
        
        real_args = [executable] + executableArgs + [script] + script_args + [None]

        cwd = self.currentInstancePrefs.getStringPref('cwd')
        if not cwd:
            cwd = os.path.dirname(script)
        environment = self._getenv()        
        return self._doLaunch(real_args, cwd, environment)
