#!python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

"""Language package for PHP"""

from xpcom import components, ServerException
import os, time
from koLanguageServiceBase import *
import autoinsert
import process
import koprocessutils

class koPHPLanguage(KoLanguageBase):
    name = "PHP"
    _reg_desc_ = "%s Language" % name
    _reg_contractid_ = "@activestate.com/koLanguage?language=%s;1" \
                       % (name)
    _reg_clsid_ = "{F472CC39-0902-4d92-8A5B-7DD0F612F119}"

    namedBlockRE = r'^[ \t]*(?:function\s+(?:[&]*?)\s*([\w_]*?)\s*(\(.*?\)))|(^[ \t]*(?<=\s)class\s+([\w_]*)\s*(?:(?:extends\s+([\w_]*)\s*)|\s*))'
    namedBlockDescription = 'PHP functions and classes'
    defaultExtension = ".php"
    styleBits = 7
    indicatorBits = 1
    cookbookName = "PHP"
    cookbookUrl = "http://aspn.activestate.com/ASPN/Cookbook/PHP"
    variableIndicators = '$'
    # XXX read from config somewhere, download from ActiveState?
    downloadURL = 'http://aspn.ActiveState.com/ASPN/PHP/'
    remoteDebuggerSupport = 1
    commentDelimiterInfo = {
        "line": [ "#", "//" ],
        "block": [ ("/*", "*/") ],
    }

    def get_debugger(self):
        if self._debugger is None:
            koDirs = components.classes["@activestate.com/koDirs;1"].\
              getService(components.interfaces.koIDirs)
            dbgpy = os.path.join(koDirs.mozBinDir, "components" ,"koPHPDebugger.py")
            if os.path.exists(dbgpy):
                self._debugger = KoDebuggerLanguageService(self.name)
            else:
                self._debugger = None
        return self._debugger

    def get_lexer(self):
        if self._lexer is None:
            self._lexer = KoLexerLanguageService()
            self._lexer.setLexer(components.interfaces.ISciMoz.SCLEX_HTML)
            # XXX config options needed?
            self._lexer.setProperty('fold.html','1')
            #self._lexer.setProperty('fold.compact','1')
            # keyword sets
            # 0 = html # 1 = JavaScript # 2 = VBScript # 3 = PythonScript # 4 = PHP
            self._lexer.setKeywords(1, self._js_keywords)
            self._lexer.setKeywords(2, self._vb_keywords)
            self._lexer.setKeywords(4, self._php_keywords)
            self._lexer.supportsFolding = 1
        return self._lexer

    def get_linter(self):
        if self._linter is None:
            self._linter = KoLinterLanguageService(
                "@activestate.com/koPHPLinter;1", None)
        return self._linter

    def get_methodcompletion(self):
        if self._methodcompletion is None:
            self._methodcompletion = components.classes["@activestate.com/koPHPCompletionLanguageService;1"].getService(components.interfaces.koICompletionLanguageService)
        return self._methodcompletion

    def get_commenter(self):
        #XXX Could override default KoCommenterLangSvc to use styles to tell
        #    if should use HTML or PHP style comment delimiters. 
        if self._commenter is None:
            self._commenter = KoCommenterLanguageService(
                self.commentDelimiterInfo)
        return self._commenter

    def get_inserter(self):
        if self._inserter is None:
            self._inserter = KoWSInsertLanguageService(
                autoinsert.PHPAutoInsert(),
                "$%s = WebService::ServiceProxy(\"%s\", \"%s\");",
                "webservice.php", 1)
        return self._inserter

    def get_interpreter(self):
        if self._interpreter is None:
            self._interpreter = components.classes["@activestate.com/koPHPInfoEx;1"].getService()
        return self._interpreter

    def get_cookbook(self):
        if self._cookbook is None:
            self._cookbook = KoCookbookLanguageService(
                self.cookbookName, self.cookbookUrl)
        return self._cookbook

    # XXX Scintilla will not work with uppercased keywords!
    
    # XXX later these need to be centralized.
    _php_keywords = [
                "define", "true", "false", 
                "int", "integer", "real", "double",
                "float", "string",  "object", "bool", "boolean",
                "this", "virtual", "parent", 
                # http://www.php.net/manual/en/reserved.php#reserved.keywords
                "and", "or", "xor", "__file__", "__line__",
                "array", "as", "break", "case", "cfunction",
                "class", "const", "continue", "declare"
                "default", "die", "do", "echo", "else",
                "elseif", "empty", "enddeclare", "endfor", "endforeach",
                "endif", "endswitch", "endwhile", "eval", "exit",
                "extends", "for", "foreach", "function", "global",
                "if", "include", "include_once", "isset", "list",
                "new", "old_function", "print", "require", "require_once",
                "return", "static", "switch", "unset", "use",
                "var", "while", "_function_", "_class_",
                # http://www.php.net/manual/en/reserved.constants.core.php
                "php_version",
                "php_os",
                "default_include_path",
                "pear_install_dir",
                "pear_extension_dir",
                "php_extension_dir",
                "php_bindir",
                "php_libdir",
                "php_datadir",
                "php_sysconfdir",
                "php_localstatedir",
                "php_config_file_path",
                "php_output_handler_start",
                "php_output_handler_cont",
                "php_output_handler_end",
                "e_error",
                "e_warning",
                "e_parse",
                "e_notice",
                "e_core_error",
                "e_core_warning",
                "e_compile_error",
                "e_compile_warning",
                "e_user_error",
                "e_user_warning",
                "e_user_notice",
                "e_all",
                # http://www.php.net/manual/en/reserved.constants.standard.php
                "extr_overwrite",
                "extr_skip",
                "extr_prefix_same",
                "extr_prefix_all",
                "extr_prefix_invalid",
                "extr_prefix_if_exists",
                "extr_if_exists",
                "sort_asc",
                "sort_desc",
                "sort_regular",
                "sort_numeric",
                "sort_string",
                "case_lower",
                "case_upper",
                "count_normal",
                "count_recursive",
                "assert_active",
                "assert_callback",
                "assert_bail",
                "assert_warning",
                "assert_quiet_eval",
                "connection_aborted",
                "connection_normal",
                "connection_timeout",
                "ini_user",
                "ini_perdir",
                "ini_system",
                "ini_all",
                "m_e",
                "m_log2e",
                "m_log10e",
                "m_ln2",
                "m_ln10",
                "m_pi",
                "m_pi_2",
                "m_pi_4",
                "m_1_pi",
                "m_2_pi",
                "m_2_sqrtpi",
                "m_sqrt2",
                "m_sqrt1_2",
                "crypt_salt_length",
                "crypt_std_des",
                "crypt_ext_des",
                "crypt_md5",
                "crypt_blowfish",
                "directory_separator",
                "seek_set",
                "seek_cur",
                "seek_end",
                "lock_sh",
                "lock_ex",
                "lock_un",
                "lock_nb",
                "html_specialchars",
                "html_entities",
                "ent_compat",
                "ent_quotes",
                "ent_noquotes",
                "info_general",
                "info_credits",
                "info_configuration",
                "info_modules",
                "info_environment",
                "info_variables",
                "info_license",
                "info_all",
                "credits_group",
                "credits_general",
                "credits_sapi",
                "credits_modules",
                "credits_docs",
                "credits_fullpage",
                "credits_qa",
                "credits_all",
                "str_pad_left",
                "str_pad_right",
                "str_pad_both",
                "pathinfo_dirname",
                "pathinfo_basename",
                "pathinfo_extension",
                "char_max",
                "lc_ctype",
                "lc_numeric",
                "lc_time",
                "lc_collate",
                "lc_monetary",
                "lc_all",
                "lc_messages",
                "abday_1",
                "abday_2",
                "abday_3",
                "abday_4",
                "abday_5",
                "abday_6",
                "abday_7",
                "day_1",
                "day_2",
                "day_3",
                "day_4",
                "day_5",
                "day_6",
                "day_7",
                "abmon_1",
                "abmon_2",
                "abmon_3",
                "abmon_4",
                "abmon_5",
                "abmon_6",
                "abmon_7",
                "abmon_8",
                "abmon_9",
                "abmon_10",
                "abmon_11",
                "abmon_12",
                "mon_1",
                "mon_2",
                "mon_3",
                "mon_4",
                "mon_5",
                "mon_6",
                "mon_7",
                "mon_8",
                "mon_9",
                "mon_10",
                "mon_11",
                "mon_12",
                "am_str",
                "pm_str",
                "d_t_fmt",
                "d_fmt",
                "t_fmt",
                "t_fmt_ampm",
                "era",
                "era_year",
                "era_d_t_fmt",
                "era_d_fmt",
                "era_t_fmt",
                "alt_digits",
                "int_curr_symbol",
                "currency_symbol",
                "crncystr",
                "mon_decimal_point",
                "mon_thousands_sep",
                "mon_grouping",
                "positive_sign",
                "negative_sign",
                "int_frac_digits",
                "frac_digits",
                "p_cs_precedes",
                "p_sep_by_space",
                "n_cs_precedes",
                "n_sep_by_space",
                "p_sign_posn",
                "n_sign_posn",
                "decimal_point",
                "radixchar",
                "thousands_sep",
                "thousep",
                "grouping",
                "yesexpr",
                "noexpr",
                "yesstr",
                "nostr",
                "codeset",
                "log_emerg",
                "log_alert",
                "log_crit",
                "log_err",
                "log_warning",
                "log_notice",
                "log_info",
                "log_debug",
                "log_kern",
                "log_user",
                "log_mail",
                "log_daemon",
                "log_auth",
                "log_syslog",
                "log_lpr",
                "log_news",
                "log_uucp",
                "log_cron",
                "log_authpriv",
                "log_local0",
                "log_local1",
                "log_local2",
                "log_local3",
                "log_local4",
                "log_local5",
                "log_local6",
                "log_local7",
                "log_pid",
                "log_cons",
                "log_odelay",
                "log_ndelay",
                "log_nowait",
                "log_perror"
                ]

    _js_keywords = ["abstract", "boolean", "break", "byte", "case", "catch",
                 "char", "class", "const", "continue", "debugger", "default",
                 "delete", "do", "double", "else", "enum", "export",
                 "extends", "final", "finally", "float", "for", "function",
                 "goto", "if", "implements", "import", "in", "instanceof",
                 "int", "interface", "long", "native", "new", "package",
                 "private", "protected", "public", "return", "short",
                 "static", "super", "switch", "synchronized", "this", "throw",
                 "throws", "transient", "try", "typeof", "var", "void",
                 "volatile", "while", "with"]

    _vb_keywords = ["and", "as", "attribute", "begin", 
                    "byref", "byval", "case", "call", "const",
                    "continue", "dim", "do", "each", "else",
                    "elseif", "end", "error", "exit", "false",
                    "for", "function", "global", "goto", "if",
                    "in", "loop", "me", "new", "next", "not",
                    "nothing", "on", "optional", "or", "private",
                    "public", "redim", "rem", "resume", "select",
                    "set", "sub", "then", "to", "true", "type",
                    "while", "with", "boolean", "byte", "currency",
                    "date", "double", "integer", "long", "object",
                    "option", "single", "string", "type", "variant"]
