# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

# An explanation of how style prefs work:
#
# The koStyleManager class reads in two files:
#   defaultStylePrefs.xml, and
#   styleMapPrefs.xml
#
# defaultStylePrefs.xml defines default fonts and font sizes,
# foreground and background colours, and italic and bold settings for
# each "style" within a particular "language". styleMapPrefs.xml
# defines a mapping between simplified style names (such as "comment")
# and style numbers within Scintilla. Entries in styleMapPrefs.xml may
# be either raw decimal numbers or identifiers which can be found in
# the ISciMoz interface.
#
# Style preference names have a very particular (and important!)
# format. Each style preference name specifies the language, style
# name, and style type of the style preference. These fields are
# separated by forward slashes. Each style name also starts with
# "style/". For example,
#
#   style/Python/comment/fore
#
# ... is the style name for the foreground colour setting of comments
# in the Python language. Generally, style names have the following
# structure:
#
#   style/[language]/[stylemap-name]/[style-type]
#
# The language field must match exactly with the identifier of a
# particular preference set within the "languageToStyleMap" preference set
# defined in styleMapPrefs.xml. The stylemap-name must match a
# particular preference defined within that language's corresponding
# preference set. i.e. "Python/comment" comes from:
#
#  <preference-set id='Python'>
#    <string id='comment'>SCE_P_COMMENTLINE</string>
#
# The language field in a style preference name may be replaced by an
# asterisk to specify that the style preference shall apply to _all_
# languages. The stylemap-name field may also be replaced by an
# asterisk, in which case the style preference shall apply to _all_
# styles within a particular language (or all languages if the
# language field is "*". For example,
#
#     <string id="style/Rx/*/font">Courier</string>
#
# This preference specifies that the font for all styles in Rx
# shall be "Courier".
#
#     <long id="style/*/*/size">10</long>
#
# This preference specifies that the default font size for all
# styles in all languages shall be 10 points.

import os, time, sys
from xpcom import components, COMException, ServerException
from koLogger import logError, logWarning

from koStyleApplier import *

keep_performance_stats = 0

if keep_performance_stats:
    class Timer:
        def __init__(self, msg):
            self._msg = msg
            self._start = time.clock()

        def __del__(self):
            end = time.clock()
            print "%s took %s seconds" % (self._msg, end - self._start)

_gKoEncodingSvc = None

_defaultFontSpec = 'iso8859-15' # no pref, use latin-9, largest coverage

def _getFontSpecFromEncoding(encoding_name):
    info = _gKoEncodingSvc.get_encoding_info(encoding_name)
    if not info:
        return _defaultFontSpec
    return info.fontspec

def _getDefaultStylePrefName(useFixedFont, encoding_name=None):
    prefservice = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService);
    if encoding_name:
        lang = _getFontSpecFromEncoding(encoding_name)
    elif prefservice.prefs.hasPref('encodingDefault'):
        encoding_name = prefservice.prefs.getStringPref('encodingDefault')
        lang = _getFontSpecFromEncoding(encoding_name)
    else:
        lang = _defaultFontSpec
    
    return _getLanguageStylePrefName(useFixedFont, lang)

def _getLanguageStylePrefName(useFixedFont, fontspec):
    if useFixedFont:
        return fontspec+"_fixed"
    else:
        return fontspec+"_proportional"


class koStyleManager:
    _com_interfaces_ = [components.interfaces.koIStyleManager]
    _reg_contractid_ = "@activestate.com/koStyleManager;1"
    _reg_clsid_ = "{5f36e73c-3360-44ce-b926-44796cd8e3b9}"
    
    def __init__(self):
        # Don't get this service at the top level so it is not run at
        # regxpcom-time.
        global _gKoEncodingSvc
        _gKoEncodingSvc = components.classes["@activestate.com/koEncodingServices;1"] \
            .getService(components.interfaces.koIEncodingServices);

        # Get the user style preferences
        self.prefservice = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService);

        self._defaultEncoding = None

        self.stylePrefs = {}
        self._reverseStyleDicts = {}

        self._initStyleMap()
        self._initStylePrefs()

        # clone set to fall back on when cancel is chosen
        self._clonePrefs()
        
        
    def _initStyleMap(self):
        # t = Timer("koStyleManager::_initStyleMap")
        self.styleMapPrefs = self.prefservice.getPrefs("styleMap")

    def _initStylePrefs(self):
        stylePrefs = self.stylePrefs["editor"] = self.prefservice.getPrefs("editorStyles")
        stylePrefs.id = 'userstyles'
        _defaultStylePrefs = self.prefservice.getPrefs("editorStylesDefaults")
        self._defaultEncoding = self.prefservice.prefs.getStringPref("encodingDefault")
        self._encoding_info = _gKoEncodingSvc.get_encoding_info(self._defaultEncoding)
        
        assert _defaultStylePrefs is not None, "Appear to have no parent!"
        # For every language in the default preference set, we need to
        # ensure the user preference set has one of the same name.
        # We then set the parent of each individual language to the
        # individual language in the default.
        for language in _defaultStylePrefs.getPrefIds():
            if not stylePrefs.hasPref(language):
                stylePrefs.setPref(language, components.classes["@activestate.com/koPreferenceSet;1"].createInstance())
            stylePrefs.getPref(language).parent = _defaultStylePrefs.getPref(language)

        #TODO: allow for user prefs for printer styles
        self.stylePrefs["printer"] = self.prefservice.getPrefs("printerStyles")
        
    def _clonePrefs(self):
        self._clonedPrefs = self.stylePrefs["editor"].clone()

    def getStyleNames(self, language):
        styleMap = self.styleMapPrefs.getPref(language)
        return styleMap.getPrefIds()

    def getStyle(self, device, language, styleName, useFixedFont):
        styleMap = self.styleMapPrefs.getPref(language)
        stylePrefs = self.stylePrefs[device].getPref(language)
        defaultStylePrefName = _getDefaultStylePrefName(useFixedFont)
        defPrefs = self.stylePrefs[device].getPref(defaultStylePrefName)
        return koStyle(language, styleMap, stylePrefs, styleName, defPrefs, self._encoding_info)
    
    def getStyleByNumber(self, device, language, styleNumber, useFixedFont):
        styleName = self._getStyleNameByNumber(language, styleNumber)
        return self.getStyle(device, language, styleName, useFixedFont)
    
    def getDefaultStyle(self, device, useFixedFont):
        prefName = _getDefaultStylePrefName(useFixedFont)
        stylePrefs = self.stylePrefs[device].getPref(prefName)
        return koDefaultStyle(stylePrefs,self._encoding_info)

    def getDefaultEncodingStyle(self, device, encoding, useFixedFont):
        prefName = _getLanguageStylePrefName(useFixedFont, encoding)
        stylePrefs = self.stylePrefs[device].getPref(prefName)
        return koDefaultStyle(stylePrefs,self._encoding_info)
    
    def getStylePrefs(self, device):
        return self.stylePrefs[device]
    
    def saveStyles(self):
        self._writeOutStyles()
        # And update the cloned set.
        self._clonePrefs()

    def revertStyles(self):
        prefset = self.stylePrefs["editor"]
        prefset.reset()
        prefset.update(self._clonedPrefs)

    def _writeOutStyles(self):
        # if any values have changed, write to file
        self.prefservice.savePrefsState("editorStyles")
        
    def _getStyleNameByNumber(self, language, styleNumber):
        # XXX Failure point -- if a style definition is not complete,
        # this fails quite badly.
        styledict = self._getReverseStyleDict(language)
        try:
            return styledict[styleNumber]
        except KeyError:
            logError("Incomplete Style definition for style number %s for language %s\nReturning random style" % (styleNumber, language))
            return styledict.values()[0] # XXXX  Am unable to propagate the error up to the UI.

    def _getReverseStyleDict(self, language):
        if not self._reverseStyleDicts.has_key(language):
            self._reverseStyleDicts[language] = \
                                        self._makeReverseStyleDict(language)
        return self._reverseStyleDicts[language]

    def _makeReverseStyleDict(self, language):
        styleMap = self.styleMapPrefs.getPref(language).\
                   QueryInterface(components.interfaces.koIPreferenceSet)
        prefIds = styleMap.getPrefIds()

        reverseStyleDict = {}
        for id in prefIds:
            rawValue = styleMap.getStringPref(id)
            try:
                value = int(rawValue)
            except ValueError:
                # if there is an error in the style map, just let it die
                value = getattr(components.interfaces.ISciMoz, rawValue)
            reverseStyleDict[value] = id
        return reverseStyleDict

    # Reset a specific language's styles back to "factory defaults".
    def resetLanguage(self, language):
        styleMap = self.stylePrefs["editor"].getPref(language)
        prefIds = styleMap.getPrefIds()
        for id in prefIds:
            styleMap.deletePref(id)

    # Reset the default proportional and default fixed fonts.
    def resetDefaults(self):
        device = "editor"
        defaultStylePrefName = _getDefaultStylePrefName(0)
        defPrefs = self.stylePrefs[device].getPref(defaultStylePrefName)
        prefIds = defPrefs.getPrefIds()
        for id in prefIds:
            defPrefs.deletePref(id)

        defaultStylePrefName = _getDefaultStylePrefName(1)
        defPrefs = self.stylePrefs[device].getPref(defaultStylePrefName)
        prefIds = defPrefs.getPrefIds()
        for id in prefIds:
            defPrefs.deletePref(id)

class koScintillaStyleApplier:
    _com_interfaces_ = [components.interfaces.koIScintillaStyleApplier,
                        components.interfaces.nsIObserver]
    _reg_contractid_ = "@activestate.com/koScintillaStyleApplier;1"
    _reg_clsid_ = "{66046EC6-DE69-4942-97D4-6032DD5C6DF9}"

    def __init__(self):
        self._scintilla = None
        self._language = None
        self._stylePrefs = None
        self._styleMapPrefs = None
        self._styleManager = components.classes["@activestate.com/koStyleManager;1"].getService()

    def destroy(self):
        self._stylePrefsLanguage.RemoveObserver(self)
        self._removeStylePrefObservers()
        
    def _removeStylePrefObservers(self):
        defStylePrefs = self._stylePrefs.getPref(_getDefaultStylePrefName(1,self._encoding_name))
        defStylePrefs.RemoveObserver(self)

        defStylePrefs = self._stylePrefs.getPref(_getDefaultStylePrefName(0,self._encoding_name))
        defStylePrefs.RemoveObserver(self)

    def initWithEncoding(self, scintilla, language, viewPrefs, encoding_name):
        self._do_init(scintilla, language, viewPrefs, encoding_name)
        
    def init(self, scintilla, language, viewPrefs):
        self._do_init(scintilla, language, viewPrefs)
        
    def _do_init(self, scintilla, language, viewPrefs, encoding_name='utf-8'):
        self._scintilla = scintilla
        self._language = language
        self._viewPrefs = viewPrefs
        self._encoding_name = encoding_name
        self._encoding_info = _gKoEncodingSvc.get_encoding_info(encoding_name)
        
        self._styleMapPrefs = self._styleManager.styleMapPrefs
        self._styleMapPrefsLanguage = self._styleMapPrefs.getPref(language)
        self._stylePrefs = self._styleManager.getStylePrefs("editor")
        self._stylePrefsLanguage = self._stylePrefs.getPref(language)
        self._stylePrefsLanguage.AddObserver(self)
        self._addStylePrefObservers()

    def _addStylePrefObservers(self):
        defStylePrefs = self._stylePrefs.getPref(_getDefaultStylePrefName(1,self._encoding_name))
        defStylePrefs.AddObserver(self)

        defStylePrefs = self._stylePrefs.getPref(_getDefaultStylePrefName(0,self._encoding_name))
        defStylePrefs.AddObserver(self)

    def Observe(self, stylePrefs, prefsetName, prefName):
        # Apply the style pref change to our scintilla
        (language, styleName, styleType) = splitStylePrefName(prefName)
        assert language=="*" or language == self._language, "Notified for wrong language - %s!" % (language,)

        if language=="*" or not stylePrefs.hasPref(prefName):
            # default changed, or preference deleted - update everything as working out exactly what this means is too hard!
            self.applyStyles()
            return
        styleMap = self._styleMapPrefsLanguage
        stylePrefs = self._stylePrefsLanguage

        if self._viewPrefs is not None:
            useFixedFont = self._viewPrefs.getBooleanPref('editUseFixedFont')
        else:
            useFixedFont = 0
        defaultStylePrefName = _getDefaultStylePrefName(useFixedFont,self._encoding_name)
        defaultPrefs = self._stylePrefs.getPref(defaultStylePrefName)
        
        sa = koStyleApplier(self._language, styleMap,
                            stylePrefs, defaultPrefs,self._encoding_info)
        sa.applyOneStyle(prefName, self._scintilla)

    def setEncoding(self, encoding_name):
        self._removeStylePrefObservers()
        self._encoding_name = encoding_name
        self._encoding_info = _gKoEncodingSvc.get_encoding_info(encoding_name)
        self._addStylePrefObservers()
        
    def applyStyles(self):
        # XXX: This algorithm is slow, and NOT because of all the XPCOM work.
        #      Probably a site for future optimization work.
        assert self._stylePrefs is not None, "_userStylePrefs not initialized!"
        assert self._language is not None, "invalid language"
        assert self._scintilla is not None, "invalid scintilla"
        
        if self._viewPrefs is not None:
            useFixedFont = self._viewPrefs.getBooleanPref('editUseFixedFont')
        else:
            useFixedFont = 0

        styleMap = self._styleMapPrefsLanguage

        # Apply default styles.
        defaultStylePrefName = _getDefaultStylePrefName(useFixedFont,self._encoding_name)
        
        defStylePrefs = self._stylePrefs.getPref(defaultStylePrefName).\
                    QueryInterface(components.interfaces.koIPreferenceSet)
        sa = koBulkStyleApplier(self._language, styleMap, \
                                defStylePrefs, None, self._encoding_info)
        sa.applyStyles(self._scintilla)

        # Apply language-specific styles.
        stylePrefs = self._stylePrefsLanguage
        sa = koBulkStyleApplier(self._language, styleMap, stylePrefs, \
                                defStylePrefs, self._encoding_info)
        sa.applyStyles(self._scintilla)


