#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

# some useful system utilities

import sys, os, operator
from xpcom import components, nsError, ServerException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from koLogger import logError, logWarning, logDebug


class koSysUtils:
    _com_interfaces_ = [components.interfaces.koISysUtils]
    _reg_clsid_ = "{56F686E0-A989-4714-A5D6-D77BC850C5C0}"
    _reg_contractid_ = "@activestate.com/koSysUtils;1"
    _reg_desc_ = "System Utilities Service"

    def __init__(self):
        self.F_OK = os.F_OK
        self.R_OK = os.R_OK
        self.W_OK = os.W_OK
        self.X_OK = os.X_OK
        self._proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        # XXX Should just cache the user env locally. It ain't changin'.
        self._userEnvSvc = components.classes["@activestate.com/koUserEnviron;1"].\
            getService(components.interfaces.koIUserEnviron)
        self._userEnvProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koIUserEnviron, self._userEnvSvc,
            PROXY_ALWAYS | PROXY_SYNC)

    def _SameFile(self, fname1, fname2):
        return ( os.path.normpath(os.path.normcase(fname1)) ==\
            os.path.normpath(os.path.normcase(fname2)) )

    def _GetRegisteredExecutable(self, exe):
        if sys.platform.startswith('win'):
            # If on Windows, look up the Path in the "App Paths" registry
            import _winreg
            try:
                #XXX might have to be smart enough to handle HKCU too
                return _winreg.QueryValue(_winreg.HKEY_LOCAL_MACHINE,
                    "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths"\
                    "\\" + exe + ".exe")
            except _winreg.error:
                pass
        return None

    def FastCheckIfHaveExecutable(self, exe):
        if sys.platform.startswith("win"):
            path = self._userEnvProxy.get("PATH")
            import win32api
            try:
                win32api.SearchPath(path, exe, ".exe")
                return 1
            except win32api.error:
                pass
            return self._GetRegisteredExecutable(exe) is not None
        # Fast check - dont go a searching down PATHEXT etc on Windows.
        return operator.truth(self.Which(exe))

    def Which(self, exeName):
        """Return the full path to the given executable using the user's PATH.
        Return the empty string if the executable could not be found.
        ** NOTE: The *user's* environment is reproduced by checking
                 <user-data-dir>/startup-env.tmp for any changes to the PATH
                 that komodo.exe might have made.
        """
        # Result: string
        # In: exeName: string
        return self._DoWhich(exeName, 1)

    def WhichAll(self, exeName):
        return self._DoWhich(exeName, 0)

    def _DoWhich(self, exeName, return_first):
        """Same as Which() except all matching executables are given."""
        # Result: Array[string]
        # In: exeName: string
        # Out: uint32
        matches = []
        # otherwise use the PATH environment variable
        path = self._userEnvProxy.get("PATH").split(os.pathsep)
        # windows has the concept of a list of extensions (PATHEXT env var)
        exts = []
        if sys.platform.startswith("win"):
            if self._userEnvProxy.has("PATHEXT"):
                exts = self._userEnvProxy.get("PATHEXT").split(os.pathsep)
            # on WinNT PATHEXT is not returned by the Win32 API
            # GetEnvironmentStrings(), which is used to get the user
            # environment, therefore, default to the current runtime
            # environment (should be safe, Komodo is not going to change
            # PATHEXT)
            elif os.environ.has_key("PATHEXT"):
                exts = os.environ("PATHEXT").split(os.pathsep)
            # If '.exe' is not in exts then obviously this is Win9x and or a
            # bogus PATHEXT (this is possible, and likely, on Win9x where
            # ActivePython adds .py;.pyc;.pyo to an non-existant PATHEXT)
            # then use a reasonable default.
            for ext in exts:
                if ext.lower() == ".exe":
                    break
            else:
                exts = ['.COM', '.EXE', '.BAT']
        # file name cannot have path separators because PATH lookup does not
        # work that way
        if os.sep in exeName or os.altsep and os.altsep in exeName:
            pass
        else:
            for dirName in path:
                # on windows the dirName *could* be quoted, drop the quotes
                if sys.platform.startswith("win") and\
                   len(dirName) >= 2 and\
                   dirName[0] == '"' and dirName[-1] == '"':
                    dirName = dirName[1:-1]
                for ext in ['']+exts:
                    absExeName = os.path.join(dirName, exeName+ext)
                    if os.path.isfile(absExeName):
                        if return_first:
                            return absExeName
                        matches.append(os.path.normpath(absExeName))
            # last look up the Path in the "App Paths" registry or whatever
            ret = self._GetRegisteredExecutable(exeName)
            if ret is not None:
                if return_first:
                    return ret
                matches.append(ret)
        if return_first: # If I want the first and got here, I must have none!
            return ""
        # check for duplicates
        matchesWithoutDups = []
        for i in range(len(matches)):
            for j in range(i+1, len(matches)):
                if self._SameFile(matches[i], matches[j]):
                    break
            else:
                matchesWithoutDups.append(matches[i])
        return matchesWithoutDups
    
    def IsFile(self, filename):
        try:
            return os.path.isfile(filename)
        except OSError, e:
            raise ServerException, (nsError.NS_ERROR_FILE_NOT_FOUND, str(e))

    def IsDir(self, dirname):
        try:
            return os.path.isdir(dirname)
        except OSError, e:
            raise ServerException, (nsError.NS_ERROR_FILE_NOT_FOUND, str(e))

    def Stat(self, filename):
        try:
            return os.stat(filename)[:10]
        except OSError, e:
            raise ServerException, (nsError.NS_ERROR_FILE_NOT_FOUND, str(e))

    def Access(self, filename, mode):
        return os.access(filename, mode)

    def Touch(self, filename):
        if os.path.exists(filename):
            raise NotImplementedError("Don't know how to easily implement "\
                                      "'touch' for a file that exists.")
        else:
            f = open(filename, 'w')
            f.close()

    def joinargv(self, argv):
        return _joinArgv(argv)

    def diff_files(self, fname1, fname2):
        # XXX upgrade to deal with remote files someday?
        import difflib
        try:
            left = open(fname1, 'rb').read().split('\n');
            right = open(fname2, 'rb').read().split('\n');
        except IOError, ex:
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                           .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_UNEXPECTED, str(ex))
        return ''.join(difflib.ndiff(left, right))
    
def _escapeArg(arg):
    """Escape the given command line argument for the shell."""
    #XXX There is a probably more that we should escape here.
    return arg.replace('"', r'\"')


def _joinArgv(argv):
    r"""Join an arglist to a string appropriate for running.

        >>> import os
        >>> _joinArgv(['foo', 'bar "baz'])
        'foo "bar \\"baz"'
    """
    cmdstr = ""
    for arg in argv:
        if ' ' in arg or ';' in arg:
            cmdstr += '"%s"' % _escapeArg(arg)
        else:
            cmdstr += _escapeArg(arg)
        cmdstr += ' '
    if cmdstr.endswith(' '): cmdstr = cmdstr[:-1]  # strip trailing space
    return cmdstr
