#!python

from xpcom import components
from koSciMozEvents import SciMozEvents
from koEditorController import CommandHandler
import threading

# XXX Read-only hack:
#     Komodo's terminals used to maintain themselves in read-only state
#     and only open up to specifically lett characters through while
#     active. This relies upon .onCharAdded() getting called even when
#     read-only. However onCharAdded() does NOT get called while
#     read-only when scintilla.newLine() is called. So new-line handling
#     on stdin in the terminal was wonky. The HACK solution here is to
#     not bother with maintaining a read-only state. The right answer is
#     to use a special shellscintilla XBL which has special keypress,
#     drag 'n drop, and paste handlers. See similar comments in
#     koRunTerminal.py (which still maintains the read-only state).

class KoTerminal(SciMozEvents):
    _com_interfaces_ = [components.interfaces.koITerminal, components.interfaces.ISciMozEvents]
    _reg_desc_ = "Komodo Terminal"
    _reg_contractid_ = "@activestate.com/koTerminal;1"
    _reg_clsid_ = "{B247496F-E335-45f4-BB7D-2C43CE60969A}"

    def __init__(self):
        self.emulateCGI = 0
        
    def init(self, scintilla, viewName, scratchDocument, language, encoding_name):
        self._scintilla = scintilla
        self._terminalController = TerminalCommandHandler(scintilla, self)
        self.acceptInput = 1
        self.lock = threading.RLock()
        self.inputReady = threading.Condition()
        self.stdin = KoTerminalInFile(self._scintilla, self)
        self.stdout = KoTerminalOutFile(self._scintilla, self)
        self.stderr = KoTerminalOutFile(self._scintilla, self)
        self.clear()

    def clear(self):
        self.lock.acquire()
        # See "Read-only hack" note above.
        #self._scintilla.readOnly = 0
        self._scintilla.text = ''
        self.lineBuffer = ''
        self.inBuffer = ''
        self._scintilla.gotoPos(self._scintilla.length)
        #self._scintilla.readOnly = 1
        self.lock.release()

    def clearInput(self):
        self.inputReady.acquire()
        self.inputReady.notify()
        self.inputReady.release()

    def get_controller(self):
        return self._terminalController

    def onCharAdded(self, char):
        if self.acceptInput:
            self.lock.acquire()
            # See "Read-only hack" note above.
            #self._scintilla.readOnly = 0
            #self._scintilla.insertText(self._scintilla.length, chr(char))
            #self._scintilla.gotoPos(self._scintilla.length)
            #self._scintilla.readOnly = 1
            self.send(chr(char))
            self.lock.release()

    def send(self, character):
        self.lineBuffer = self.lineBuffer + character

    def onZoom(self):
        s = self._scintilla
        if s.getMarginWidthN(0):
            s.setMarginWidthN(0, s.textWidth(0, str(s.lineCount))+5)

class TerminalCommandHandler(CommandHandler):
    
    def __init__(self, scimoz, terminal):
        self.wrapper = None
        self._terminal = terminal  #XXX creates a cycle!
        self.scimoz = scimoz

    def _do_cmd_newline(self):
        if self._terminal.acceptInput:
            self._terminal.lock.acquire()

            # scimoz.newLine() will cause cause the terminal's onCharAdded
            # handler to be called, adding the EOL marker to lineBuffer
            # so copy lineBuffer beforehand, and clear it afterward
            self._terminal.inBuffer = self._terminal.lineBuffer + '\n'
            #print "inBuffer = ", self._terminal.lineBuffer
            self.scimoz.newLine()
            self._terminal.lineBuffer = ''
            
            self._terminal.lock.release()

            self._terminal.inputReady.acquire()
            self._terminal.inputReady.notify()
            self._terminal.inputReady.release()

    def _do_cmd_backSmart(self):
        if self._terminal.acceptInput and self._terminal.lineBuffer:
                self._terminal.lock.acquire()
                # See "Read-only hack" note above.
                #self.scimoz.readOnly = 0
                self.scimoz.selectionStart = self.scimoz.length - 1
                self.scimoz.selectionEnd = self.scimoz.length
                self.scimoz.replaceSel('');
                #self.scimoz.readOnly = 1
                self._terminal.lineBuffer = self._terminal.lineBuffer[:-1]
                self._terminal.lock.release()

class KoTerminalFile:
    _com_interfaces_ = [components.interfaces.koIFile]

    def __init__(self, scintilla, terminal):
        self._terminal = terminal
        self._scintilla = scintilla
        
    def read(self, aCount):
        return ''
    
    def write(self, data):
        pass

    def flush(self):
        pass

    def close(self):
        pass


# this expects to be used from another thread
class KoTerminalOutFile(KoTerminalFile):
    
    def write(self, data):
        self._terminal.lock.acquire()
        # See "Read-only hack" note above.
        #self._scintilla.readOnly = 0
        self._scintilla.addText(len(data), data)
        self._scintilla.gotoPos(self._scintilla.length)
        #self._scintilla.readOnly = 1
        self._terminal.lock.release()

class KoTerminalInFile(KoTerminalFile):

    def __init__(self, scintilla, terminal):
        KoTerminalFile.__init__(self, scintilla, terminal)
        
    def read(self, aCount):
        #print "waiting for input ",self._readWait
        self._terminal.inputReady.acquire()
        self._terminal.inputReady.wait(self._readWait) 
        self._terminal.inputReady.release()

        #print "KoTerminalInFile.read(): have input, waiting for lock"
        self._terminal.lock.acquire()
        if aCount >= len(self._terminal.inBuffer):
            data = self._terminal.inBuffer
            self._terminal.inBuffer = ''
        else:
            data = self._terminal.inBuffer[:aCount-1]
            self._terminal.inBuffer = self._terminal.inBuffer[aCount-1:]
        self._terminal.lock.release()
        #print "KoTerminalInFile.read(): input is '%s'" % data
        return data


# classes specificly for use in the debugger

class KoDebuggerTerminal(KoTerminal):
    _com_interfaces_ = [components.interfaces.koIDebuggerTerminal, components.interfaces.ISciMozEvents]
    _reg_desc_ = "Komodo Debugger Terminal"
    _reg_contractid_ = "@activestate.com/KoDebuggerTerminal;1"
    _reg_clsid_ = "{F1E249E4-DB45-453b-BC83-2F8FDE7986A6}"
    
    def init(self, scintilla, viewName, scratchDocument, language, encoding_name):
        KoTerminal.init(self, scintilla, viewName, scratchDocument, language, encoding_name)
        self.stdin = KoMemoryIOFile(self._scintilla, self)
        self.emulateCGI = 0

    def set_acceptInput(self, acceptInput):
        if not self.emulateCGI:
            self.acceptInput = acceptInput

class KoMemoryIOFile(KoTerminalFile):
    def __init__(self, scintilla, terminal):
        KoTerminalFile.__init__(self, scintilla, terminal)
        self._readWait = None #block indefinitly for user input
        
    def read(self, aCount):
        # XXX logicaly, we should block on read if we are not running
        # cgi, however, that makes debugging output flaky, not sure why
        # but if we do not wait, everything works great.  Need to investigate
        # this logic much more later on.
        # readWait should be zero for if emulateCGI is false
        
        # if emulating CGI, don't block on waiting
        self._readWait = 1
        #if self._terminal.emulateCGI: self._readWait = 1

        # only wait if NO data is available to read!
        if len(self._terminal.inBuffer) < 1:
            #print "waiting for input ",self._readWait
            self._terminal.inputReady.acquire()
            self._terminal.inputReady.wait(self._readWait) 
            self._terminal.inputReady.release()

        #print "KoTerminalInFile.read(): have input, waiting for lock"
        self._terminal.lock.acquire()
        if aCount >= len(self._terminal.inBuffer):
            data = self._terminal.inBuffer
            self._terminal.inBuffer = ''
        else:
            data = self._terminal.inBuffer[:aCount-1]
            self._terminal.inBuffer = self._terminal.inBuffer[aCount-1:]
        self._terminal.lock.release()
        #print "KoTerminalInFile.read(): input size sent is '%d'" % len(data)
        return data
    
    def write(self, data):
        self._terminal.lock.acquire()
        self._terminal.inBuffer += data
        #print "size of data is ",len(data)
        #print "size of inbuffer is now ",len(self._terminal.inBuffer)
        self._terminal.lock.release()
        self._terminal.inputReady.acquire()
        self._terminal.inputReady.notify()
        self._terminal.inputReady.release()
    
