# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

import sys
import thread, threading
import urlparse
import convertURL

from xpcom import components, nsError, ServerException, Exception, _xpcom
from xpcom.server.enumerator import SimpleEnumerator
from koLogger import logError, logDebug, logWarning

# This base class handles an "asynch breakpoint".  If the engine
# can not currently accept breakpoints (as it is running or stopped)
# we mark them as "pending".  Our engine base class knows to
# automatically move breakpoints to unpending.
#
# all the child-class need do is implement the language
# name, and the code for moving the break-point info
# into the engine, and connection listeners, and IO
# redirection, and ... :)
class koBreakpointBase:
    _com_interfaces_ = components.interfaces.koIBreakpoint
    def __init__(self, engine, where, lineno):
        self.engine = engine
        self.uri = where

        filename = convertURL.URLToLocalFile(where)
        self.remoteURI = where
        if filename is not None:
            self.filename = filename.path
        elif where.startswith("kodebugger:"):
            uriobj = components.classes["@activestate.com/koDebuggerURI"].\
                       createInstance(components.interfaces.koIDebuggerURI)
            uriobj.init(where)
            self.filename = uriobj.path
            self.remoteURI = uriobj.getBaseURI(where)
        else:
            logError("Can only debug local files or kodebugger URLS - not '%s'" % (where,))
            raise ServerException(nsError.NS_ERROR_FAILURE)

        self.lineno = lineno
        self.state = components.interfaces.koIBreakpoint.BREAKPOINT_ENABLED
        self.pending = 1
    def get_file( self ):
        # Result: string
        return self.filename
    def get_lineNumber( self ):
        # Result: uint32
        return self.lineno
    def set_state( self, state ):
        if self.state == components.interfaces.koIBreakpoint.BREAKPOINT_DELETED and not self.pending:
            logError("You can't set the state on a deleted breakpoint")
            raise ServerException(nsError.NS_ERROR_FAILURE)
        self.pending = 1
        self.state = state
        self._tryUpdatePending()
    def get_state( self ):
        # Result: short
        return self.state
    def get_pending( self ):
        # read-only book
        return self.pending
    def _reset(self):
        self.state = components.interfaces.koIBreakpoint.BREAKPOINT_DELETED
        self.pending = 0
        self.engine = None
    def _tryUpdatePending(self):
        try:
            # This winds around to call _updatePending!
            self.engine._release_command( self.engine._processPendingBreakpoints )
        except ServerException:
            pass

    # The ones needed to be overwritten.
    def _updatePending(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
        # OK - valid breakpoint with some state -set it.
    def get_language(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

# This class is a good base for a "blocking", command-based debug
# engine.  It implements the correct blocking semantics for calling
# back to the engine when it is in the blocked (ie, at a breakpoint)
# state.
class koDebugEngineBase:
    def __init__(self):
        self.sync_request_lock = threading.RLock()
        self.sync_target_lock = threading.Lock()
        self.sync_request_event = threading.Event()
        self.sync_response_event = threading.Event()
        self.stack_frame_list = None
        self.cmdloop_thread_id = None
        self.state = components.interfaces.koIDebugger.DBG_STATE_STOPPED

        sm = _xpcom.GetGlobalServiceManager()
        self.komodo_debugger = components.classes["@activestate.com/debugger;1"] \
                               .getService(components.interfaces.koIDebuggingFramework)

        # May be added on any thread - may be otherwise processed
        # only on the debugging thread (as this requires communication
        # with the otherwise blocked engine.)
        self.breakpoints = []
        self.bplock = threading.Lock()

    def stop(self):
        try:
            self._release_command( None, components.interfaces.koIDebugger.RESUME_STOP )
        except ServerException:
            pass
        self._stop()

    def resume(self, how):
        return self._release_command( None, how )

    def enumBreakpoints(self, uri):
        try:
            self.bplock.acquire()
            if not uri:
                breakpoints = self.breakpoints
            else:
                breakpoints = [(bp) for bp in self.breakpoints if bp.uri.lower()==uri.lower()]
            return SimpleEnumerator(breakpoints)
        finally:
            self.bplock.release()

    def enumStackFrames(self):
        return self._release_command( self._enumStackFrames )

    def getSourceCode(self, filename):
        return self._release_command(self._getSourceCode, filename)

    def _refreshBreakpoints(self):
        # re-add breakpoints so they are entered into the new debug session
        try:
            self.bplock.acquire()
            breakpoints = self.breakpoints
            self.breakpoints = [] # new list doesn't affect old list above
        finally:
            self.bplock.release()

        # don't do in lock, addBreakpoint calls updatepending which locks
        for breakpoint in breakpoints:
            if breakpoint.state != \
                   components.interfaces.koIBreakpoint.BREAKPOINT_DELETED:
                self.addBreakpoint( breakpoint.uri,
                                    breakpoint.lineno)

    def get_state(self):
        return self.state
    # Internal helper implementations.
    def _set_state(self, state, state_extra): # Read-only - Not in the interface
        self.state = state
        self.stack_frame_list = None
        if state == components.interfaces.koIDebugger.DBG_STATE_STOPPED:
            # Every break-point gets set to pending.
            try:
                self.bplock.acquire()
                for bp in self.breakpoints:
                    bp.pending = 1
            finally:
                self.bplock.release()
        self.komodo_debugger.onStateChange( self, state, state_extra )

    # Helpers for thread synchronization.
    def _release_command(self, command, *args):
        self.sync_request_lock.acquire()
        # Now the lock is acquired, we know the state of the target will
        # be stable unless we change it (by triggering the event)
        try:
            if not self.sync_target_lock.locked():
                raise ServerException(nsError.NS_ERROR_UNEXPECTED)
            self.sync_request_request = command
            self.sync_request_args = args
            self.sync_request_event.set()
            self.sync_response_event.wait()
            self.sync_response_event.clear()
            if self.sync_response_exception is not None:
                raise self.sync_response_exception[0], self.sync_response_exception[1], self.sync_response_exception[2]
            return self.sync_response_result
        finally:
            self.sync_request_lock.release()

    def _processPendingBreakpoints(self):
        i = 0
        try:
            self.bplock.acquire()
            while i < len(self.breakpoints):
                bp = self.breakpoints[i]
                bp._updatePending()
                if bp.state == components.interfaces.koIBreakpoint.BREAKPOINT_DELETED:
                        del self.breakpoints[i]
                        continue
                i = i +1
        finally:
            self.bplock.release()

    def _runCommandLoop( self, breakReason ):
        # Result: int32
        # In: param0: int32
        self.sync_target_lock.acquire()
        self.cmdloop_thread_id = thread.get_ident()
        self._processPendingBreakpoints()

        rc = None
        try:
            #print "QQQ: Setting state to DBG_STATE_BREAK"
            self._set_state(components.interfaces.koIDebugger.DBG_STATE_BREAK, breakReason)
            # Start a callback loop
            while 1:
                self.sync_request_event.wait()
                self.sync_request_event.clear()
                # If we process an exit request, we can not signal the request
                # as complete until our target_lock is released, otherwise we
                # can deadlock if a second thread gets in for a request before
                # our exit request has actually got out and released the lock.
                release_response_at_exit = 0
                self.sync_response_exception = None
                try:
                    if self.sync_request_request is None:
                        rc = self.sync_response_result = self.sync_request_args[0]
                        release_response_at_exit = 1
                        break
                    try:
                        self.sync_response_result = apply(self.sync_request_request, self.sync_request_args)
                    except:
                        self.sync_response_exception = sys.exc_info()
                        self.sync_response_result = None
                finally:
                    if not release_response_at_exit:
                        self.sync_response_event.set()
        finally:
            self.cmdloop_thread_id = None
            self.sync_target_lock.release()

        # If we are exiting due to a stop request, only now can the lock be released.
        if release_response_at_exit:
            self.sync_response_event.set()

        self._set_state(components.interfaces.koIDebugger.DBG_STATE_RUNNING, rc)
        return rc

    # determine whether the installed language can debug, default always return true
    def _canDebug( self ):
        return 1
    
        
