"""This module provides functions to do end-of-line detection and manipulation"""

EOL_LF = 0
EOL_CR = 1
EOL_CRLF = 2
EOL_MIXED = 3
EOL_NOEOL = 4

eolMappings = {
    EOL_LF: '\n',
    EOL_CR: '\r',
    EOL_CRLF: '\r\n'
}

import sys

# 'newl' is the current newline delimiter for the current platform
if sys.platform.startswith('win'):
    newl = eolMappings[EOL_CRLF]
elif sys.platform.startswith('mac'):
    newl = eolMappings[EOL_CR]
else:
    newl = eolMappings[EOL_LF]
    
def detectEOLFormat(buffer):
    """detectEOLFormat('test\r\n\n\n') => (EOL_MIXED, EOL_LF)
    
    Return a tuple containing a code for the detected end-of-line format and
    a code for the suggested end-of-line format to use for the given buffer,
    which may be different if the end-of-line terminator was ambiguous. If the
    buffer does not contain any end-of-lines, the return value will be
    (EOL_NOEOL, -1)"""
    
    crlf = buffer.count("\r\n")
    cr = buffer.count("\r") - crlf
    lf = buffer.count("\n") - crlf

    if crlf == lf == cr == 0:
        return (EOL_NOEOL, -1)

    eols = [(crlf, EOL_CRLF), (cr, EOL_CR), (lf, EOL_LF)]
    eols.sort()

    if eols[0][0] or eols[1][0]:
        return (EOL_MIXED, eols[2][1])
    else:
        return (eols[2][1], eols[2][1])
    

def convertToEOLFormat(buffer, format):
    """convertToEOLFormat("Test\r\n", EOL_LF) => "Test\n"

    Return the specified buffer with the end-of-line format
    converted to the one specified."""
    
    import re
    return re.sub( '\r\n|\r|\n', eolMappings[format], buffer )        

def test():
    assert detectEOLFormat('\r') == (EOL_CR, EOL_CR)
    assert detectEOLFormat('\r\n') == (EOL_CRLF, EOL_CRLF)
    assert detectEOLFormat('\n') == (EOL_LF, EOL_LF)
    assert detectEOLFormat('') == (EOL_NOEOL, -1)
    assert detectEOLFormat('\rTest\r\n\r') == (EOL_MIXED, EOL_CR)

    assert convertToEOLFormat('\rTest\r\nTest2\nTest\n\r\n',EOL_LF) == '\nTest\nTest2\nTest\n\n'
    assert convertToEOLFormat('\r\r\r\nTest\n\nTest2\nTest\n\n\n',EOL_CRLF) == '\r\n\r\n\r\nTest\r\n\r\nTest2\r\nTest\r\n\r\n\r\n'

if __name__ == '__main__':
    test() 