# Copyright (c) 2000 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

# Code that allows Komodo to pretend it is IDLE
# (at least as far as most IDLE extensions are concerned)
# This includes a Tk text widget emulation for Scintilla.
#
# Forked from Pythonwin's IDLEEnvironment.py

import string
import sys

wordchars = string.uppercase + string.lowercase + string.digits
# XXX this may have been fixed another way with setting default
# encoding in koinitsrvc.
# wordchars may contain Unicode characters, and this is just too confusing for us ATM :)
wordchars = filter(lambda c: ord(c)<128, wordchars)

from xpcom import components

all_string_styles = (components.interfaces.ISciMoz.SCE_P_STRING,
					 components.interfaces.ISciMoz.SCE_P_CHARACTER,
					 components.interfaces.ISciMoz.SCE_P_TRIPLE,
					 components.interfaces.ISciMoz.SCE_P_TRIPLEDOUBLE,
					 components.interfaces.ISciMoz.SCE_P_STRINGEOL,
					 )

class TextError(Exception): # When a TclError would normally be raised.
	pass

class EmptyRange(Exception): # Internally raised.
	pass

def GetIDLEModule(module):
	# We keep out idle derived extensions in a package
	modname = "idle." + module
	__import__(modname)
	return sys.modules[modname]

# A class that is injected into the IDLE auto-indent extension.
# It allows for decent performance when opening a new file,
# as auto-indent uses the tokenizer module to determine indents.
# The default AutoIndent readline method works OK, but it goes through
# this layer of Tk index indirection for every single line.  For large files
# without indents (and even small files with indents :-) it was pretty slow!
def fast_readline(self):
	if self.finished:
		return ""
	if not self.__dict__.has_key("_scint_lines"):
		# XXX - note - assumes this is only called once the file is loaded!
		self._scint_lines = string.split(self.text.edit.getTextRange(0, self.text.edit.textLength-1), "\n")
		self._scint_lines.append('') # Pretend an extra newline
	sl = self._scint_lines
	i = self.i = self.i + 1
	if i >= len(sl):
		return ""
	return sl[i]+"\n"

try:
	GetIDLEModule("AutoIndent").IndentSearcher.readline = fast_readline
except ImportError:
	pass

# A class that attempts to emulate an IDLE editor window.
# Construct with an ISciMoz interface.
class IDLEEditorWindow:
	def __init__(self, edit, extensionUI):
		self.edit = edit
		self.extensionUI = extensionUI
		self.text = TkText(edit)
		self.extensions = {}
		self.extension_menus = {}

	def close(self):
		self.edit = self.text = None
		try:
			for ext in self.extensions.values():
				closer = getattr(ext, "close", None)
				if closer is not None:
					closer()
		finally:
			self.extensions = {}

	def IDLEExtension(self, extension):
		ext = self.extensions.get(extension)
		if ext is not None: return ext
		mod = GetIDLEModule(extension)
		if mod is None: return None
		klass = getattr(mod, extension)
		ext = self.extensions[extension] = klass(self)
		return ext

	######################################################################
	# The IDLE "Virtual UI" methods that are exposed to the IDLE extensions.
	#
	def askinteger(self, caption, prompt, parent=None, initialvalue=0, minvalue=None, maxvalue=None):
		if minvalue is None: minvalue = -sys.maxint
		if maxvalue is None: maxvalue = sys.maxint
		ok, newvalue = self.extensionUI.askInteger(caption, prompt, minvalue, maxvalue, initialvalue)
		if not ok:
			raise KeyboardInterrupt
		return newvalue
	def askyesno(self, caption, prompt, parent=None):
		return self.extensionUI.askYesNo(caption, prompt)

	######################################################################
	# The IDLE "Virtual Text Widget" methods that are exposed to the IDLE extensions.
	#

	# Is character at text_index in a Python string?  Return 0 for
	# "guaranteed no", true for anything else.
	def is_char_in_string(self, text_index):
		# A helper for the code analyser - we need internal knowledge of
		# the colorizer to get this information
		text_index = self.text._getoffset(text_index)
		assert text_index <= self.edit.endStyled, "This assumes the colorizer has got to this point!"
		rc = self.edit.getStyleAt(text_index) in all_string_styles
		return rc

	# If a selection is defined in the text widget, return
	# (start, end) as Tkinter text indices, otherwise return
	# (None, None)
	def get_selection_indices(self):
		try:
			first = self.text.index("sel.first")
			last = self.text.index("sel.last")
			return first, last
		except TextError:
			return None, None

	def set_tabwidth(self, width ):
		self.edit.tabWidth = width

	def get_tabwidth(self):
		return self.edit.tabWidth

# A class providing the generic "Call Tips" interface
class CallTips:
	def __init__(self, edit):
		self.edit = edit
	def showtip(self, tip_text):
		self.edit.callTipShow(self.edit.selectionStart, tip_text)
	def hidetip(self):
		self.edit.callTipCancel()

########################################
#
# Helpers for the TkText emulation.
def TkOffsetToIndex(offset, edit):
	lineoff = 0
	# May be 1 > actual end if we pretended there was a trailing '\n'
	offset = min(offset, edit.textLength)
	line = edit.lineFromPosition(offset)
	lineIndex = edit.positionFromLine(line)
	return "%d.%d" % (line+1, offset-lineIndex)

def _NextTok(str, pos):
	# Returns (token, endPos)
	end = len(str)
	if pos>=end: return None, 0
	while pos < end and str[pos] in string.whitespace:
		pos = pos + 1
	# Special case for +-
	if str[pos] in '+-':
		return str[pos],pos+1
	# Digits also a special case.
	endPos = pos
	while endPos < end and str[endPos] in string.digits+".":
		endPos = endPos + 1
	if pos!=endPos: return str[pos:endPos], endPos
	endPos = pos
	while endPos < end and str[endPos] not in string.whitespace + string.digits + "+-":
		endPos = endPos + 1
	if pos!=endPos: return str[pos:endPos], endPos
	return None, 0

def get_character_start_pos(edit, pos):
	# Given a random position in a UTF8 buffer,
	# return the position of the first byte of the character.
	# (ripped from npscimoz.cxx)
	byte = 0xFF & edit.getCharAt(pos)
	if (byte < 0x80):
		#Characters in the ASCII charset.
		# Also treats \0 as a valid characters representing itself.
		return pos
	while byte < 0xC0 and byte >= 0x80 and pos > 0:
		#* Naked trail byte.  We asked for an index in the middle of
		#* a UTF-8 char sequence.  Back up to the beginning.  We should
		#* end up with a start byte >= 0xC0 and <= 0xFD, but check against
		#* 0x80 still in case we have a screwy buffer.
		pos -= 1
		byte = 0xFF & edit.getCharAt(pos)
	return pos

def get_character_length(edit, pos):
	# Given a start position of a char in a UTF8 buffer,
	# return how many bytes make up the character.
	length = 1
	textLength = edit.textLength
	getCharAt = edit.getCharAt
	assert pos == get_character_start_pos(edit, pos), "get_character_length, but not at the start of the character"
	# count to next sync byte.
	length = 1
	while 1:
		if pos == textLength:
			break
		pos += 1
		byte = getCharAt(pos) & 0xFF
		if byte < 0x80 or byte >= 0xC0:
			break
		length += 1
	return length

def TkIndexToOffset(bm, edit, marks):
	base, nextTokPos = _NextTok(bm, 0)
	if base is None: raise ValueError, "Empty bookmark ID!"
	if string.find(base,".")>0:
		try:
			line, col = string.split(base, ".", 2)
			if col=="first" or col=="last":
				# Tag name
				if line != "sel": raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
				start = edit.selectionStart
				end = edit.selectionEnd
				if start == end:
					raise EmptyRange
				if col=="first":
					pos = start
				else:
					pos = end
			else:
				# Lines are 1 based for tkinter
				line = int(line)-1
				if line > edit.lineCount:
					pos = edit.textLength+1
				else:
					pos = edit.positionFromLine(line)
					if pos==-1: pos = edit.textLength
					pos = pos + int(col)
		except (ValueError, IndexError):
			raise ValueError, "Unexpected literal in '%s'" % base 
	elif base == 'insert':
		pos = edit.selectionStart
	elif base=='end':
		pos = edit.textLength
		# Pretend there is a trailing '\n' if necessary
		if pos and edit.getWCharAt(pos-1) != "\n":
			pos = pos+1
	else:
		try:
			pos = marks[base]
		except KeyError:
			raise ValueError, "Unsupported base offset or undefined mark '%s'" % base

	while 1:
		word, nextTokPos = _NextTok(bm, nextTokPos)
		if word is None: break
		if word in ['+','-']:
			num, nextTokPos = _NextTok(bm, nextTokPos)
			if num is None: raise ValueError, "+/- operator needs 2 args"
			what, nextTokPos = _NextTok(bm, nextTokPos)
			if what is None: raise ValueError, "+/- operator needs 2 args"
			if what[0] <> "c": raise ValueError, "+/- only supports chars"
			# Correctly handling UTF8 chars is expensive - ie, for
			# "1.0+2000c" (not too uncommon), it would require us
			# looping 2000 times, checking each character for its length.
			# The extensions which use this code are not that sophisticaed WRT
			# Unicode anyway, so there is usually no need for this correctness
			# (indeed, this may actually be *incorrect*, as the extensions may
			# be passing, eg, the cursor position, which is a byte index anyway.
			# SO - the best we do is ensure what whatever is calculated, it is
			# never 1/2 way in a UTF8 sequence.
			if word=='+':
				pos += int(num)
				# If 1/2 way in a sequence, move to the next character.
				while get_character_start_pos(edit, pos) != pos:
					pos += 1
			else:
				pos -= int(num)
				# If half way through, move to start of character.
				pos = get_character_start_pos(edit, pos)
		elif word=='wordstart':
			while pos > 0 and edit.getWCharAt(pos-1) in wordchars:
				pos = pos - 1
		elif word=='wordend':
			end = edit.textLength
			while pos < end and edit.getWCharAt(pos) in wordchars:
				pos = pos + 1
		elif word=='linestart':
			while pos > 0 and edit.getWCharAt(pos-1) not in '\n\r':
				pos = pos - 1
		elif word=='lineend':
			end = edit.textLength
			while pos < end and edit.getWCharAt(pos) not in '\n\r':
				pos = pos + 1
		else:
			raise ValueError, "Unsupported relative offset '%s'" % word
	return pos
#	return max(pos, 0) # Tkinter is tollerant of -ve indexes.

# A class that resembles an IDLE (ie, a Tk) text widget.
# Construct with an edit object (eg, an editor view)
class TkText:
	def __init__(self, edit):
		self.calltips = None
		self.edit = edit
		self.marks = {}
	def make_calltip_window(self):
		if self.calltips is None:
			self.calltips = CallTips(self.edit)
		return self.calltips
	def _getoffset(self, index):
		return TkIndexToOffset(index, self.edit, self.marks)
	def _getindex(self, off):
		return TkOffsetToIndex(off, self.edit)
	def _fix_eol_indexes(self, start, end):
		if start>0 and self.edit.getWCharAt(start)=='\n' and self.edit.getWCharAt(start-1)=='\r':
			start = start - 1
		if end < self.edit.textLength and self.edit.getWCharAt(end-1)=='\r' and self.edit.getWCharAt(end)=='\n':
			end = end + 1
		return start, end
	def bind(self, binding, handler):
		raise RuntimeError, "We don't support bindings"
	def get(self, start, end = None):
		try:
			start = self._getoffset(start)
			if end is None:
				end = start + get_character_length(self.edit, start)
			else:
				end = self._getoffset(end)
		except EmptyRange:
			return ""
		# Simple semantic checks to conform to the Tk text interface
		if end <= start: return ""
		max = self.edit.textLength
		checkEnd = 0
		if end > max:
			end = max
			checkEnd = 1
		start, end = self._fix_eol_indexes(start, end)
		ret = self.edit.getTextRange(start, end)
		# pretend a trailing '\n' exists if necessary.
		if checkEnd and (not ret or ret[-1] != '\n'): ret = ret + '\n'
		return string.replace(ret, "\r", "")
	def index(self, spec):
		try:
			return self._getindex(self._getoffset(spec))
		except EmptyRange:
			return ""
	def insert(self, pos, text):
		try:
			pos = self._getoffset(pos)
		except EmptyRange:
			raise TextError, "Empty range"
		self.edit.selectionStart = pos
		self.edit.selectionEnd = pos
		# IDLE only deals with "\n" - we will be nicer
		bits = string.split(text, '\n')
		self.edit.addText(len(bits[0]), bits[0])
		for bit in bits[1:]:
			self.edit.newLine()
			self.edit.addText(len(bit), bit)

	def delete(self, start, end=None):
		try:
			start = self._getoffset(start)
			if end is not None: end = self._getoffset(end)
		except EmptyRange:
			raise TextError, "Empty range"
		# If end is specified and == start, then we must delete nothing.
		if start==end: return
		# If end is not specified, delete one char
		if end is None:
			end = start + get_character_length(self.edit, start)
		else:
			# Tk says not to delete in this case, but our control would.
			if end<start: return
		if start==self.edit.textLength: return # Nothing to delete.
		old = self.edit.selectionStart # Lose a selection
		# Hack for partial '\r\n' removal
		start, end = self._fix_eol_indexes(start, end)
		self.edit.selectionStart = start
		self.edit.selectionEnd = end
		self.edit.clear()
		if old>=start and old<end:
			old=start
		elif old>=end:
			old = old - (end-start)
		self.edit.selectionStart = old
		self.edit.selectionEnd = old

	def bell(self):
		beeper = components.classes["@mozilla.org/sound;1"] \
				 .createInstance().queryInterface(components.interfaces.nsISound)
		beeper.beep()

	def see(self, pos):
		# Most commands we use in Scintilla actually force the selection
		# to be seen, making this unnecessary.
		pass

	def mark_set(self, name, pos):
		try:
			pos = self._getoffset(pos)
		except EmptyRange:
			raise TextError, "Empty range '%s'" % pos
		if name == "insert":
			self.edit.selectionStart = pos
			self.edit.selectionEnd = pos
		else:
			self.marks[name]=pos

	def tag_add(self, name, start, end):
		if name != "sel": raise ValueError, "Only sel tag is supported"
		try:
			start = self._getoffset(start)
			end = self._getoffset(end)
		except EmptyRange:
			raise TextError, "Empty range"
		self.edit.selectionStart = start
		self.edit.selectionEnd = end

	def tag_remove(self, name, start, end):
		if name !="sel" or start != "1.0" or end != "end":
			raise ValueError, "Cant remove this tag"
		# Turn the sel into a cursor
		self.edit.selectionEnd = self.edit.selectionStart

	def compare(self, i1, op, i2):
		try:
			i1=self._getoffset(i1)
		except EmptyRange:
			i1 = ""
		try:
			i2=self._getoffset(i2)
		except EmptyRange:
			i2 = ""
		return eval("%d%s%d" % (i1,op,i2))

	def undo_block_start(self):
		self.edit.beginUndoAction()

	def undo_block_stop(self):
		self.edit.endUndoAction()

# Pythonwin has some test code that should be adopted!
