# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

import re
from xpcom import components
from koLogger import logError, logWarning, logDebug
sci_constants = components.interfaces.ISciMoz

# NOTE: TabWidth isn't strictly relevant to this code - but is still
# maintained here for consistency.
def _makemask(bits):
    stylingBitsMask = 0    
    for bit in range(bits):
        stylingBitsMask <<= 1
        stylingBitsMask |= 1
    return stylingBitsMask

class CommandHandler:
    _com_interfaces_ = components.interfaces.nsIController
    def __init__(self, wrapper):
        self.wrapper = wrapper
        self.languageObj = wrapper.languageOb
        self.prefs = wrapper.viewPrefs
        self.guessedIndents = 0 
        self.scimoz = wrapper._scintilla
        self.stylingBitsMask = _makemask(self.languageObj.styleBits)
        self._loadIndentSettings()
        import idle # XXX Try to get rid of this if at all possible!
        self.idlewindow = idle.IDLEEditorWindow(self.scimoz, wrapper.extensionUI)
        self.idleIndentExt = self.idlewindow.IDLEExtension("AutoIndent")
        self.doBraceMatching = 0

    def __del__(self):
        self.idleIndentExt = None
        self.idlewindow.close()

    def _loadIndentSettings(self):
        # Set indentation to default for the file type.
        self.indentWidth = self.prefs.getLongPref("indentWidth")
        self.useTabs = self.prefs.getBooleanPref("useTabs")
        self.tabWidth = self.prefs.getLongPref("tabWidth")
        self.scimoz.tabWidth = self.tabWidth
        # Flag that we will allow guessing to override defaults...
        self.smartTabs = self.prefs.getBooleanPref("useSmartTabs")

    def configureIndentationParams(self):
        # Tell scintilla the indent width, as we fall back to it in
        # really bad situations.
        self.scimoz.indent = self.indentWidth

    def applyIndentationParams(self):
        if self.guessedIndents or not self.smartTabs:
            # User doesnt want smart tabs, or we have already guessed - use
            # existing settings.
            return

        if self.scimoz.textLength == 0:
            data = None
        else:
            data = self._guessFileIndentation()
        # data == None when we cannot guess for some reason,
        # like empty file or no lines with leading white-space
        if data is not None:
            # Set the setting via the preferences - this will ensure that
            # "Current File Preferences" shows our newly guessed state.  It
            # will also force a notification, and cause the notification to
            # set the actual value, back via our editor parent.  These
            # indent settings are for this file, not the global settings.
            indentWidth, useTabs = data

            self.prefs.setLongPref("indentWidth", indentWidth)
            self.prefs.setBooleanPref("useTabs", useTabs)
            assert useTabs == self.useTabs and \
                   self.indentWidth == indentWidth, \
                   "The data didn't get reflected back into me!"
            #logDebug("useTabs %s indentWidth %s tabWidth %s" % \
            #         (useTabs, self.indentWidth, self.tabWidth));

        self.guessedIndents = 1
        self.smartTabs = 0 # no longer need to be smart about this.
        self.prefs.setBooleanPref("useSmartTabs", 0)
        self.configureIndentationParams()

    def isCommandEnabled( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        meth = getattr(self, "_is_%s_enabled" % (str(command_name),), None)
        if meth is None:
            return 1
        return meth()

    def supportsCommand( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        rc = hasattr(self, "_do_" + command_name)
        #print "koEditorController - SupportsCommand '%s' -> %d" % (command_name, rc)
        return rc

    def doCommand( self, command_name ):
        # Result: void - None
        # In: param0: wstring
        meth = getattr(self, "_do_" + command_name)
        try:
            # Some CommandHandler's do not have a wrapper, e.g. TerminalCommandHandler
            if self.wrapper is not None:
                self.wrapper.beginBatchOperation()
            try:
                ret = meth()
            finally:
                if self.wrapper is not None:
                    self.wrapper.endBatchOperation()
        except KeyboardInterrupt: # User cancelled a dialog.
            pass

    def onEvent( self, param0 ):
        # Result: void - None
        # In: param0: wstring
        # print "koScintillaController::onEvent", param0
        pass

    def keyPressed(self, char):
        pass

# A class that provides "generic" text capabilities.
# Other controllers may derive from this if they only
# want to provide certain functionality.
# Although this is a generic handler, we _do_ make use of the
# Python "AutoIndent" extension - just not full use.
# Features such as "tabify" are pretty generic.
class TextCommandHandler(CommandHandler):

    # Constants
    INDENT_BRACE = 1 
    LINEUP_BRACE = 2

    def __init__(self, braceData, *args):
        CommandHandler.__init__(self, *args)
        self.braceData = braceData
        # Need to ensure that we configure the indentation params,
        # because self.applyIndentationParams only does that when
        # smartTabs is on.
        self.configureIndentationParams()

    def _guessFileIndentation(self):
        # Heuristic to determine what file indentation settings the user
        # likely wants for this file.
        usesTabs = usesSpaces = linesChecked = 0
        # In the first 150 lines of the file, search for the non-blank
        # lines with leading white-space.  Searching farther takes too long.
        for lineNo in xrange(min(self.scimoz.lineCount, 150)):
            lineLen, thisLine = self.scimoz.getLine(lineNo)
            thisLine = thisLine[:lineLen].rstrip()
            if thisLine[:1] == "\t":
                # If first char is a tab, recognize that and move on
                linesChecked += 1
                usesTabs = 1
            elif thisLine[:2] == "  ":
                # If first 2 chars are spaces, recognize that and move on
                # Require at least two spaces on the line to count it
                linesChecked += 1
                usesSpaces = 1
            if linesChecked == 25:
                # Only check up to 25 lines with indentation
                break
        if linesChecked:
            # We found some lines with indentation
            if usesTabs and usesSpaces:
                # If we found both space and tab indentation, leave the
                # indentWidth setting as default, but do use tabs where
                # appropriate
                return self.indentWidth, 1
            elif usesTabs:
                # If only tab indentation was found, set the indentWidth
                # to the tabWidth, so we essentially always use tabs.
                return self.tabWidth, 1
            else:
                # Otherwise we won't use tabs at all.
                return self.indentWidth, 0
        return None

    def configureIndentationParams(self):
        #print "configureIndentationParams: useTabs", self.useTabs, \
        #      "indentwidth", self.indentWidth, "tabWidth", self.tabWidth
        CommandHandler.configureIndentationParams(self)
        self.scimoz.tabWidth = self.tabWidth
        self.idleIndentExt.config(usetabs=self.useTabs, indentwidth=self.indentWidth, tabwidth=self.tabWidth)

    def _do_cmd_cancel(self):
        # Cancel command completion or autocompletion.
        sm = self.scimoz
        if sm.callTipActive():
            sm.callTipCancel()
        if sm.autoCActive():
            sm.autoCCancel()

    def _do_cmd_newline(self):
        # koEditorWrapper.onModified needs to work when we hit enter, so we
        # have to cancel the batch operation mode.
        # kind of hacky to have this here, but no better way right now
        autoIndentStyle = self.prefs.getStringPref('editAutoIndentStyle')
        self._do_newline(autoIndentStyle)

    def _do_newline(self, style):
        # style is the value of the 'editAutoIndentStyle' pref - smart, plain, or none.
        if self.wrapper is not None:
            self.wrapper.endBatchOperation()
        if style == 'none':
            self.scimoz.newLine()
            return
        self.scimoz.beginUndoAction()
        try:
            # it's important not to use self._guessIndentation here
            # in the case where _do_newline is bein called by the
            # newline handler in the BraceCommandHandler, as a result
            # of the 'indent to current indent level' pref.
            indent = TextCommandHandler._guessIndentation(self, self.scimoz.currentPos)
            # need to remove whitespace immediately after the current character
            # iff that's all that is to the right of the current character.
            endOfLine = self.scimoz.positionFromLine(self.scimoz.lineFromPosition(self.scimoz.currentPos)+1)
            thisline = self.scimoz.getTextRange(self.scimoz.currentPos, endOfLine)
            if not thisline.strip():
                while self.scimoz.getWCharAt(self.scimoz.currentPos) in ' \t':
                    self.scimoz.clear()
            self.scimoz.newLine()
            numtabs = indent.count('\t')
            numspaces = indent.count(' ')
            numspaces += numtabs * self.tabWidth
            if self.useTabs:
                numtabs, numspaces = divmod(numspaces, self.tabWidth)
                indent = '\t'*numtabs + ' '*numspaces
            self.scimoz.addText(len(indent), indent)
        finally:
            self.scimoz.endUndoAction()

    def _guessIndentation(self, index):
        scimoz = self.scimoz
        lineNumber = scimoz.lineFromPosition(index)
        startIndex = scimoz.positionFromLine(lineNumber)
        previousIndent = "" # the tabs and spaces preceding the previous line.
        # Starts from the beginning of the previous line until it finds the
        # first non whitespace character. Records the indent of this line
        # to use as a guide for the indent of the next line.
        for i in range(startIndex, index):
            ch = scimoz.getWCharAt(i)
            if ch != ' ' and ch != '\t':
                break
            previousIndent += ch
        return previousIndent

    def _do_cmd_indent(self):
        # This code is fairly complex, as it deals with many situations.
        # If you modify this code, please make sure you understand:
        # The cases being handled:
        #    - No selection, cursor in the middle of non-whitespace chars
        #    - No selection, cursor with only whitespace (or nothing) to its left
        #    - Selection spanning part of a line with only whitespace to its left
        #      - The above both with the left edge of the selection aligned with
        #        indentation lines or not.
        #    - Selection spanning an entire line (excluding EOL markers)
        #    - selection spanning multiple lines
        #    - Dealing with tabs and spaces
        
        scimoz = self.scimoz

        # Do tab autocompletion, assuming it's in progress.
        if scimoz.autoCActive():
            scimoz.autoCComplete()
            return

        scimoz.beginUndoAction()
        try:
            # "smart tab" behavior is somewhat tricky.
            # 1) If more than one line is selected, it's a region indent.
            oldAnchor = scimoz.anchor
            currentLineNo = scimoz.lineFromPosition(scimoz.currentPos)
            lastLineNo = scimoz.lineFromPosition(scimoz.anchor)
            # stash away old selection information
            oldCurPos = scimoz.currentPos
            oldAnchorPos = scimoz.anchor
            oldCurLine, oldCurCol = currentLineNo, scimoz.getColumn(scimoz.currentPos)
            oldAnchorLine, oldAnchorCol = lastLineNo, scimoz.getColumn(scimoz.anchor)
    
            currentLineNo, lastLineNo = min(currentLineNo, lastLineNo), max(currentLineNo, lastLineNo)
            # need to detect if we've selected a whole line
            a, b = min(scimoz.currentPos, scimoz.anchor), max(scimoz.currentPos, scimoz.anchor)
            # How many lines do we need to iterate over depends on whether we're at the "end" of the last line.
            atBeginningOfNextLine = 0
            if scimoz.getColumn(b) == 0:
                lastLineNo -= 1
                atBeginningOfNextLine = 1
            if (scimoz.getColumn(a) == 0) and (scimoz.getColumn(b) == 0 or b == scimoz.getLineEndPosition(lastLineNo)):
                wholeLine = a != b
            else:
                wholeLine = 0
            processLastLine = (atBeginningOfNextLine == 0 and wholeLine == 0)
            if a != b and (currentLineNo != lastLineNo or wholeLine):
                oldPos = scimoz.currentPos
                effectiveIndent = None
                for lineNo in range(currentLineNo, lastLineNo + (wholeLine or processLastLine)):
                    self.scimoz.currentPos = self.scimoz.positionFromLine(lineNo)
                    effectiveIndent = self._indentLine(lineNo, effectiveIndent)
                scimoz.currentPos = oldPos
                # restore selection
                if (oldCurPos < oldAnchorPos):
                    scimoz.currentPos = scimoz.positionFromLine(currentLineNo)
                    if atBeginningOfNextLine:
                        scimoz.anchor = scimoz.positionFromLine(lastLineNo+1)
                    else:
                        scimoz.anchor = scimoz.getLineEndPosition(lastLineNo)
                else:
                    if atBeginningOfNextLine:
                        scimoz.currentPos = scimoz.positionFromLine(lastLineNo+1)
                    else:
                        scimoz.currentPos = scimoz.getLineEndPosition(lastLineNo)
                    scimoz.anchor = scimoz.positionFromLine(currentLineNo)
            else:
                # If there's a selection, delete the selection
                scimoz.replaceSel('')
                self._indentLine(currentLineNo, None)
            self._DoTheSciMozWiggle()
        finally:
            scimoz.endUndoAction()
        
    def _DoTheSciMozWiggle(self):
        scimoz = self.scimoz
        # this is the wiggle needed to set lastX
        c = scimoz.currentPos
        a = scimoz.anchor
        scimoz.currentPos = c
        scimoz.anchor = c
        if scimoz.currentPos > 0:
            scimoz.charLeft()
            scimoz.charRight()
        else:
            scimoz.charRight()
            scimoz.charLeft()
        scimoz.currentPos = c
        scimoz.anchor = a

    def _indentLine(self, lineNo, forcedIndent):
        # if forcedIndent is not None, then it must be used as the indent _and_ returned
        # as the second argument.  if forcedIndent is None, then the function must compute
        # the "effective" indent that it applies to the line and return _that_ to be used
        # for subsequent lines in a multi-line indent.
        scimoz = self.scimoz
        startofLinePos = scimoz.positionFromLine(lineNo)
        endofLinePos = scimoz.getLineEndPosition(lineNo)
        wholeLine = scimoz.getTextRange(startofLinePos, endofLinePos)
        lineLength = scimoz.lineLength(lineNo)
        positioninLine, buffer = scimoz.getCurLine()
        if wholeLine[:positioninLine].strip() != '':
            if self.useTabs:
                scimoz.replaceSel('\t')
            else:
                scimoz.replaceSel(' '*self.indentWidth)
            return
        firstNonWhitespace = re.search(r'\S', wholeLine)
        if (firstNonWhitespace):
            index = firstNonWhitespace.start()
        else:
            index = len(wholeLine)
        scimoz.currentPos = startofLinePos + index
        scimoz.anchor = startofLinePos
        newwhitespace, effectiveIndent = self._analyzeAndIndent(wholeLine[:index], forcedIndent)
        scimoz.replaceSel(newwhitespace)
        return effectiveIndent

    def _indentFromColumn(self, numspaces):
        if self.useTabs:
            numnewtabs, numspaces = divmod(numspaces, self.tabWidth)
            return '\t'*numnewtabs + ' '*numspaces
        else:
            return ' '*numspaces
        
    def _analyzeAndIndent(self, whitespaceline, forcedIndent):
        # if forcedIndent is not None, then it must be used as the indent _and_ returned
        # as the second argument.  if forcedIndent is None, then the function must compute
        # the "effective" indent that it applies to the line and return _that_ to be used
        # for subsequent lines in a multi-line indent.
        orignumtabs = numtabs = whitespaceline.count('\t')
        orignumspaces = numspaces = whitespaceline.count(' ')
        newline = ''
        numspaces += numtabs * self.tabWidth
        if forcedIndent is None:
            numindents, numspaces = divmod(numspaces, self.indentWidth)
            numindents += 1
            numspaces = self.indentWidth * numindents
        else:
            numspaces += forcedIndent
        if self.useTabs:
            numnewtabs, numspaces = divmod(numspaces, self.tabWidth)
            if forcedIndent is None:
                returnedIndent = self.tabWidth * (numnewtabs-numtabs) + (numspaces - orignumspaces)
            else:
                returnedIndent = forcedIndent
            return '\t'*numnewtabs + ' '*numspaces, returnedIndent
        else:
            if forcedIndent is None:
                returnedIndent = numspaces - orignumspaces 
            else:
                returnedIndent = forcedIndent
            return ' '*numspaces, returnedIndent

    def _do_cmd_dedent(self):
        scimoz = self.scimoz
        scimoz.beginUndoAction()
        try:
            # If no selection is set, we don't want a selection afterwards.
            if scimoz.anchor == scimoz.currentPos:
                oldCol, thisLine = self.scimoz.getCurLine()
                if oldCol != 0 and thisLine[0:oldCol].strip() == '':
                    # We're in in the "leading whitespace" region -- we want to end
                    # at the first non-whitespace character.
                    match = re.search(r'\S', thisLine)
                    if (match):
                        oldCol = match.start()
                    else:
                        oldCol = 0
            else:
                oldCol = 0
    
            oldCurPos = scimoz.currentPos
            oldAnchorPos = scimoz.anchor
    
            currentLineNo = scimoz.lineFromPosition(scimoz.currentPos)
            lastLineNo = scimoz.lineFromPosition(scimoz.anchor)
            currentLineNo, lastLineNo = min(currentLineNo, lastLineNo), max(currentLineNo, lastLineNo)
            # need to detect if we've selected a whole line
            a, b = min(scimoz.currentPos, scimoz.anchor), max(scimoz.currentPos, scimoz.anchor)
            # How many lines do we need to iterate over depends on whether we're at the "end" of the last line.
            atBeginningOfNextLine = 0
            if scimoz.getColumn(b) == 0:
                if a != b:
                    lastLineNo -= 1
                atBeginningOfNextLine = 1
            if (scimoz.getColumn(a) == 0) and (scimoz.getColumn(b) == 0 or b == scimoz.getLineEndPosition(lastLineNo)):
                wholeLine = a != b
            else:
                wholeLine = 0
            processLastLine = (atBeginningOfNextLine == 0 and wholeLine == 0)
            if (currentLineNo != lastLineNo or wholeLine):
                oldPos = scimoz.currentPos
                effectiveIndent = None
                for lineNo in range(currentLineNo, lastLineNo + (wholeLine or processLastLine)):
                    self.scimoz.currentPos = self.scimoz.positionFromLine(lineNo)
                    self._dedentLine(lineNo)
                scimoz.currentPos = oldPos
                # restore selection
                if (oldCurPos < oldAnchorPos):
                    scimoz.currentPos = scimoz.positionFromLine(currentLineNo)
                    scimoz.anchor = scimoz.getLineEndPosition(lastLineNo)
                else:
                    scimoz.currentPos = scimoz.getLineEndPosition(lastLineNo)
                    scimoz.anchor = scimoz.positionFromLine(currentLineNo)
            else:
                effectiveDedent = self._dedentLine(currentLineNo)
                if oldCol:
                    newCol = (oldCol - effectiveDedent)
                else:
                    newCol = 0
                self.scimoz.anchor = self.scimoz.currentPos = scimoz.positionFromLine(currentLineNo) + newCol
            self._DoTheSciMozWiggle()
        finally:
            scimoz.endUndoAction()
        

        # Won't try and guess exactly where to put the cursor, as mixed tabs 
        # and spaces may confuse us.
        #scimoz = self.scimoz
        #if scimoz.selectionStart == scimoz.selectionEnd:
        #    lineRestore = scimoz.lineFromPosition(self.scimoz.selectionStart)
        #else:
        #    lineRestore = None
        #self.idleIndentExt.dedent_region_event(None)
        #if lineRestore is not None:
        #    scimoz.selectionStart = scimoz.selectionEnd = \
        #        scimoz.positionFromLine(lineRestore)

    def _dedentLine(self, lineNo):
        scimoz = self.scimoz
        startofLinePos = scimoz.positionFromLine(lineNo)
        endofLinePos = scimoz.getLineEndPosition(lineNo)
        wholeLine = scimoz.getTextRange(startofLinePos, endofLinePos)
        lineLength = scimoz.lineLength(lineNo)
        positioninLine, buffer = scimoz.getCurLine()
        firstNonWhitespace = re.search(r'\S', wholeLine)
        if (firstNonWhitespace):
            index = firstNonWhitespace.start()
        else:
            index = len(wholeLine)
        scimoz.currentPos = startofLinePos + index
        scimoz.anchor = startofLinePos
        newwhitespace, effectiveDedent = self._analyzeAndDedent(wholeLine[:index])
        scimoz.replaceSel(newwhitespace)
        return effectiveDedent

    def _analyzeAndDedent(self, whitespaceline):
        numtabs = whitespaceline.count('\t')
        orignumtabs = numtabs
        numspaces = whitespaceline.count(' ')
        orignumspaces = numspaces
        newline = ''
        numspaces += numtabs * self.tabWidth
        numindents, numspaces = divmod(numspaces, self.indentWidth)
        if numspaces == 0:
            numindents -= 1  # any extra spaces are considered an indent
        numspaces = self.indentWidth * numindents
        if self.useTabs:
            numnewtabs, numspaces = divmod(numspaces, self.tabWidth)
            effectiveDedent = (orignumspaces + orignumtabs) - (numspaces + numnewtabs)
            return '\t'*numnewtabs + ' '*numspaces, effectiveDedent
        effectiveDedent = orignumspaces - numspaces
        return ' '*numspaces, effectiveDedent

    def _do_cmd_backSmart(self):
        # Our backspace is by default fairly "smart"
        # If there is only whitespace to the left of us (on the current line), we delete enough characters
        #    to do a proper alignment ( this is currently done by IDLE, and works well w.r.t. tabs, spaces, etc.
        # Else we just do a regular backspace.
        #
        # The IDLE code "by itself"  is too aggressive deleting
        # whitespace in the middle of lines (Where alignment is not relevant).
        #
        currentLineNo = self.scimoz.lineFromPosition(self.scimoz.currentPos)
        startofLinePos = self.scimoz.positionFromLine(currentLineNo)
        endofLinePos = self.scimoz.getLineEndPosition(currentLineNo)
        wholeLine = self.scimoz.getTextRange(startofLinePos, endofLinePos)
        lineLength = self.scimoz.lineLength(currentLineNo)
        positioninLine, buffer = self.scimoz.getCurLine()
        if wholeLine[:positioninLine].strip() == '':
            # we're at the beginning of the line -- let IDLE do the smart backtabbing
            self.idleIndentExt.smart_backspace_event(None)
        else:
            # we're somewhere in the line with text to our right and left -- just do a backspace.
            self.scimoz.deleteBack()

    def _do_cmd_completeWord(self):
        ext = self.idlewindow.IDLEExtension("AutoExpand")
        ext.expand_word_event(None)
        self.scimoz.gotoPos(self.scimoz.currentPos);

    def _getFoldLevel(self, line):
        """Return a fold level for the current line that guarantees that:
            - if at the top level, the fold level is zero
            - the deeper you are the higher the fold level
        It does *not* guarantee that:
            - a depth of three means then level is three
        """ 
        sm = self.scimoz
        # if the line being asked about has never been lexed, then lex it.
        _startOfLine = sm.positionFromLine(line)
        _style = sm.getStyleAt(_startOfLine)
        if _style == 0:
            # NOTE: assumption that style = 0 means that it hasn't been lexed.            _startOfNextLine = sm.positionFromLine(line+20)
            sm.colourise(0, -1)

        topFoldLevel = sm.getFoldLevel(0)
        topFoldLevel &= sm.SC_FOLDLEVELNUMBERMASK
        foldLevel = sm.getFoldLevel(line)
        foldLevel &= sm.SC_FOLDLEVELNUMBERMASK
        foldLevel -= topFoldLevel
        return foldLevel

    def _do_cmd_blockSelect(self):
        """Use the lexer's fold markers to increase the selection.
        If the current selection *is* a complete "fold block" then increase
        to the next fold block. If the current selection is *not* a complete
        fold block then select the fold block of the shallowest end of the
        selection.

        The logic is complicated somewhat because, the line preceeding a
        block at a certain level should be included in the block selection.
        For example, in the following code, if "Ctrl-B" is hit with the
        cursor on the "print 1" line, the "for element in mylist:" line
        should be included in the selection.

        LEVEL   CODE
         0       def foo(a):
         1           mylist = ["hello", 42, "there", 3.14159]
         1           while 1:
         2               for element in mylist:
         3                   print 1
         3                   if type(element) == type(""):
         4                       print "element %s is a string" % element
         1           print
        """
        #logDebug("----------------- BLOCK SELECT ---------------------")
        sm = self.scimoz
        startLine = sm.lineFromPosition(sm.selectionStart)
        endLine = sm.lineFromPosition(sm.selectionEnd)
        lastLine = sm.lineFromPosition(sm.textLength-1)

        # if the lexer doesn't support folding, select all and get out of here fast.
        lexer = self.languageObj.getLanguageService(
            components.interfaces.koILexerLanguageService)
        if not lexer.supportsFolding:
            sm.selectAll()
            return

        # If the next line after the startLine is "deeper" then move the
        # startLine to there. See the doc string for why this is done.
        if startLine != lastLine \
           and self._getFoldLevel(startLine+1) > self._getFoldLevel(startLine):
            startLine += 1

        # If there is a selection and selectionEnd is at the start of a line
        # then move up the end of the previous line because this simplifies
        # the subsequent logic.
        if sm.selText != "" and sm.getColumn(sm.selectionEnd) == 0:
            endLine -= 1
            sm.selectionEnd = sm.getLineEndPosition(endLine)
        #logDebug("BLOCKSELECT: startLine=%d, endLine=%d, lastLine=%d"\
        #         % (startLine, endLine, lastLine))

        # Determine the minimum fold level in the current selection.
        minFoldLevel = self._getFoldLevel(startLine)
        for line in range(startLine, endLine):
            #logDebug("BLOCKSELECT: foldLevel for line %d is: %x" \
            #         % (line, self._getFoldLevel(line)))
            minFoldLevel = min(minFoldLevel, self._getFoldLevel(line))
        #logDebug("BLOCKSELECT: minFoldLevel is %x" % minFoldLevel)

        # Determine if the current selection is a complete fold block.
        # - selection start must be at the start of a line
        # - selection end must be at the end of a line
        # - start and end fold level must both be at the bound of the
        #   minFoldLevel
        if sm.getColumn(sm.selectionStart) == 0 \
           and sm.selectionEnd == sm.getLineEndPosition(endLine) \
           and (startLine == 0 \
                or (self._getFoldLevel(startLine-1) < minFoldLevel \
                    and self._getFoldLevel(startLine) \
                        > self._getFoldLevel(startLine-1))
               ) \
           and (endLine == lastLine \
                or (self._getFoldLevel(endLine+1) < minFoldLevel \
                    and self._getFoldLevel(endLine) \
                        > self._getFoldLevel(endLine+1))
               ):
            #logDebug("BLOCKSELECT: *is* a complete fold block")
            if startLine == 0:
                targetFoldLevel = 0
            else:
                targetFoldLevel = self._getFoldLevel(startLine-1)
        else:
            #logDebug("BLOCKSELECT: condition: start of selection on column 0? %s" \
            #         % (sm.getColumn(sm.selectionStart) == 0))
            #logDebug("BLOCKSELECT: condition: end of selection at eol? %s" \
            #         % (sm.selectionEnd == sm.getLineEndPosition(endLine)))
            #logDebug("BLOCKSELECT: condition: startLine at minFoldLevel boundary? %s" \
            #         % (startLine == 0 \
            #            or (self._getFoldLevel(startLine-1) < minFoldLevel \
            #                and self._getFoldLevel(startLine) \
            #                    > self._getFoldLevel(startLine-1))))
            #logDebug("BLOCKSELECT: condition: endLine at minFoldLevel boundary? %s" \
            #         % (endLine == lastLine \
            #            or (self._getFoldLevel(endLine+1) < minFoldLevel \
            #                and self._getFoldLevel(endLine) \
            #                    > self._getFoldLevel(endLine+1))))
            #logDebug("BLOCKSELECT: is *not* a complete fold block")
            targetFoldLevel = minFoldLevel
        #logDebug("BLOCKSELECT: targetFoldLevel is %x" % targetFoldLevel)

        # increase the selection to the target fold level
        if targetFoldLevel == 0:
            #logDebug("BLOCKSELECT: select all")
            sm.selectionStart = 0
            sm.selectionEnd = sm.textLength
        else:
            # expand upwards
            while startLine > 0:  #XXX are line's in SciMoz 0- or 1-based???
                #logDebug("BLOCKSELECT: level(%x) up one line(%d) > target level(%x)?" \
                #         % (self._getFoldLevel(startLine-1), startLine-1,
                #            targetFoldLevel))
                if self._getFoldLevel(startLine-1) >= targetFoldLevel:
                    startLine -= 1
                else:
                    break 
            # select the leading line in the block as well
            if startLine != 0:
                startLine -= 1

            # expand downwards
            while endLine < lastLine:
                #logDebug("BLOCKSELECT: level(%x) down one line(%d) > target level(%x)?" \
                #         % (self._getFoldLevel(endLine+1), endLine+1,
                #            targetFoldLevel))
                if self._getFoldLevel(endLine+1) >= targetFoldLevel:
                    endLine += 1
                else:
                    break 

            # make the selection
            #logDebug("BLOCKSELECT: lineStart=%d, lineEnd=%d" % (startLine, endLine))
            sm.selectionStart = sm.positionFromLine(startLine)
            if endLine == lastLine:
                sm.selectionEnd = sm.textLength
            else:
                sm.selectionEnd = sm.positionFromLine(endLine+1)


    def _do_cmd_tabify(self):
        self.idleIndentExt.tabify_region_event(None)

    def _do_cmd_untabify(self):
        self.idleIndentExt.untabify_region_event(None)

    def _is_cmd_comment_enabled(self):
        commenter = self.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        return commenter is not None

    def _do_cmd_comment(self):
        commenter = self.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        commenter.comment(self.scimoz)

    def _is_cmd_uncomment_enabled(self):
        commenter = self.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        return commenter is not None

    def _do_cmd_uncomment(self):
        commenter = self.languageObj.getLanguageService(
            components.interfaces.koICommenterLanguageService)
        commenter.uncomment(self.scimoz)

    def _do_cmd_convertUpperCase(self):
        selStart = self.scimoz.selectionStart
        selEnd = self.scimoz.selectionEnd
        self.scimoz.replaceSel(self.scimoz.selText.upper())
        self.scimoz.selectionStart = selStart
        self.scimoz.selectionEnd = selEnd

    def _do_cmd_convertLowerCase(self):
        selStart = self.scimoz.selectionStart
        selEnd = self.scimoz.selectionEnd
        self.scimoz.replaceSel(self.scimoz.selText.lower())
        self.scimoz.selectionStart = selStart
        self.scimoz.selectionEnd = selEnd

    def _is_cmd_selectToMatchingBrace_enabled(self):
        scimoz = self.scimoz   
        isInside, braceAtCaret, braceOpposite = self._findMatchingBracePosition(scimoz.currentPos,1)
        if braceAtCaret == -1 or braceOpposite == -1:
            return 0
        return 1
    _is_cmd_jumpToMatchingBrace_enabled = _is_cmd_selectToMatchingBrace_enabled

    def _do_cmd_selectToMatchingBrace(self):
        scimoz = self.scimoz   
        isInside, braceAtCaret, braceOpposite = self._findMatchingBracePosition(scimoz.currentPos,1)
        if braceAtCaret == -1 or braceOpposite == -1:
            return
        if braceAtCaret < braceOpposite:
            scimoz.currentPos = braceOpposite+1
            scimoz.anchor = braceAtCaret
        else:
            scimoz.currentPos = braceOpposite
            scimoz.anchor = braceAtCaret+1
        scimoz.scrollCaret()

    def _do_cmd_jumpToMatchingBrace(self):
        scimoz = self.scimoz   
        isInside, braceAtCaret, braceOpposite = self._findMatchingBracePosition(scimoz.currentPos,1)
        if braceAtCaret == -1 or braceOpposite == -1:
            return
        if braceAtCaret < braceOpposite:
            targetPos = braceOpposite+1
        else:
            targetPos = braceOpposite
        scimoz.selectionStart = targetPos
        scimoz.selectionEnd = targetPos
        scimoz.scrollCaret()

    # If the caret is positioned next to a brace, the brace and its
    # matching brace will be highlighted for the user.
    def showMatchingBrace(self):
        scimoz = self.scimoz   

        isInside, braceAtCaret, braceOpposite = self._findMatchingBracePosition(scimoz.currentPos,1)
        
        # print "_findMatchingBracePosition(",isInside,",",braceAtCaret,",",braceOpposite,")"
        if braceAtCaret != -1 and braceOpposite == -1:
            scimoz.braceBadLight(braceAtCaret)
            scimoz.highlightGuide = 0

        else:
            chBrace = scimoz.getWCharAt(braceAtCaret)
            scimoz.braceHighlight(braceAtCaret, braceOpposite)
            columnAtCaret = scimoz.getColumn(braceAtCaret)
            if chBrace == ':':
                lineStart = scimoz.lineFromPosition(braceAtCaret)
                indentPos = scimoz.getLineIndentPosition(lineStart)
                indentPosNext = scimoz.getLineIndentPosition(lineStart+1)
                columnAtCaret = scimoz.getColumn(indentPos)
                columnAtCaretNext = scimoz.getColumn(indentPosNext)
                indentSize = scimoz.indent
                if (columnAtCaretNext - indentSize > 1):
                    columnAtCaret = columnAtCaretNext - indentSize 
            columnOpposite = scimoz.getColumn(braceOpposite)
            scimoz.highlightGuide = min(columnAtCaret, columnOpposite)


    # Find the brace that matches the brace before or after the given caret
    # position. The character before has precedence if it is a brace.
    # Returns three values: whether the caret is inside the braces,
    # and the indexes of the current and the matching brace.
    def _findMatchingBracePosition(self, caretPos, sloppy):
        s = self.scimoz
        isInside = 0
        braceAtCaret = -1
        braceOpposite = -1
        charBefore = '\0'
        styleBefore = 0
        textLength = s.textLength
        if caretPos > textLength:
            caretPos = textLength
        if (caretPos > 0) :
            charBefore = s.getWCharAt(caretPos - 1)
            styleBefore = s.getStyleAt(caretPos - 1) & self.stylingBitsMask
        # Priority goes to character before caret
        if charBefore and self._getBraceIndentStyle(charBefore,styleBefore):
            braceAtCaret = caretPos - 1
        
        # TODO: This should be done in PythonCommandHandler
        colonMode = 0
        if isinstance(self,PythonCommandHandler) and ':' == charBefore:
            braceAtCaret = caretPos - 1
            colonMode = 1

        isAfter = 1
        if sloppy and (braceAtCaret < 0) and  0 <= caretPos < textLength: # XXX check last edge condition
            # No brace found so check other side
            charAfter = s.getWCharAt(caretPos)
            styleAfter = s.getStyleAt(caretPos)  & self.stylingBitsMask
            
            if charAfter and (charAfter in "[](){}") and self._getBraceIndentStyle(charAfter,styleAfter):
                braceAtCaret = caretPos
                isAfter = 0

            # TODO: This should be done in PythonCommandHandler
            if isinstance(self,PythonCommandHandler) and ':' == charAfter:
                braceAtCaret = caretPos
                colonMode = 1
                
        if braceAtCaret >= 0:
            if colonMode:
                lineStart = s.lineFromPosition(braceAtCaret)
                lineMaxSubord = s.getLastChild(lineStart, -1)
                braceOpposite = s.getLineEndPosition(lineMaxSubord)
            else:
                braceOpposite = s.braceMatch(braceAtCaret)
                
        if braceOpposite > braceAtCaret:
            isInside = isAfter
        else:
            isInside = not isAfter
        return isInside, braceAtCaret, braceOpposite


    # returns 0 if the character at the given index is not a valid brace, or
    # self.INDENT_BRACE/self.LINEUP_BRACE depending on its type.
    def _getBraceIndentStyle(self,ch,style):
        style = style  & self.stylingBitsMask
        if ch in self.braceData[0] and style in self.braceData[1]:
            return self.INDENT_BRACE
        
        if ch in self.braceData[2] and style in self.braceData[3]:
            return self.LINEUP_BRACE

        return 0        
    
    # Find the previous non-blank line, and copy its indentation.
    def guessIndentation(self,index):
        lineNumber = self.scimoz.lineFromPosition(index)-1
        while lineNumber > 0:
            lineLen, thisLine = self.scimoz.getLine(lineNumber)
            thisLine = thisLine[:lineLen].rstrip()
            if thisLine:
                return thisLine[:-len(thisLine.lstrip())]
            lineNumber -= 1
        return ""

    def _is_cmd_transpose_enabled(self):
        index, text = self.scimoz.getCurLine()
        return index > 1

    def _do_cmd_transpose(self):
        # transpose two characters to the left
        # this particular sequence (others are possible) has good Undo behavior
        index, text = self.scimoz.getCurLine()
        if index < 2: return

        sm = self.scimoz
        sm.beginUndoAction()
        try:
            sm.charLeftExtend()
            sm.cut()
            sm.charLeft()
            sm.paste()
            sm.charRight()
            #sm.charRight() # this one is required for Emacs compatibility, but I don't see the use case.
        finally:
            sm.endUndoAction()

    def _do_cmd_transposeWords(self):
        # transpose two words to the left (or current word and previous word if in a word)
        # this particular sequence (others are possible) has good Undo behavior

        sm = self.scimoz
        sm.beginUndoAction()
        try:
            sm.wordLeft()
            sm.wordLeftExtend()
            sm.cut()
            sm.wordLeft()
            sm.paste()
            sm.wordRight()
        finally:
            sm.endUndoAction()


class PythonCommandHandler(TextCommandHandler):
    def __init__(self, *args):
        TextCommandHandler.__init__(self, *args)
        self.indentation_params = None
        self.doBraceMatching = 1

    # Amost directly from Pythonwin!
    # Returns None, or a tuple of (indentSize, useTabs)

    # Auto-indent has very complicated behaviour.  In a nutshell, the only
    # way to get sensible behaviour from it is to ensure tabwidth != indentsize.
    # Further, usetabs will only ever go from 1->0, never 0->1.
    # This is _not_ the behaviour Pythonwin wants:
    # * Tab width is arbitary, so should have no impact on smarts.
    # * bUseTabs setting should reflect how new files are created, and
    #   if Smart Tabs disabled, existing files are edited
    # * If "Smart Tabs" is enabled, bUseTabs should have no bearing
    #   for existing files (unless of course no context can be determined)
    #
    # So for smart tabs we configure the widget with completely dummy
    # values (ensuring tabwidth != indentwidth), ask it to guess, then
    # look at the values it has guessed, and re-configure
    # TODO: is this a hack? i think so!
    def _guessFileIndentation(self):
        self.idleIndentExt.config(usetabs=1, tabwidth=5, indentwidth=4)
        self.idleIndentExt.set_indentation_params(1)
        if self.idleIndentExt.indentwidth==5:
            # Either 5 literal spaces, or a single tab character. Assume a tab
            return self.tabWidth, 1
        # Have an answer we can trust.
        return self.idleIndentExt.indentwidth, 0

    def _do_cmd_newline(self):
        # koEditorWrapper.onModified needs to work when we hit enter, so we
        # have to cancel the batch operation mode.
        # kind of hacky to have this here, but no better way right now
        if self.wrapper is not None: self.wrapper.endBatchOperation()
        autoIndentStyle = self.prefs.getStringPref('editAutoIndentStyle')
        if autoIndentStyle != 'smart':
            TextCommandHandler._do_newline(self, autoIndentStyle)
            return
        self.idleIndentExt.newline_and_indent_event(None)
        # if last statement is pass, return, raise, _dedent_

    #def keyPressed(self, char):
    #    if key != ':' or not self.prefs.getBooleanPref('editElectricBrace'):
    #        return
    #     find out if the colon is a block delimiting colon
    #     find indent of current line
    #     find the first word in the current logical line
    #     if it is else, and if the else starts at the same indent level as the previous logical line,
    #     dedent the current logical line.
    #    

# This class handles autoindenting for languages which use braces. We use the
# term brace refers to any grouping characters like round and square brackets.
class BraceCommandHandler(TextCommandHandler):

    # Constants
    OPEN_BRACES = "{(["
    CLOSE_BRACES = "})]"

    def __init__(self, *args):
        TextCommandHandler.__init__(self, *args)
        self.doBraceMatching = 1

    def keyPressed(self, key):
        # If a close brace is typed, find the matching open brace and
        # ensure that the close brace is positioned respectively.
        if key not in self.CLOSE_BRACES or not self.prefs.getBooleanPref('editElectricBrace'):
            return
        scimoz = self.scimoz
        # If the brace is preceded only by whitespace, count the number
        # of spaces.
        currentPos = scimoz.currentPos
        startIndex = currentPos - 2
        while startIndex >= 0:
            ch = scimoz.getWCharAt(startIndex)
            if ch != ' ' and ch != '\t':
                if ch == '\n' or ch == '\r':
                    startIndex += 1
                    break
                return None # the brace is not preceded by whitespace
            startIndex -= 1

        desiredIndent = self._getIndentAt(key, startIndex, currentPos)
        if desiredIndent is None: return
        # Replace the current indent with the desired indent.
        scimoz.selectionStart = startIndex
        scimoz.selectionEnd = currentPos-1
        #print 'replace %s to %s with "%s"' % \
        #      (startIndex, currentPos-1, desiredIndent)
        scimoz.replaceSel(desiredIndent)                
        scimoz.selectionEnd += 1
        scimoz.selectionStart = scimoz.selectionEnd

    def _getIndentAt(self, key, startIndex, currentPos):
        scimoz = self.scimoz
        isInside, braceAtCaret, braceOpposite = self._findMatchingBracePosition(currentPos,1)  
        if braceAtCaret != -1 and braceOpposite == -1:
            return None

        if key == '}':
            # find out if it's an indenting { or a line-up brace
            style = self._getBraceIndentStyle('}',scimoz.getStyleAt(braceOpposite))
            if style == self.LINEUP_BRACE:
                desiredIndent = self._indentFromColumn(scimoz.getColumn(braceOpposite))
            else:
                desiredIndent = self._guessIndentation(braceOpposite-1)
        else:
            desiredIndent = self._indentFromColumn(scimoz.getColumn(braceOpposite))
        return desiredIndent

    # returns 0 if the character at the given index is not a valid brace, or
    # self.INDENT_BRACE/self.LINEUP_BRACE depending on its type.
    #def _getBraceIndentStyle(self,ch,style):
    #    style = style & 31 # XXX need proper stylemask for language!!!
    #    if not ch in "{}()[]" or \
    #       (not style & sci_constants.SCE_PL_OPERATOR and \
    #        not style & sci_constants.SCE_PL_VARIABLE_INDEXER and \
    #        not style & sci_constants.SCE_PL_REGEX and \
    #        not style & sci_constants.SCE_PL_REGSUBST):
    #        return 0
    #                
    #    if (ch == '{' or ch == '}'):
    #        if style & sci_constants.SCE_PL_OPERATOR:
    #            return self.INDENT_BRACE
    #        elif style & sci_constants.SCE_PL_VARIABLE_INDEXER:
    #            return self.LINEUP_BRACE
    #        
    #    return self.LINEUP_BRACE
    #
    def _do_cmd_newline(self):
        # koEditorWrapper.onModified needs to work when we hit enter, so we
        # have to cancel the batch operation mode.
        # kind of hacky to have this here, but no better way right now
        if self.wrapper is not None: self.wrapper.endBatchOperation()

        autoIndentStyle = self.prefs.getStringPref('editAutoIndentStyle')
        if autoIndentStyle != 'smart':
            TextCommandHandler._do_newline(self, autoIndentStyle)
            return
        self.scimoz.beginUndoAction()
        try:
            indent = self._guessIndentation(self.scimoz.currentPos)
            # need to remove whitespace immediately after the current character
            # XXX should be pref'able
            # iff that's all that is to the right of the current character.
            endOfLine = self.scimoz.positionFromLine(self.scimoz.lineFromPosition(self.scimoz.currentPos)+1)
            thisline = self.scimoz.getTextRange(self.scimoz.currentPos, endOfLine)
            if not thisline.strip():
                while self.scimoz.getWCharAt(self.scimoz.currentPos) in ' \t':
                    self.scimoz.clear()
            # if there are non-whitespace characters to the left,
            # remove any whitespace characters which occur after them
            # XXX should be pref'able
            beginningOfLine = self.scimoz.positionFromLine(self.scimoz.lineFromPosition(self.scimoz.currentPos))
            leftwards = self.scimoz.getTextRange(beginningOfLine, self.scimoz.currentPos)
            if leftwards.strip():
                while self.scimoz.getWCharAt(self.scimoz.currentPos-1) in ' \t':
                    self.scimoz.deleteBack()
            endOfLine = self.scimoz.positionFromLine(self.scimoz.lineFromPosition(self.scimoz.currentPos)+1)
            rightwards = self.scimoz.getTextRange(self.scimoz.currentPos, endOfLine)
            if not rightwards.strip():
                while self.scimoz.getWCharAt(self.scimoz.currentPos) in ' \t':
                    self.scimoz.clear()
            startIndex = self.scimoz.currentPos
            self.scimoz.newLine()
            self.scimoz.addText(len(indent), indent)
        finally:
            self.scimoz.endUndoAction()

    # Calculate the indentation for the given index based on the previous
    # lines.
    def guessIndentation(self,index):
        lineNumber = self.scimoz.lineFromPosition(index)-1
        return self._guessIndentation(self.scimoz.getLineEndPosition(lineNumber)-1)

    # Return the indentation string for the line after the index given, using
    # the contents of the relevant line, up to the given index.
    def _guessIndentation(self,endIndex):
        scimoz = self.scimoz

        lineNumber = scimoz.lineFromPosition(endIndex)        
        startIndex = scimoz.positionFromLine(lineNumber)
        previousIndent = "" # the tabs and spaces preceding the previous line.
        
        # Starts from the beginning of the previous line until it finds the
        # first non whitespace character. Records the indent of this line
        # to use as a guide for the indent of the next line.
        while startIndex < endIndex:
            ch = scimoz.getWCharAt(startIndex)
            if ch != ' ' and ch != '\t':
                break
            previousIndent += ch
            startIndex += 1
        # At this point we have previousIndent, which is the indent of the current line
        
        closeBraceCount = [0,0,0] # the number of close braces without a matching open brace
        openCurlyBraceCount = 0 # the number of open curly braces without a matching close brace
        lastLineupOpenBrace = -1 # the last unmatched lineup open brace
        lastCloseBrace = -1 # the last unmatched close brace
        lastCloseBraceType = -1 # the last unmatched close brace
        
        # Get the values for the above variables.
        index = endIndex

        # set the default indent string
        indentStr = (' ' * self.indentWidth)
        while index >= startIndex:
            # here we're going backwards to find out if we have closed enough
            # braces.  A problem is that if the current line closes some but not all
            # braces, it'll be wrong, because the exit condition is too early
            ch = scimoz.getWCharAt(index)
            braceIndent = self._getBraceIndentStyle(ch,scimoz.getStyleAt(index))
            if braceIndent:
                braceType = self.CLOSE_BRACES.find(ch)

                # Recognise and count the number of closing braces. If
                # there are currently none unmatched, record this as the
                # last close brace.  Ignore any leading close curly brace
                # as it has already been handled.
                if braceType >= 0 and not (braceType == 0 and index == startIndex):
                    if closeBraceCount == [0,0,0]:
                        lastCloseBrace = index
                        lastCloseBraceType = braceType
                    closeBraceCount[braceType] += 1
                else:

                    braceType = self.OPEN_BRACES.find(ch)
                    if braceType >= 0:
                        if closeBraceCount[braceType] == 0:
                            if braceIndent == self.INDENT_BRACE:
                                openCurlyBraceCount += 1
                            elif lastLineupOpenBrace == -1:
                                lastLineupOpenBrace = index
                        else:
                            closeBraceCount[braceType] -= 1
                            if lastCloseBraceType == braceType and closeBraceCount[braceType] == 0:
                                lastCloseBrace = -1
                                lastCloseBraceType = -1
            index -= 1

        desiredIndent = "" # will contain the line indent characters

        # The line contains an unmatched open lineup brace - indent the
        # next line with extra spaces so that it lines up underneath the
        # brace.
        if lastLineupOpenBrace >= 0:
            desiredIndent = previousIndent + ' ' * (lastLineupOpenBrace + 1 - startIndex)
        else:

            if lastCloseBrace < 0:
                desiredIndent = previousIndent

            # The line contains an unmatched close brace. Find the indent
            # of the line matching the last close brace and use that as a
            # base.
            else:
                # There are two cases -- linup braces and indenting braces
                isInside, braceAtCaret, braceOpposite = self._findMatchingBracePosition(lastCloseBrace+1,1)
                if braceOpposite >= 0:
                    index = scimoz.positionFromLine(scimoz.lineFromPosition(braceOpposite))
                    while index < braceOpposite:
                        ch = scimoz.getWCharAt(index)
                        if ch != ' ' and ch != '\t':
                            break
                        desiredIndent += ch
                        index += 1
                
            # Add extra spaces/tabs for the extra indent provided by open
            # curly braces.
            if openCurlyBraceCount:
                desiredIndent += indentStr * openCurlyBraceCount

        # XXX fix for: http://bugs.activestate.com/Komodo/show_bug.cgi?id=20474
        # the above algorithm is wrong in the case of a matching brace which is
        # the first character on the line -- the following fixes it, although
        # I'd rather it weren't a special case like this.
        # There are a few cases which come in play here -- matching braces as in the above
        # referenced bug.  Also, newline handling if the cursor is on the same line
        # as an opening brace, with sub-cases involving immediately before the brace,
        # one character to the left, and more characters to the left, and after the opening brace.
        if index + 1 == endIndex and \
           scimoz.getWCharAt(endIndex) in ' \t' and\
           scimoz.getWCharAt(endIndex+1) not in ' \t\r\n' and\
           endIndex != scimoz.currentPos:  
            desiredIndent += scimoz.getWCharAt(endIndex)
        
        if self.useTabs and self.tabWidth:
            # replace spaces with tabs by first making sure we are all spaces
            desiredIndent = desiredIndent.expandtabs(self.tabWidth)
            # and then converting appropriately to tabs
            desiredIndent = desiredIndent.replace((' ' * self.tabWidth), '\t')

        #dumpStr = desiredIndent.replace(' ', 's').replace('\t', 't')
        #print "guessIndentation() = \"" + dumpStr + "\""

        return desiredIndent

PerlBraceData = ["{}", 
                     [
                      sci_constants.SCE_PL_OPERATOR, 
                     ],
                     "{}()[]",
                     [
                      sci_constants.SCE_PL_OPERATOR,
                      sci_constants.SCE_PL_VARIABLE_INDEXER,
                      sci_constants.SCE_PL_REGEX,
                      sci_constants.SCE_PL_REGSUBST
                     ],
                    ]
PHPBraceData = ["{}",
                     [sci_constants.SCE_HPHP_OPERATOR],
                     "()[]",
                     [sci_constants.SCE_HPHP_OPERATOR],
                    ]
DefaultBraceData = ["{}",
                     [sci_constants.SCE_PL_OPERATOR],
                     "()[]",
                     [sci_constants.SCE_PL_OPERATOR],
                    ]


def GetControllerForLanguage(wrapper):
    #XXX Eventually all the language specific stuff here should be deferred
    #    language services.
    languageName = wrapper.languageOb.name;
    if languageName == "Perl":
        braceData = PerlBraceData;
    elif languageName == "PHP":
        braceData = PHPBraceData;
    else:
        braceData = DefaultBraceData;

    if languageName == "Python":
        handler = PythonCommandHandler(braceData, wrapper)
    else:
        #TODO: this should be handled by language services.
        if languageName in ["C#","C++","Java","JavaScript","Perl","Tcl","PHP"]:
            handler = BraceCommandHandler(braceData, wrapper)
        else:
            handler = TextCommandHandler(braceData, wrapper)
    return handler



