from xpcom.server import UnwrapObject
from xpcom import components, COMException
import sys

def splitStylePrefName(styleName):
    return styleName.split("/")[1:]

# A couple of caches.
color_map = {}
style_name_to_number_map = {}

class koStyleApplier:
    def __init__(self, language, styleMap, stylePrefs, defaultPrefs, encoding_info):
        #print "koStyleApplier ",encoding_info.python_encoding_name
        self._language = language
        self._styleMap = UnwrapObject(styleMap) # Unwrap as a perf opt.
        self._stylePrefs = UnwrapObject(stylePrefs) # Unwrap as a perf opt.
        self._scintilla = None
        self._defaultPrefs = defaultPrefs
        self._encoding_info = encoding_info # koIEncodingInfo
        self._linux = sys.platform.startswith("linux")

    # _stylePrefInfo entries are lists containing the following
    # [0] get pref function
    # [1] set pref function
    # [2] true means it is a "color" preference, used when applying styles to Scintilla
    # [3] function for setting the Scintilla property
    # [4] temporary storage of the value
    # !!! font MUST always be LAST!
    _stylePrefInfo = {'size': ['getLongPref', 'setLongPref', 0, 'styleSetSize',None],
                     'italic': ['getBooleanPref', 'setBooleanPref', 0, 'styleSetItalic',None],
                     'eolfilled': ['getBooleanPref', 'setBooleanPref', 0, 'styleSetEOLFilled',None],
                     'bold': ['getBooleanPref', 'setBooleanPref', 0, 'styleSetBold',None],
                     'back': ['getStringPref', 'setStringPref', 1, 'styleSetBack',None],
                     'fore': ['getStringPref', 'setStringPref', 1, 'styleSetFore',None],
                     'font': ['getStringPref', 'setStringPref', 0, 'styleSetFont',None]}

    def _getPrefName(self, styleName, styleType):
        prefName = self._buildPrefName(self._language, styleName, styleType)
        if not self._stylePrefs.hasPref(prefName):
            prefName = self._buildPrefName(self._language, "*", styleType)
            if not self._stylePrefs.hasPref(prefName):
                prefName = self._buildPrefName("*", "*", styleType)
                if not (self._stylePrefs.hasPref(prefName) or \
                        (self._defaultPrefs and \
                         self._defaultPrefs.hasPref(prefName))):
                        prefName = None

        assert prefName is not None, "no default pref for style %s, %s, %s" % \
               (self._language, styleName, styleType)
        return prefName

    def _buildPrefName(self, language, styleName, styleType):
        return "style/%s/%s/%s" % (language, styleName, styleType)


    class StyleNumberError:
        pass

    # This function is too slow, so we cache the values
    def _getStyleNumber(self, styleName):
        try:
            return style_name_to_number_map[self._language,styleName]
        except KeyError:
            pass
        if self._styleMap.hasPref(styleName):
            rawPref = self._styleMap.getStringPref(styleName)
        else:
            raise self.StyleNumberError
        
        try:
            styleNumber = int(rawPref)
        except ValueError:
            try:
                styleNumber = getattr(self._scintilla, rawPref)
            except AttributeError:
                logError("SciMoz doesn't have attribute '%s' for style '%s'" % \
                         (rawPref, styleName))
                styleNumber = 0
        style_name_to_number_map[self._language,styleName] = styleNumber
        return styleNumber

    def _getScinColor(self, colorName):
        # No colour name means black.
        if len(colorName) < 1:
            return 0
        if not color_map:
            import colors
            color_map.update(colors.colors)
        try:
            return color_map[colorName]
        except KeyError:
            pass
        try:
            r,g,b = map(int, colorName.split(','))
            colorVal = r+g*256+b*256*256
        except ValueError:
            if colorName[0] == "#":
                r = int(colorName[1:3], 16)
                g = int(colorName[3:5], 16)
                b = int(colorName[5:7], 16)
                colorVal = r + (g << 8) + (b << 16)
            else:
                logError("color %s is invalid" % colorName)
                colorVal = 0
        color_map[colorName] = colorVal
        return colorVal

    def _getCSSColor(self, colorName):
        color = self._getScinColor(colorName)
        r = (color & 0x0000FF)
        g = (color & 0x00FF00) >> 8
        b = (color & 0xFF0000) >> 16
        return "#%02x%02x%02x" % (r,g,b)

    def _buildFontSpec(self):
        # UNIX ONLY
        # we need to build a font spec from information provided
        # so far in prefs
        #
        # format: condense to one line
        #
        #    -adobe-helvetica-bold-r-normal--12-120-75-75-p-70-iso8859-1
        #
        #    -foundary
        #    -fontName
        #    -weight (normal,bold,medium)
        #    -slant (R,I,O - Roman,Italic,Oblique)
        #    -width ('normal','condensed')
        #    -style (leave emtpy)
        #    -pixelsize
        #    -pointsize  (10ths of a point)
        #    -resulotionX
        #    -resolutionY
        #    -spacing (p,m,c - proportional,monospaced or charcell)
        #    -aveWidth
        #    -charsetRegistry (eg. 8859)
        #    -charsetEncoding (eg. 1 through 15)
        
        # define the defaults
        #print "FONT 1: ",self._stylePrefInfo['font'][4]
        encoding = self._encoding_info.fontspec.split('-')
        fontinfo = ['*','*',encoding[0],encoding[1]]
        if self._stylePrefInfo['font'][4] and \
           self._stylePrefInfo['font'][4][0] == '-':
            if self._stylePrefInfo['font'][4].find(',') >= 0:
                self._stylePrefInfo['font'][4] = self._stylePrefInfo['font'][4].split(',')[0]
            parts = self._stylePrefInfo['font'][4].split('-')
            fontinfo = [parts[1],parts[2],encoding[0],encoding[1]]
        # font pref looks like: adobe-courier-iso8859-15
        elif self._stylePrefInfo['font'][4]:
            fontinfo = self._stylePrefInfo['font'][4].split('-')
            if len(fontinfo) != 4:
                fontinfo = ['*',self._stylePrefInfo['font'][4].lower(),encoding[0],encoding[1]]
                #print "font??: ",self._stylePrefInfo['font'][4],repr(fontinfo)
        #print "FONT 2: ",self._stylePrefInfo['font'][4]

        if fontinfo[2] != 'iso10646':
            spec = "%s-%s-%s-%s" % \
                (fontinfo[0],fontinfo[1],fontinfo[2],fontinfo[3])
            return spec

        # this will happen for UTF-8 buffers
        fontset = []
        spec = "%s-%s-%s-%s" % \
            (fontinfo[0],fontinfo[1],fontinfo[2],fontinfo[3])
        fontset.append(spec)
        spec = "*-%s-%s-%s" % \
            (fontinfo[1],fontinfo[2],fontinfo[3])
        fontset.append(spec)
        #spec = "*-*-%s-%s" % \
        #    (fontinfo[2],fontinfo[3])
        #fontset.append(spec)
        #spec = "*-*-%s-*" % fontinfo[2]
        #fontset.append(spec)
        #spec = "*"
        #fontset.append(spec)
        #print "fontspec: ",(','.join(fontset)),self._encoding_info.python_encoding_name,self._encoding_info.fontspec
        return ','.join(fontset)

    
    def applyDefaultStyles(self, prefNames, scintilla):
        # This relies on the fact that styleClearAll actually copies the STYLE_DEFAULT
        # style to all other styles.  So we only need to set the default, and styleClearAll()
        # to set the basis for *all* styles.  We can then set any language specific ones.
        styleNumber = scintilla.STYLE_DEFAULT
        fontPref = None
        for prefName in prefNames:
            if prefName.find('font')>=0:
                fontPref = prefName
                continue
            self.applyOneStyle(prefName,scintilla)
        # reset the font correctly
        if fontPref:
            self.applyOneStyle(fontPref,scintilla)
        scintilla.styleClearAll()

    def applyOneStyle(self, prefName, scintilla):
        self._scintilla = scintilla
        (language, styleName, styleType) = splitStylePrefName(prefName)
        if styleType not in self._stylePrefInfo: return

        # Grab the style pref.
        assert self._stylePrefs.hasPref(prefName), "invalid style pref id %s" % prefName
        styleGetPrefFun = getattr(self._stylePrefs, self._stylePrefInfo[styleType][0])
        self._stylePrefInfo[styleType][4] = styleGetPrefFun(prefName)

        if self._linux and styleType =='font':
            self._stylePrefInfo[styleType][4] = self._buildFontSpec()
            
        # Is this style a color style?
        if self._stylePrefInfo[styleType][2]:
            self._stylePrefInfo[styleType][4] = self._getScinColor(self._stylePrefInfo[styleType][4])

        self._applyOneStyle2(styleName, styleType, self._stylePrefInfo[styleType][4])

    def _applyOneStyle2(self, styleName, styleType, value):
        #print "_applyOneStyle2 ",styleName, styleType, value
        try:
            if styleName == '*':
                styleNumber = components.interfaces.ISciMoz.STYLE_DEFAULT
            else:
                styleNumber = self._getStyleNumber(styleName)
        except self.StyleNumberError:
            logError("style %s isn't mapped in the style map prefs for language %s" % (styleName, self._language))
            return

        # Finally, apply the style to the Scintilla.
        getattr(self._scintilla, self._stylePrefInfo[styleType][3])(styleNumber, value)

class koBulkStyleApplier(koStyleApplier):
    """
    Applies ALL the styles for a particular language in bulk to
    a particular Scintilla.
    """

    # Apply ALL styles to a particular Scintilla widget in one shot.
    # applyStyles() is vastly more efficient than the get_ and set_ methods, which
    # get and set individual styles only.
    def applyStyles(self, scintilla):
        styleNames = self._stylePrefs.getAllPrefIds()
        def_styles = []
        specific_styles = []
        for styleName in styleNames:
            if styleName.find("*") >= 0:
                def_styles.append(styleName)
            else:
                specific_styles.append(styleName)
        self.applyDefaultStyles(def_styles, scintilla)
        for styleName in specific_styles:
            self.applyOneStyle(styleName, scintilla)

class koStyle(koStyleApplier):
    """
    Manage _one_ style for a particular language.
    """
    _com_interfaces_ = [components.interfaces.koIStyle]

    def __init__(self, language, styleMap, stylePrefs, styleName, defaultPrefs, encoding_info):
        #print "koStyle init encoding ",encoding_info.python_encoding_name
        koStyleApplier.__init__(self, language, styleMap, stylePrefs, defaultPrefs, encoding_info)
        self._styleName = styleName

    # A function object which stores a style type and
    # style get-or-set method and then forwards calls to
    # the method, providing the style type as the first argument.
    class _StyleSetGet:
        def __init__(self, styleType, method):
            self._styleType = styleType
            self._method = method

        def __call__(self, *args):
            return self._method(self._styleType, *args)

    # The koIStyle Interface
    def __getattr__(self, attr):
        styleType = attr[4:]
        if attr.startswith("get_"):
            if self._stylePrefInfo.has_key(styleType):
                return self._StyleSetGet(styleType, self._getStyle)
        elif attr.startswith("set_"):
            if self._stylePrefInfo.has_key(styleType):
                return self._StyleSetGet(styleType, self._setStyle)
        else:
            raise AttributeError, attr

    # Some useful routines to access foreground and
    # background colours in different formats.
    def get_cssFore(self):
        fore = self._getStyle('fore')
        return self._getCSSColor(fore)

    def set_cssFore(self, fore):
        self._setStyle('fore', fore)

    def get_cssBack(self):
        back = self._getStyle('back')
        return self._getCSSColor(back)

    def set_cssBack(self, back):
        self._setStyle('back', back)

    # Utility Functions

    # Grab this style's value from the style prefs.
    # For instance, _getStyle("font", "identifier")
    def _getStyle(self, styleType):
        prefName = self._getPrefName(self._styleName, styleType)

        if self._stylePrefs.hasPref(prefName):
            prefSet = self._stylePrefs
        else:
            prefSet = self._defaultPrefs
        
        styleGetPrefFun = getattr(prefSet, self._stylePrefInfo[styleType][0])
        return styleGetPrefFun(prefName)

    # Set a style's value in the style prefs.
    # For instance, _setStyle("font", "identifier", "courier new")
    # Note: The style setting information will be forwarded via the preferences
    # observer system to the style manager, which will actually apply changes
    # to the style in all Scintillas under its management.
    def _setStyle(self, styleType, styleValue):
        prefName = self._buildPrefName(self._language, self._styleName, styleType)
        styleSetPrefFun = getattr(self._stylePrefs, self._stylePrefInfo[styleType][1])
        return styleSetPrefFun(prefName, styleValue)
    
class koDefaultStyle(koStyle):

    def __init__(self, stylePrefs, encoding_info):
        # XXX need proper encoding
        #print "koDefaultStyle init ",encoding_info.python_encoding_name
        koStyle.__init__(self, "*", None, stylePrefs, "*", None, encoding_info)
            
