"""

This is the file that controls the progressbar in the splash screen.

It works by first finding out the window handle of the splash screen, stashing
references to the window handles of the progress bar control (and, optionally the
handle to the static text control).

There are two entry points to the module:

initProgressBar and progressStep.

initProgressBar takes no arguments and sets up module globals.

progressStep takes two arguments - a label, and a number of tick marks
by which the progress meter should be moved.

There is a module global TOTALNUMSTEPS, which corresponds to the total number of tick marks.
This number should be updated to reflect the total number of tick marks throughout a startup.

Note that the # of tick marks needs to be _fixed_.
"""



###
import sys
_done = 0

if sys.platform.startswith('win'):
    TOTALNUMSTEPS = 138  # this needs to be tweaked depending on when "ticks" are sent.
    import win32gui, win32api  
    import sys

    # constants pulled from win32con to save memory
    WM_USER = 0x400
    GWL_USERDATA = 0xffffffeb
    PBM_SETRANGE = WM_USER + 1
    PBM_GETRANGE = WM_USER + 7
    PBM_SETRANGE32 = WM_USER + 111
    PBM_SETPOS = WM_USER + 2
    PBM_DELTAPOS = WM_USER + 3
    PBM_SETSTEP = WM_USER + 4
    PBM_STEPIT = WM_USER + 5

    progressControl = None
    labelControl = None

    numSteps = 0
    def _MyCallback( hwnd, extra ):
        userdata = win32api.GetWindowLong(hwnd, GWL_USERDATA)
        # This is a special random number set in nsNativAppSupportWin.cpp that identifies the splash screen
        # the number is chosen to not possibly being a valid pointer, and to try and be somewhat random.
        if userdata == 11757453:
            extra.append(hwnd)

    def _findWindow():
        windows = []
        try:
            win32gui.EnumWindows(_MyCallback, windows)
            if len(windows) == 1:
                return windows[0]
        except win32gui.error:
            return None

    def initProgressBar():
        if not sys.platform.startswith('win'): return
        global progressControl, labelControl
        dialogWindow = _findWindow()
        if dialogWindow is not None:
            try:
                progressControl = win32gui.GetDlgItem( dialogWindow, 200 )
                if progressControl:
                    win32gui.SendMessage( progressControl,
                                          PBM_SETSTEP,
                                          1,
                                          0);
                    win32gui.SendMessage( progressControl,
                                          PBM_SETRANGE,
                                          0,
                                          TOTALNUMSTEPS << 16);
                try:
                    labelControl = win32gui.GetDlgItem( dialogWindow, 300 )
                except win32gui.error:
                    pass
            except win32gui.error:
                pass

    def progressStep(message, numTicks):
        global progressControl, labelControl, numSteps, _done
        if _done: return
        if progressControl is None: initProgressBar()
        numSteps += numTicks
        if numSteps > TOTALNUMSTEPS:
            # if this is true then TOTALNUMSTEPS needs to be updated (see first line in file)
            # XXX Should unload the import hook, etc.
            return
        if progressControl is not None:
            try:
                for i in range(numTicks):
                    win32gui.SendMessage( progressControl,
                                          PBM_STEPIT,
                                          0,
                                          0);
            except win32gui.error, e:
                if e.args[0] == 1400:
                    # the dialog is gone.
                    progressControl = None
        if labelControl is not None:
            try:
                win32gui.SetWindowText(labelControl, message)
            except win32gui.error, e:
                if e.args[0] == 1400:
                    # the dialog is gone.
                    labelControl = None
    def progressEnd():
        global _done
        _done = 1
        if progressControl:
            win32gui.SendMessage( progressControl, PBM_SETSTEP, TOTALNUMSTEPS << 16, 0);
        
else:
    # Do something purely curses based for now
    TOTALNUMSTEPS = 123  # this needs to be tweaked depending on when "ticks" are sent.
    PBWIDTH = 40
    numSteps = 0
    def initProgressBar():
        pass

    def progressStep(message, numTicks):
        if _done: return
        global numSteps, lastNumHashes
        if numSteps == 0:
            firstStep = 1
        else:
            firstStep = 0
        numSteps += numTicks
        if numSteps > TOTALNUMSTEPS:
            return
        numHashes = int(round(float(numSteps) / float(TOTALNUMSTEPS) * PBWIDTH))

        if firstStep:
            sys.stdout.write("\n") # start on a fresh line
        else:
            length = len('Starting Komodo: [') + PBWIDTH + len(']')
            sys.stdout.write('\b'*length)
        sys.stdout.write('Starting Komodo: [' + '#'*numHashes +
                         ' '*(PBWIDTH-numHashes) + ']')
        sys.stdout.flush()

    def progressEnd():
        global _done
        _done = 1
        length = len('Starting Komodo: [') + PBWIDTH + len(']')
        sys.stdout.write('\b'*length)
        sys.stdout.write('Starting Komodo: [' + '#'*PBWIDTH +'] done.\n')
        sys.stdout.flush()
        
