# Copyright (c) 2003 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

r"""
    A Python module to manage a dictionary of named, cache-able system
    data.

    Here is the problem. Komodo needs to know stuff about the system on
    which it is running. E.g. the version of the currently configured
    VPM, or PDK; whether the current Perl is ActivePerl, etc. This info
    is sometimes expensive the determine each time it is needed.
    Therefore a system of caching this data would be helpful.

    When you start caching data you need to know when to flush the
    cache. For example, if the user selects a new "current" perl then
    all info about the current perl (and likely the current PDK must be
    flushed from the cache).
    
    This module provides a simple interface for all of this:
        get(<name>)
        flush(<name>)
    At the pure Python level, the data can be of any type.  There is a
    koISystemDataService to wrap this interface to XPCOM (with the
    expected limitations on return types).

    Any piece of system data is assigned a <name>. For convenience (XXX
    this may end up being for more than convenience, i.e. required) a
    name hierarchy is implied by using dotted names. For example:
        perl                C:\Perl\bin\perl.exe
        perl.version        5.8.0
        perl.installDir     C:\Perl
        perl.buildNum       804
    Then, when a .flush("perl") is called on this module to indicate
    that a new "current" perl installation has been selected, this
    module can know that all "perl.*" data must be flushed from the
    cache.

    XXX The flushing interface may still need to be fleshed out. This
        basically boils down to deciding who manages the knowledge that
        when "perl" is flushed, "perl.version" must also be flushed.
        ...and how.
"""

import os
import sys
import pprint
import getopt
import cmd

from sysdata.common import SysDataError, log
from sysdata import pdk


#---- globals

_version_ = (0, 1, 0)

# Registry of known datum names to there "getter".
#
# A "getter" can be a raw value, a callable getter, or an exception.
# If a callable the value is expected to be the result of calling the
# getter with one argument: the datum name.
getters = {
    # <name>: <getter>
    "pdk.perlapp": pdk.perlapp,
    "pdk.perlctrl": pdk.perlctrl,
    "pdk.perlsvc": pdk.perlsvc,
    "pdk.perlnet": pdk.perlnet,
    "pdk.perltray": pdk.perltray,
    "pdk.installed": pdk.installed,
    "pdk.haveLicense": pdk.haveLicense,
    "pdk.version": pdk.version,
    "pdk.localHelpFile": pdk.localHelpFile,
    "pdk.webHelpFile": pdk.webHelpFile,
}

# The system data cache.
cache = {
    # <name>: <cached-value>
}



#---- internal support stuff

class _ListCmd(cmd.Cmd):
    """Pass arglists instead of command strings to commands.

    Modify the std Cmd class to pass arg lists instead of command lines.
    This seems more appropriate for integration with sys.argv which handles
    the proper parsing of the command line arguments (particularly handling
    of quoting of args with spaces).
    """
    name = "_ListCmd"
    
    def cmdloop(self, intro=None):
        raise NotImplementedError

    def onecmd(self, argv):
        # Differences from Cmd
        #   - use an argv, rather than a command string
        #   - don't specially handle the '?' redirect to 'help'
        #   - don't allow the '!' shell out
        if not argv:
            return self.emptyline()
        self.lastcmd = argv
        cmdName = argv[0]
        try:
            func = getattr(self, 'do_' + cmdName)
        except AttributeError:
            return self.default(argv)
        try:
            return func(argv)
        except TypeError, ex:
            log.error("%s: %s", cmdName, ex)
            log.error("try '%s help %s'", self.name, cmdName)
            if 1:   # for debugging
                print
                import traceback
                traceback.print_exception(*sys.exc_info())

    def default(self, args):
        log.error("unknown syntax: '%s'", " ".join(args))
        return 1

    def _do_one_help(self, arg):
        try:
            # If help_<arg1>() exists, then call it.
            func = getattr(self, 'help_' + arg)
        except AttributeError:
            try:
                doc = getattr(self, 'do_' + arg).__doc__
            except AttributeError:
                doc = None
            if doc: # *do* have help, print that
                sys.stdout.write(doc + '\n')
                sys.stdout.flush()
            else:
                log.error("no help for '%s'", arg)
        else:
            return func()

    # Technically this improved do_help() does not fit into _ListCmd, and
    # something like this would be more appropriate:
    #    def do_help(self, argv):
    #        cmd.Cmd.do_help(self, ' '.join(argv[1:]))
    # but I don't want to make another class for it.
    def do_help(self, argv):
        if argv[1:]:
            for arg in argv[1:]:
                retval = self._do_one_help(arg)
                if retval:
                    return retval
        else:
            doc = self.__class__.__doc__  # try class docstring
            if doc:
                sys.stdout.write(doc + '\n')
                sys.stdout.flush()
            elif __doc__:  # else try module docstring
                sys.stdout.write(__doc__)
                sys.stdout.flush()

    def emptyline(self):
        # Differences from Cmd
        #   - Don't repeat the last command for an emptyline.
        pass



#---- public module interface

def get(name):
    """Get the named datum.

    <name> is the name of the system datum to get.

    Return the named datum. If the datum does not exist in the cache it
    will be calculated and cached first. If the name is unknown a
    SysDataError will be raised.
    """
    from sysdata import cache, getters
    log.debug("get(name=%r)", name)

    if name in cache:
        log.info("get(): getting '%s' from cache", name)
        return cache[name]
    elif name not in getters:
        raise SysDataError("unknown system data name: '%s'" % name)
    else:
        # The "getter" in the getters dictionary can be a raw value, a
        # callable getter, or an exception.
        log.info("get(): determining '%s'", name)
        getter = getters[name]
        if callable(getter):
            value = getter(name)
        elif isinstance(getter, Exception):
            raise getter
        else:
            value = getter
        cache[name] = value
        return value


def flush(name):
    """Flush the named datum (or data) from the cache.

    <name> is a "name spec" for a system datum or data.

    A "name spec" is a just a simple name or a something of the for
    "<name>.*" to indicate a whole data tree.

    This does not return anything. A SysDataError is raised if the name
    is illegal or if the name is unknown. It is not an error to flush a
    known name that does not happen to be in the cache.
    """
    from sysdata import cache, getters
    log.debug("flush(name=%r)", name)
    if name.endswith(".*"):
        prefix = name[:-1]
        names = [n for n in cache.keys() if n.startswith(prefix)]
        for n in names:
            log.info("flushing '%s' from cache", n)
            del cache[n]
    elif "*" in name:
        raise SysDataError("illegal name for flush(): '%s'" % name)
    elif name not in getters:
        raise SysDataError("unknown system data name: '%s'" % name)
    else:
        if name in cache:
            log.info("flushing '%s' from cache", name)
            del cache[name]



#---- command line interface

class SysDataShell(_ListCmd):
    """
    sysdata - a command-line interface for debugging the sysdata module

    Usage:
        sysdata [<options>...] <command> [<args>...]

    Options:
        -h, --help              Print this help and exit.
        -V, --version           Print the version info and exit.
        -v, --verbose           More verbose output.

    Sysdata's usage is intended to feel like p4's command line
    interface.

    Getting Started:
        sysdata help            Print this help.
        sysdata help <command>  Help on a specific command.

    Commands:
        get <name>              Get the named datum.
        flush <name>            Flush the named datum (or data).
        getters                 Dump the set of getters.
        cache                   Dump the current data cache.
    """
    name = "sysdata"

    def emptyline(self):
        self.do_help(["help"])

    def help_usage(self):
        sys.stdout.write(__doc__)
        sys.stdout.flush()

    def do_get(self, argv):
        """
    get -- Get the named datum.

    sysdata get [<names>...]

        <names> is a list of system data names to get

        Return the named datum. If the datum does not exist in the cache
        it will be calculated and cached first. It is an error if the
        name is unknown.
        """
        try:
            optlist, args = getopt.getopt(argv[1:], "")
        except getopt.GetoptError, ex:
            log.error("get: %s", ex)
            log.error("get: try 'sysdata help get'")
            return 1

        names = args

        for name in names:
            try:
                datum = get(name)
                print "%s: %r" % (name, datum)
            except SysDataError, ex:
                log.error(str(ex))
                if log.isDebugEnabled():
                    import traceback
                    traceback.print_exception(*sys.exc_info())
                return 1

    def do_flush(self, argv):
        """
    flush -- Flush the named datum (or data) from the cache.

    sysdata flush [<names>...]

        <names> is a list of name "specs" of system data to flush.

        You may specify a simple name of a "name spec", for example,
            perl.*
        to flush the whole "perl..." tree. The asterisk is only legal at
        the end of the name and following a ".".
    """
        try:
            optlist, args = getopt.getopt(argv[1:], "")
        except getopt.GetoptError, ex:
            log.error("flush: %s", ex)
            log.error("flush: try 'sysdata help flush'")
            return 1

        names = args

        for name in names:
            try:
                flush(name)
            except SysDataError, ex:
                log.error(str(ex))
                if log.isDebugEnabled():
                    import traceback
                    traceback.print_exception(*sys.exc_info())
                return 1

    def do_cache(self, argv):
        """
    cache -- Dump the current system data cache

    sysdata cache
    """
        try:
            optlist, args = getopt.getopt(argv[1:], "")
        except getopt.GetoptError, ex:
            log.error("cache: %s", ex)
            log.error("cache: try 'sysdata help cache'")
            return 1

        pprint.pprint(cache)

    def do_getters(self, argv):
        """
    getters -- Dump the set of system data getters

    sysdata getters
    """
        try:
            optlist, args = getopt.getopt(argv[1:], "")
        except getopt.GetoptError, ex:
            log.error("getters: %s", ex)
            log.error("getters: try 'sysdata help getters'")
            return 1

        pprint.pprint(getters)



def _main(argv):
    try:
        optlist, args = getopt.getopt(argv[1:], "hVv",
            ["help", "version", "verbose"])
    except getopt.GetoptError, msg:
        log.error("%s. Your invocation was: %s", msg, argv)
        log.error("Try 'sysdata --help'.")
        return 1
    for opt, optarg in optlist:
        if opt in ("-h", "--help"):
            sys.stdout.write(SysDataShell.__doc__)
            return 0
        elif opt in ("-V", "--version"):
            print "sysdata %s" % '.'.join([str(i) for i in _version_])
            return 0
        elif opt in ("-v", "--verbose"):
            log.setLevel(log.DEBUG)

    shell = SysDataShell()
    return shell.onecmd(args)


if __name__ == "__main__":
    __file__ = os.path.abspath(sys.argv[0])
    sys.exit( _main(sys.argv) )


