# Copyright (c) 2003 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

r"""
    Gather system data about PDK (the Perl Dev Kit).

    This is part of the "sysdata" package.

    Data:
        "pdk.perlapp": pdk.perlapp
        "pdk.perlctrl": pdk.perlctrl
        "pdk.perlsvc": pdk.perlsvc
        "pdk.perlnet": pdk.perlnet
        "pdk.perltray": pdk.perltray
        "pdk.installed": pdk.installed
        "pdk.haveLicense": pdk.haveLicense
        "pdk.version": pdk.version
        "pdk.localHelpFile": pdk.localHelpFile
        "pdk.webHelpFile": pdk.webHelpFile

    XXX Currently the "current" PDK is just the first one on the PATH.
        It might be preferable to have the "current" PDK be indicated by
        the currently selected Perl installation.
"""

import os
import sys
import re

import which
import process

from sysdata.common import SysDataError, log


#---- internal support stuff

def _getLicenseAndVersionInfo():
    """Return a tuple (<haveLicense>, <version>)."""
    from sysdata import get

    perlapp = get("pdk.perlapp")
    if perlapp is None:
        return (0, None)
    argv = [perlapp, "--version"]
    p = process.ProcessOpen(argv)
    output = p.stdout.read()
    error = p.stderr.read()
    retval = p.close()
    
    # Annoyingly, perlapp 4.x put the invalid lic warning on stdout.
    licWarn = "You don't seem to have a valid license."
    if error  and error .splitlines()[0].find(licWarn) != -1 or\
       output and output.splitlines()[0].find(licWarn) != -1:
        haveLicense = 0
        version = None
    else:
        haveLicense = 1
        # Parse out the version.
        # Example version line output:
        #   (PerlApp version: 3.0.1 ...
        #   PerlApp 4.0.0 build 401 ...
        #   PerlApp 5.0.3 build 503 ...
        patterns = [
            "PerlApp (?P<version>[0-9.]+)",
            "\(PerlApp version: (?P<version>[0-9.]+)\)",
        ]
        for pattern in patterns:
            match = re.search(pattern, output)
            if match:
                version = match.group("version")
                break
        else:
            version = None

    return (haveLicense, version)



#---- module interface

def perlapp(name):
    try:
        # I hate this generator interface. -- TM
        return which.which("perlapp").next()
    except StopIteration:
        return None

def perlctrl(name):
    try:
        # I hate this generator interface. -- TM
        return which.which("perlctrl").next()
    except StopIteration:
        return None

def perlsvc(name):
    try:
        # I hate this generator interface. -- TM
        return which.which("perlsvc").next()
    except StopIteration:
        return None

def perlnet(name):
    try:
        # I hate this generator interface. -- TM
        return which.which("plc").next()
    except StopIteration:
        return None

def perltray(name):
    try:
        # I hate this generator interface. -- TM
        return which.which("perltray").next()
    except StopIteration:
        return None

def installed(name):
    from sysdata import get
    perlapp = get("pdk.perlapp")
    return perlapp is not None


def haveLicense(name):
    from sysdata import cache
    haveLicense, version = _getLicenseAndVersionInfo()
    cache["pdk.version"] = version
    return haveLicense


def version(name):
    from sysdata import cache
    haveLicense, version = _getLicenseAndVersionInfo()
    cache["pdk.haveLicense"] = haveLicense
    return version


def localHelpFile(name):
    from sysdata import get
    perlapp = get("pdk.perlapp")
    if perlapp is None:
        return None
    installPath = os.path.dirname(os.path.dirname(perlapp))

    candidates = []
    if sys.platform.startswith("win"):
        candidates.append( os.path.join(installPath, "Docs", "PDK",
                                        "PDKUserGuide.chm") )
    candidates.append( os.path.join(installPath, "Docs", "PDK", "HTML",
                                    "index.html") )

    for candidate in candidates:
        if os.path.isfile(candidate):
            return candidate
    else:
        return None


def webHelpFile(name):
    return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-PERLDEVKI_"


