#!/usr/bin/env python

"""
    Relocate an ActivePython installation.

    Usage:
        python reloc_python.py [<options>...] <fromdir> <todir>

    Options:
        -h, --help      Show this help and exit.
    
    Moves a python installation wholesale to a new location.
    Note: No binary and text files editing of the install prefix is currently
          done.  This wil be implemented if the need arises. If binary
          replacement *is* added then:
            assert len(<fromdir>) >= len(<todir>)

    (c) 1999-2001 ActiveState Tool Corp.  All rights reserved.
"""

import os, sys, tempfile, getopt, shutil


#---- utility functions

def _samefile(a, b):
    try:
        return os.path.samefile(a, b)
    except AttributeError:
        pass
    # fall back to comparing the file name string (may not be that robust)
    if sys.platform[:3] == "win":
        # case-insensitive: lowercase everything
        a = a.lower()
        b = b.lower()
    return os.path.normpath(a) == os.path.normpath(b)
        

def _isdir(dirname):
    """os.path.isdir() doesn't work for UNC mount points. Fake it."""
    if sys.platform[:3] == 'win' and dirname[:2] == r'\\':
        if os.path.exists(dirname):
            return os.path.isdir(dirname)
        try:
            os.listdir(dirname)
        except WindowsError:
            return 0
        else:
            return os.path.ismount(dirname)
    else:
        return os.path.isdir(dirname)


def _mkdir(newdir):
    """works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
    """
    if _isdir(newdir):
        pass
    elif os.path.isfile(newdir):
        raise OSError("a file with the same name as the desired " \
                      "dir, '%s', already exists." % newdir)
    else:
        head, tail = os.path.split(newdir)
        if head and not _isdir(head):
            _mkdir(head)
        #print "_mkdir %s" % repr(newdir)
        if tail:
            os.mkdir(newdir)


def _copy(src, dst):
    """works the way a good copy should :)
        - no source, raise an exception
        - destination directory, make a file in that dir named after src
        - source directory, recursively copy the directory to the destination
        - filename wildcarding allowed
    """
    import string, glob, shutil

    assert src != dst, "You are try to copy a file to itself. Bad you! "\
                       "src='%s' dst='%s'" % (src, dst)
    # determine if filename wildcarding is being used
    # (only raise error if non-wildcarded source file does not exist)
    if string.find(src, '*') != -1 or \
       string.find(src, '?') != -1 or \
       string.find(src, '[') != -1:
        usingWildcards = 1
        srcFiles = glob.glob(src)
    else:
        usingWildcards = 0
        srcFiles = [src]

    for srcFile in srcFiles:
        if os.path.isfile(srcFile):
            if usingWildcards:
                srcFileHead, srcFileTail = os.path.split(srcFile)
                srcHead, srcTail = os.path.split(src)
                dstHead, dstTail = os.path.split(dst)
                if dstTail == srcTail:
                    dstFile = os.path.join(dstHead, srcFileTail)
                else:
                    dstFile = os.path.join(dst, srcFileTail)
            else:
                dstFile = dst
            dstFileHead, dstFileTail = os.path.split(dstFile)
            if dstFileHead and not _isdir(dstFileHead):
                _mkdir(dstFileHead)
            if _isdir(dstFile):
                dstFile = os.path.join(dstFile, os.path.basename(srcFile))
            #print "copy %s %s" % (srcFile, dstFile)
            shutil.copy(srcFile, dstFile)
        elif _isdir(srcFile):
            srcFiles = os.listdir(srcFile)
            if not os.path.exists(dst):
                _mkdir(dst)
            for f in srcFiles:
                s = os.path.join(srcFile, f)
                d = os.path.join(dst, f)
                try:
                    _copy(s, d)
                except (IOError, os.error), why:
                    raise OSError("Can't copy %s to %s: %s"\
                                  % (repr(s), repr(d), str(why)))
        elif not usingWildcards:
            raise OSError("Source file %s does not exist" % repr(srcFile))


def relocate(fromPath, toPath):
    print "Relocating Python from '%s' to '%s'.\n" % (fromPath, toPath)

    expandedFromPath = os.path.expanduser(fromPath)
    expandedToPath = os.path.expanduser(toPath)

    if os.path.isdir(expandedFromPath) and os.path.isdir(expandedToPath)\
       and _samefile(expandedFromPath, expandedToPath):
        print "<frompath>, '%s', and <topath>, '%s', are the same. Nothing "\
              "to do." % (fromPath, toPath)
        return 0
    elif not os.path.isdir(expandedFromPath):
        sys.stderr.write("The given source directory, '%s', does not "\
                         "exist. Aborting." % fromPath)
        return 1
    elif os.path.exists(expandedToPath):
        sys.stderr.write("The given destination directory, '%s', already "\
                         "exists. You must first remove the target dir "\
                         "or select an alternate location. Aborting."\
                         % toPath)
        return 1

    # Copy the files.
    _copy(expandedFromPath, expandedToPath)



def main(argv):
    # process the options
    try:
        optlist, args = getopt.getopt(argv[1:], 'h', ['help'])
    except getopt.GetoptError, ex:
        msg = "%s: ERROR: %s\n" % (argv[0], ex)
        sys.stderr.write(msg)
        sys.stderr.write(__doc__)
        return 1
    for opt, optarg in optlist:
        if opt in ('-h', '--help'):
            print __doc__
            return 0

    # get arguments
    if len(args) != 2:
        sys.stderr.write("%s: ERROR: incorrect number of arguments\n"\
                         % argv[0])
        sys.stderr.write(__doc__)
        return 1
    fromPath, toPath = args

    return relocate(fromPath, toPath)


if __name__ == '__main__':
    sys.exit( main(sys.argv) )
